/* Main controller for the FB execution engine
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.controller;

import java.util.Set;

import com.bbn.morph.evaluator.Evaluator;
import com.bbn.morph.evaluator.MainEvaluationController;
import com.bbn.morph.executer.FBExecutor;
import com.bbn.morph.executer.FBModel;
import com.bbn.morph.executer.StressInputs;
import com.bbn.morph.executer.Update;
import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.model.DesiredPhysicalModel;
import com.bbn.morph.model.InvalidModelException;
import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.PhysicalModel;
import com.bbn.morph.model.RobotAttributeValues;
import com.bbn.morph.model.Units;
import com.bbn.morph.ui.InformationReporter;
import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.DesiredAttributeValueView;
import com.bbn.morph.view.RobotAttribute;

public class Controller {

  private Evaluator m_evaluationControl;
  private FBExecutor m_fbExecutor;
  private AttributeValueView m_view;
  private FBModel m_fbModel;
  private DesiredAttributeValueView m_desiredRobotView;
  private ExecuterController m_updateController;
  private CompositeModel m_compositeModel;
  private InformationReporter m_reporter = null;

  

  public Controller(CompositeModel compositeModel) throws InvalidModelException {
    m_compositeModel = compositeModel;
    m_view = new AttributeValueView(compositeModel.getRobotModel());
    m_fbModel = compositeModel.getFbModel();
    m_desiredRobotView = new DesiredAttributeValueView(compositeModel.getDesiredRobotModel());
    m_evaluationControl = new MainEvaluationController(compositeModel, m_view, m_desiredRobotView);
    m_fbExecutor = new FBExecutor(m_view, m_fbModel);
    m_updateController = new ExecuterController(compositeModel.getRobotModel(), compositeModel.getEnvironmentModel());
  }
  
  /**
   * @param reporter the information reporter to use to report information
   */
  public void setInformationReporter(InformationReporter reporter) {
    m_reporter = reporter;
    m_fbExecutor.setInformationReporter(reporter);
  }
  
  public boolean emptyUpdates(Set<Update> updates) {
	  for (Update update : updates) {
	    if (update.getVal().getStandardValue() != 0)
	      return false;
	  }
	  return true;
	}
  
  /**
   * @return true iff updates have been made
   */
  public boolean runIteration() {
    Set<StressInputs> stressInputs = m_evaluationControl.getStressInputs();
    Set<Update> updates = m_fbExecutor.run(stressInputs);
    if (emptyUpdates(updates))
      return false;
    m_updateController.update(updates);
    if (m_reporter != null)
      m_reporter.reportInformation(this.toString());
    else
      System.out.println(this.toString());
    return true;
  }
  
//  public void changeRobotModel(RobotAttribute a, double finalValue ){
//    m_compositeModel.getDesiredRobotModel().addDesiredAttribute(new RobotAttributeValues(a.getName(), finalValue));
//	  m_fbModel.addFB(a,finalValue, m_view.get(a), m_compositeModel.getRobotModel());
//  }
  
  public void changeDesiredModel(DesiredPhysicalModel desiredModel, PhysicalModel model, RobotAttribute a, Measurement newValue) {
    desiredModel.addDesiredAttribute(new RobotAttributeValues(a.getName(), newValue));
    Measurement initialValue = model.getAttributeValueView().get(a);
    m_fbModel.addFB(a, newValue, initialValue, model);
  }
  
  /**
   * Remove the specified attribute from the desired model and the FB model.
   * @param desiredModel the desired model to modify
   * @param a the attribute to remove
   */
  public void removeDesiredModel(DesiredPhysicalModel desiredModel, RobotAttribute a) {
    desiredModel.removeDesiredAttribute(new RobotAttributeValues(a.getName(), new Measurement(Units.getNone(), 0)));
    m_fbModel.removeFB(a);
  }
  
  
  @Override
  public String toString() {
	  m_view.refresh();
    return m_view.toString();
  }
  
  
  /**
   * @param maxIterations the maximum number of iterations to run through
   * @return true iff equilibrium is reached within maxIterations
   */
  public boolean runToEquilibrium(int maxIterations) {
    for (int i=1; i<=maxIterations; i++) {
      boolean changed = runIteration();
      if (!changed)
        return true;
    }

    return false; // equilibrium never reached
  }
  
}
