/** Functional Blueprint Evaluator
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. 
**/

package com.bbn.morph.executer;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.Units;
import com.bbn.morph.ui.InformationReporter;
import com.bbn.morph.view.FB;
import com.bbn.morph.view.RobotAttribute;
import com.bbn.morph.view.AttributeValueView;


/**
 * @author fyaman
 *
 */
public class FBExecutor {
  AttributeValueView robotModel;
  FBModel fbModel;
  private InformationReporter m_reporter = null;
  private List<Double> stressHistory= new LinkedList<Double>();
  private boolean isSilent=false;
  private List<Double> partialStressHistory= new LinkedList<Double>();
  
  private int m_iteration = 0;

  public FBExecutor(AttributeValueView m, FBModel fbModel) {
    robotModel = m;
    this.fbModel = fbModel;
  }

  /**
   * @param reporter the information reporter to use to report information
   */
  public void setInformationReporter(InformationReporter reporter) {
    m_reporter = reporter;
  }

  public AttributeValueView getRobotModel() {
    return robotModel;
  }
  public FBModel getFbModel() {
    return fbModel;
  }

  public boolean emptyUpdates(Set<Update> updates) {
    for (Update update : updates) {
      if (update.getVal().getStandardValue() != 0)
        return false;
    }
    return true;
  }

  /**
   * Main function that interprets the evaluator outputs 
   * and adjusts the robot model based on FB stress. The robot
   * model is modified
   * @param inputs
   * @return true iff there were updates performed
   */
  public Set<Update> run(Set<StressInputs> inputs){
    m_iteration++;
    robotModel.refresh();
    // Compute the stress on each FB
    Map<FB, Double >  fbStress = compoundStress(inputs);
    
    double maxNonUserStress = 0.0;
    for (FB fb : fbStress.keySet()) {
      if (!fb.isUserInduced())
        maxNonUserStress = Math.max(maxNonUserStress, Math.abs(fbStress.get(fb)));
    }
    
    // This modulation is based on the idea that we want to turn off the
    // user induced FBs if ANY absolute value of a non-user induced FB stress is 
    // 0.5 or above. This function linearly decreases the effect of the user
    // induced FBs as the stress value rises toward 0.5. 
    double userInducedModulation = Math.max(0, (1 - (2 * maxNonUserStress)));

    printStress(fbStress);
    reportStress(fbStress);

    // Compute the suggested updates by each FB
    Map<RobotAttribute, Set<Update>> singleFBUpdates = fbUpdates(fbStress, userInducedModulation);
    // Combine the suggestions and produce one update per FB
    Set<Update>  compositeFBUpdate= compoundUpdate(singleFBUpdates);

    return compositeFBUpdate;
  }
  
  public double getTotalStress(Set<StressInputs> inputs){
	  robotModel.refresh();
	  // Compute the stress on each FB
	  Map<FB, Double >  fbStress = compoundStress(inputs);
	  double totalStress= 0;
	  for(FB fb: fbStress.keySet()){
		  String id = fb.getFeature().getName();
		  Double stress = fbStress.get(fb);
		  totalStress += Math.abs(stress);
	  }
	  return totalStress;
  }

  private void reportStress(Map<FB, Double> fbStress) {
    double totalStress= 0;
    for(FB fb: fbStress.keySet()){
      String id = fb.getFeature().getName();
      Double stress = fbStress.get(fb);
      if (m_reporter!= null)
        m_reporter.reportStress(m_iteration, id, stress);

      totalStress += Math.abs(stress);
    }

    if (m_reporter!= null)
      m_reporter.reportStress(m_iteration, "Total Stress", totalStress);
  }

  private void printStress(Map<FB, Double >  fbStress) {
    StringBuffer buf = new StringBuffer();
    double totalStress= 0;
    double partialStress=0;
    for(FB fb: fbStress.keySet()){
      buf.append("Stress for " + fb.getFeature() + " : " + fbStress.get(fb)+ " ; ");
      // FIXME this could probably use the fb.isUserInduced() test instead of the name test:
      if (!fb.getFeature().getName().startsWith("FB_"))
        partialStress +=Math.abs(fbStress.get(fb));
      totalStress +=Math.abs(fbStress.get(fb));
    }
    if (m_reporter!= null)
      m_reporter.reportInformation("Total System Stress: "+ totalStress + " --- " + buf);
    else if (!isSilent)
      System.out.println("Total System Stress: "+ totalStress + " --- " + buf);
    stressHistory.add(totalStress);
    partialStressHistory.add(partialStress);
  }
  /**
   * Main function that interprets the evaluator outputs 
   * and adjusts the robot model based on FB stress. The robot
   * model is modified
   * @param inputs
   * @return true iff there were updates performed
   */
  public boolean localRun(Set<StressInputs> inputs){
    m_iteration++;
    Set<Update>  compositeFBUpdate= run(inputs);
    if (emptyUpdates(compositeFBUpdate))
      return false;
    //Update the local model
    robotModel.update(compositeFBUpdate);
    return true;
  }


  /**
   *  Computes the stress for each input set and then combines the stress for each FB
   * @param inputs
   * @return one stress value per FB
   */
  public Map<FB, Double> compoundStress(Set<StressInputs> inputs) {
    Map<FB, List<Double>> stressSets = new HashMap<FB, List<Double>>();
    for (StressInputs in: inputs){
      FB fb= fbModel.get(in.getFeature());
      Double fbStress = fb.stress(in);
      List<Double> oldSet = stressSets.get(fb);
      if (oldSet==null)
        oldSet= new LinkedList<Double>();
      oldSet.add(fbStress);
      stressSets.put(fb, oldSet);
    }
    Map<FB, Double> stress = new HashMap<FB, Double>();
    for (FB key: stressSets.keySet())
      stress.put(key, blend(stressSets.get(key)));
    return stress;
  }

  /** This is the function that needs to be changed for implementing a  
   * different stress combination mechanism.
   * @param old - previous stress value
   * @param fbStress - new  stress value
   * @return maximum of the old and new stress
   */
  public Double combineStress(Double oldStress, Double newStress) {
    if (oldStress< newStress)
      return newStress;
    else
      return oldStress;
  }


  /** Computes all suggested updates to the model
   * @param fbStress - the stress on each functional blue print
   * @param userInducedModulation the factor to modulate all user induced functional blueprints by. Ranges from 1 (no change) to 0 (no updates from user induced FBs).
   * @return the updates recommended by each FB - can have more than one update for an attribute
   */
  public Map<RobotAttribute,Set<Update>> fbUpdates(Map<FB, Double> fbStress, double userInducedModulation) {
    Map<RobotAttribute,Set<Update>>  allUpdates = new HashMap<RobotAttribute,Set<Update>>();
    for(FB fb: fbStress.keySet()){
      Set<Update> singleUpdates = fb.updates(fbStress.get(fb));
      
      for(Update u: singleUpdates){
        // modulate the update if appropriate
        if (fb.isUserInduced() && (userInducedModulation != 1.0)) {
          u.modulate(userInducedModulation);
        }
        
        Set<Update> attributeUpdates= allUpdates.get(u.getAttribute());
        if (attributeUpdates==null)
          attributeUpdates=new HashSet<Update>();
        attributeUpdates.add(u);
        allUpdates.put(u.getAttribute(), attributeUpdates);
      }
    }
    return allUpdates;
  }


  /** Outer function that for all attributes combines the recommended updates.
   * @param singleFBUpdates - updates suggested by each FB grouped by attributes that relate to
   * @return a set of updates which does not contain more than one update per attribute
   */
  public Set<Update> compoundUpdate(Map<RobotAttribute, Set<Update>> singleFBUpdates) {
    Set<Update> compositeUpdates = new HashSet<Update>();
    for (RobotAttribute key: singleFBUpdates.keySet()){
      Set<Update> updates = singleFBUpdates.get(key);
      List<Measurement> values = updateValues(updates);
      Measurement val = blendM(values);
      Update first = updates.iterator().next();
      Update compUpdate= new Update(first.getAttribute(), val);
      compositeUpdates.add(compUpdate);
    }
    return compositeUpdates;
  }

  /** This is the function that needs to be changed for implementing a  
   * different blending of FB updates to combine the multiple update suggestions.
   * @param updates - set of all proposed updates for a specific robot attribute
   * @return - single update construct for the attribute
   */
  public List<Measurement> updateValues(Set<Update> updates) {
    if (updates==null || updates.isEmpty())
      return null;
    List<Measurement>  values = new LinkedList<Measurement>();
    for(Update u: updates){
      values.add(u.getVal());
    }
    return values;
  }


  /** Generic Blending function: 
   * 
   * deltaP_i = sign(sum deltaP_ij).max(dP_ij. (sum deltaP_ij)/(sum |deltaP_ij|)
   *  
   * @param updates - set of all proposed updates for a specific robot attribute
   * @return - single update construct for the attribute
   */
  
  static public Measurement blendM(List<Measurement> values) {
	  if(values==null || values.isEmpty())
		  return null;
	  double absSum =0;
	  double sum=0;
	  Measurement positiveMax = new Measurement(Units.getMeters(), 0.0);
	  Measurement negativeMax = new Measurement(Units.getMeters(), 0.0);
	  for(Measurement d: values){
		  sum=d.getStandardValue()+sum;
		  absSum=Math.abs(d.getStandardValue())+absSum;
		  if(d.getStandardValue()>0)
			  positiveMax= Measurement.max(d, positiveMax);
		  if(d.getStandardValue()<0)
			  negativeMax= Measurement.min(d, negativeMax);
	  }
	  Measurement result = new Measurement(Units.getMeters(), 0.0);
	    if (sum > 0)
	      result= new Measurement(positiveMax.getUnits(), positiveMax.getStandardValue() * sum / absSum) ;
	    if (sum < 0)
	      result= new Measurement(negativeMax.getUnits(), negativeMax.getStandardValue() * Math.abs(sum) / absSum) ;

	    return result;
  }

  static public Double blend(List<Double> values) {
    if (values==null || values.isEmpty())
      return null;
    double absSum =0;
    double sum=0;
    double positiveMax=0;
    double negativeMax =0;
    for(Double d: values){
      sum+=d;
      absSum+= Math.abs(d);
      if (d > 0)
        positiveMax= Math.max(d,positiveMax);
      if (d < 0)
        negativeMax= Math.min(d,negativeMax);
    }
    double result = 0.0;
    if (sum > 0)
      result= positiveMax * sum / absSum ;
    if (sum < 0)
      result= negativeMax * Math.abs(sum) / absSum ;

    return result;
  }
  
  @Override
  public String toString() {
    return "FBExecutor [fbModel=" + fbModel + ", robotModel=" + robotModel
    + "]";
  }

  public List<Double> getStressHistory() {
    return stressHistory;
  }

  public  void resetStressHistory() {
    stressHistory.clear();
  }

  public boolean isSilent() {
    return isSilent;
  }

  public void setSilent(boolean isSilent) {
    this.isSilent = isSilent;
  }

  public List<Double> getPartialStressHistory() {
    return partialStressHistory;
  }
  
}
