/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;

/**
 * A collection of ComponentInstances
 * 
 * @author aadler
 *
 */
public class ComponentLibrary {

  private List<ComponentInstance> m_libraryComponents;
  private String m_name;
  
  private List<Header> m_headers;
  
  // FIXME we need to figure out a way to select the "acceptable" ComponentInstances
  private double MASS_RANGE = 0.003; // in kg
  
  
  /**
   * Read a {@link ComponentLibrary} in from the file. The file should be in 
   * csv format with the first line a set of headers containing one field called "name"
   * For example the following is a sample file format:
   * name,lengthx(meters),mass(grams)
   * flipper1,0.20,50
   * @param libraryFile the file to read the component library from.
   * @throws IOException if an exception occurs reading the file
   */
  public ComponentLibrary(File libraryFile) throws IOException {
    m_libraryComponents = new ArrayList<ComponentInstance>();
    m_headers = new ArrayList<Header>();
    
    // first get the library name
    String name = libraryFile.getName();
    m_name = name.substring(0, name.lastIndexOf("."));
    
    // now the first line is the key -- it tells us what attributes the components have
    Scanner lineScanner = new Scanner(libraryFile);
    if (lineScanner.hasNextLine())
      parseHeader(lineScanner.nextLine());
    while(lineScanner.hasNextLine()) 
      parseComponent(lineScanner.nextLine());
  }
  
  private void parseHeader(String headerLine) {
    Scanner nameScanner = new Scanner(headerLine).useDelimiter("\\s*,\\s*");
    while (nameScanner.hasNext()) {
      String name = nameScanner.next();
      // it could have a unit type with it
      Units unit = Units.getNone();
      int openP = name.indexOf("(");
      if (openP != -1) {
        String unitStr = name.substring(openP+1, name.lastIndexOf(")"));
        unit = Units.getUnit(unitStr);
        name = name.substring(0, openP);
      }
      m_headers.add(new Header(name, unit));
    }
  }
  
  private void parseComponent(String componentLine) {
    Scanner componentScanner = new Scanner(componentLine).useDelimiter("\\s*,\\s*");
    
    String componentName = "";
    List<RobotAttributeValues> attributes = new ArrayList<RobotAttributeValues>();
    
    for (int i=0; i<m_headers.size(); i++) {
      if (!componentScanner.hasNext()) {
        System.err.println("While component library: "+m_name+", malformed line: "+componentLine);
      }
      Header h = m_headers.get(i);
      if (h.getName().equals("name")) {
        componentName = componentScanner.next();
      }
      else {
        // it's a number:
        double d = componentScanner.nextDouble();
        attributes.add(new RobotAttributeValues(h.getName(), new Measurement(h.getUnits().copy(), d)));
      }
    }
    
    // now create the ComponentInstance
    ComponentInstance component = new ComponentInstance(componentName, attributes);
    // add it to the library
    m_libraryComponents.add(component);
  }
  
  /**
   * @return the name of the component library
   */
  public String getName() {
    return m_name;
  }
  
  /**
   * Return a list of components that could be used for the specified baseBody
   * @param baseBody the ModelBody to find acceptable components for
   * @return a list of acceptable components
   */
  public List<ComponentInstance> getComponents(ModelBody baseBody) {
    List<ComponentInstance> acceptable = new ArrayList<ComponentInstance>();
    
    for (ComponentInstance candidate : m_libraryComponents) {
      boolean candidateOkay = true;
      
      for (RobotAttributeValues value : candidate.getAttributes()) {
        String attr = value.getAttribute();
        double diff = value.getMeasurement().getStandardValue() - 
            baseBody.getAttributeValue(attr).getStandardValue();
        //System.out.println(value.getValue().getStandardValue()+" "+baseBody.getAttributeValue(attr).getStandardValue());
        //System.out.println(candidate.getName()+" "+attr+" "+diff);
        if (attr.equals("mass")){
          if (Math.abs(diff) > MASS_RANGE)
            candidateOkay = false;
        }
        if (attr.equals("lengthz")) {
          // FIXME maybe we want to change this later
          // only longer flippers will work! 
          if (diff < 0)
            candidateOkay = false;
        }
        // we have to have more torque than we need
        if (attr.equals("torque")) {
          if (diff < 0)
            candidateOkay = false;
        }
      }
      
      if (candidateOkay)
        acceptable.add(candidate);
    }
    
    // if we have no acceptable candidates, that's a runtime exception
    // FIXME be more graceful!
    if (acceptable.isEmpty()) {
      System.err.println("There were no acceptable candidate parts for "+baseBody.getId()+" in component library "+getName()+".");
      throw new RuntimeException("There were no acceptable candidate parts for "+baseBody.getId()+" in component library "+getName()+".");
    }
    
    System.out.println("trying these components: ");
    for (ComponentInstance componentInstance : acceptable) {
      System.out.println(componentInstance.getName());
    }
    
    // return the acceptable candidates
    return acceptable;
  }
  
  /**
   * A class to store the header info
   * @author aadler
   *
   */
  private class Header {
    private String m_name;
    private Units m_units;
    
    public Header(String name, Units units) {
      m_name = name;
      m_units = units;
    }
    
    public String getName() {
      return m_name;
    }
    
    public Units getUnits() {
      return m_units;
    }
  }
  
}
