/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */

package com.bbn.morph.model;

import java.util.List;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Represents a cylinder in our model
 * @author aadler
 *
 */
public class ModelCylinder extends ModelBody {

  protected final static String TYPE_NAME = "cylinder";
  private Measurement m_cylinderLength;
  private Measurement m_cylinderRadius;  
  
  /**
   * Creates a cylinder
   * @param id the id 
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position
   * @param rotation the rotation, can be null
   * @param cylinderRadius the cylinder radius
   * @param cylinderLength the cylinder length
   * @param mass the mass
   */
  public ModelCylinder(String id, String libraryName, ModelPosition position, ModelRotation rotation, 
      Measurement cylinderRadius, Measurement cylinderLength, Measurement mass) {
    super(id, libraryName, position, rotation, mass);
    m_cylinderLength = cylinderLength.copy();
    m_cylinderRadius = cylinderRadius.copy();
    // FIXME not sure if I want to update the cylinder to be in the right position or if I just care that it has the right dimensions?
  }
  
  @Override
  public ModelObject copy() {
    ModelRotation rotation = (getRotation() == null)?null:getRotation().copy();
    ModelCylinder cylinder = new ModelCylinder(getId(), getLibraryName(), getPosition().copy(), rotation,
        getRadius(), getLength(), getMass());
    cylinder.addExtraParameters(getExtraParameters());
    return cylinder;
  }
  
  /* (non-Javadoc)
   * @see com.bbn.morph.model.ModelBody#clone()
   */
  public ModelCylinder clone() {
    ModelRotation rotation = (getRotation() == null)?null:getRotation().copy();

    ModelCylinder cylinder =  new ModelCylinder(getId(), getLibraryName(), getPosition().copy(), 
        rotation, getRadius(), getLength(), getMass());
    cylinder.setParentObject(getParentObject());
    cylinder.addExtraParameters(getExtraParameters());
    return cylinder;
  }

  @Override
  public Measurement getAttributeValue(String attribute) {
    if (attribute.equals("length"))
      return getLength();
    else if (attribute.equals("radius"))
      return getRadius();
    else
      // try our super class:
      return super.getAttributeValue(attribute);
  }
  
  @Override
  public Set<RobotAttributeValues> getAttributeValues() {
    Set<RobotAttributeValues> rav = super.getAttributeValues();
    // set the values in meters
    rav.add(new RobotAttributeValues(getId()+".length", m_cylinderLength.getValueIn(UnitPrefixes.NONE)));
    rav.add(new RobotAttributeValues(getId()+".radius", m_cylinderRadius.getValueIn(UnitPrefixes.NONE)));
    return rav;
  }
  
  public Measurement getLength() {
	  return m_cylinderLength.copy();
  }
  
  public Measurement getRadius() {
	  return m_cylinderRadius.copy();
  }
  
  @Override
  public void updateRobotAttributeFromDeltas(RobotAttributeValues robotAttributeValues, List<ModelJoint> joints) {
    // TODO joints not handled yet
    // handle what we can handle:
    String attribute = robotAttributeValues.getAttribute();
    Measurement delta = robotAttributeValues.getMeasurement();
    
    if (attribute.equals("length"))
      m_cylinderLength = getLength().add(delta);
    else if (attribute.equals("radius"))
      m_cylinderRadius = getRadius().add(delta);
    else
      // try our super class:
      super.updateRobotAttributeFromDeltas(robotAttributeValues, joints);
  }
  
  /**
   * Creates a {@link ModelCylinder} from the given XML element
   * @param element the XML element representing a {@link ModelCylinder}
   * @param id the id
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position of the cylinder
   * @param rotation the rotation of the cylinder, can be null
   * @param mass the mass
   * @param tagNames the set of tag names that have not been handled (tags removed as they are handled)
   * @return the {@link ModelCylinderBox} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelBody createFromXML(Element element, String id, String libraryName,
      ModelPosition position, ModelRotation rotation, Measurement mass, Set<String> tagNames) throws InvalidModelException {
    
    try {
      NodeList radiusNodes = element.getElementsByTagName("radius");
      if (radiusNodes.getLength() != 1)
        throw new InvalidModelException("Did not have one and only one radius");
      double radius = Double.parseDouble(((Element)radiusNodes.item(0)).getTextContent());

      // FIXME should probably be length
      NodeList lengthNodes = element.getElementsByTagName("height");
      if (lengthNodes.getLength() != 1)
        throw new InvalidModelException("Did not have one and only one length");
      double length = Double.parseDouble(((Element)lengthNodes.item(0)).getTextContent());

      Measurement radiusM = new Measurement(Units.getMeters(), radius);
      Measurement lengthM = new Measurement(Units.getMeters(), length);
      
      tagNames.remove("radius");
      tagNames.remove("height");
      
      return new ModelCylinder(id, libraryName, position, rotation, radiusM, lengthM, mass);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }
  
  @Override
  public Element toXMLElement(Document doc) {
    Element baseElement = super.toXMLElement(doc);
    baseElement.setAttribute("type", TYPE_NAME);
    
    Element radius = doc.createElement("radius");
    radius.setTextContent(Double.toString(getRadius().getStandardValue()));
    baseElement.appendChild(radius);
    
    // FIXME should be length
    Element length = doc.createElement("height");
    length.setTextContent(Double.toString(getLength().getStandardValue()));
    baseElement.appendChild(length);
    
    return baseElement;
  }
}
