/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Represents a slider joint
 * @author aadler
 *
 */
public class ModelSliderJoint extends ModelJoint {
  
  protected static final String TYPE_NAME = "sliderJoint";
  
  private Double m_axisVx;
  private Double m_axisVy;
  private Double m_axisVz;
  private Double m_lowStop;
  private Double m_highStop;
  
  /**
   * Creates a slider joint
   * @param bodyId1 the id of the first body
   * @param bodyId2 the id of the second body
   * @param axisVx the axis x position
   * @param axisVy the axis y position
   * @param axisVz the axis z position
   * @param lowStop the low stop
   * @param highStop the high stop
   */
  public ModelSliderJoint(String bodyId1, String bodyId2, Double axisVx, 
      Double axisVy, Double axisVz, Double lowStop, Double highStop) {
    super(bodyId1, bodyId2);
    m_axisVx = axisVx;
    m_axisVy = axisVy;
    m_axisVz = axisVz;
    m_lowStop = lowStop;
    m_highStop = highStop;
  }
  
  @Override
  public ModelObject copy() {
    return new ModelSliderJoint(getBodyId1(), getBodyId2(), getAxisVx(), 
        getAxisVy(), getAxisVz(), getLowStop(), getHighStop());
  }
  
  public Double getAxisVx() {
    return m_axisVx;
  }
  
  public Double getAxisVy() {
    return m_axisVy;
  }
  
  public Double getAxisVz() {
    return m_axisVz;
  }
  
  public Double getLowStop() {
    return m_lowStop;
  }
  
  public Double getHighStop() {
    return m_highStop;
  }
  
  /**
   * Creates a {@link ModelSliderJoint} from the given XML element
   * @param element the XML element representing a {@link ModelSliderJoint}
   * @param id1 the first id
   * @param id1 the second id
   * @return the {@link ModelSliderJoint} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelJoint createFromXML(Element element, String id1, 
      String id2) throws InvalidModelException {
    NodeList axisNodes = element.getElementsByTagName("axis");
    if (axisNodes.getLength() != 1)
      throw new InvalidModelException("Did not have one and only one axis");
    Element axis = ((Element)axisNodes.item(0));
    
    // FIXME are these required???
    
    NodeList lowNodes = element.getElementsByTagName("lowStop");
    Element lowNode = null;
    if (lowNodes.getLength() == 1)
      lowNode = ((Element)lowNodes.item(0));

    NodeList highNodes = element.getElementsByTagName("highStop");
    Element highNode = null;
    if (highNodes.getLength() != 1)
      highNode = ((Element)highNodes.item(0));

    try {
      if (!axis.hasAttribute("x") || !axis.hasAttribute("y") ||
          !axis.hasAttribute("z"))
        throw new InvalidModelException("Missing required anchor attribute");
      double axisX = Double.parseDouble(axis.getAttribute("x"));
      double axisY = Double.parseDouble(axis.getAttribute("y"));
      double axisZ = Double.parseDouble(axis.getAttribute("z"));
      
      Double lowStop = (lowNode == null)?null:Double.parseDouble(lowNode.getTextContent());
      Double highStop = (highNode == null)?null:Double.parseDouble(highNode.getTextContent());
    
      return new ModelSliderJoint(id1, id2, axisX, axisY, axisZ, lowStop, highStop);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }

  @Override
  public Element toXMLElement(Document doc) {
    Element baseElement = super.toXMLElement(doc);
    baseElement.setAttribute("type", TYPE_NAME);
    
    Element axis = doc.createElement("axis");
    axis.setAttribute("x", Double.toString(getAxisVx()));
    axis.setAttribute("y", Double.toString(getAxisVy()));
    axis.setAttribute("z", Double.toString(getAxisVz()));
    baseElement.appendChild(axis);

    if (m_lowStop != null) {
      Element lowStop = doc.createElement("lowStop");
      lowStop.setTextContent(Double.toString(m_lowStop));
      baseElement.appendChild(lowStop);
    }
    
    if (m_highStop != null) {
      Element highStop = doc.createElement("highStop");
      highStop.setTextContent(Double.toString(m_highStop));
      baseElement.appendChild(highStop);
    }
    
    return baseElement;
  }

}
