/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import java.util.List;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Represents a sphere in our model
 * @author aadler
 *
 */
public class ModelSphere extends ModelBody {

  protected final static String TYPE_NAME = "sphere";
  private Measurement m_radius;
  
  /**
   * Create a sphere 
   * @param id the id
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position
   * @param rotation the rotation, can be null
   * @param radius the sphere radius
   * @param mass the mass
   */
  public ModelSphere(String id, String libraryName, ModelPosition position, ModelRotation rotation, 
      Measurement radius, Measurement mass) {
    super(id, libraryName, position, rotation, mass);
    m_radius = radius.copy();
    // FIXME not sure if I want to update the sphere to be in the right position or if I just care that it has the right dimensions?
  }
  
  @Override
  public ModelObject copy() {
    ModelRotation rotation = (getRotation() == null)?
    		  null:getRotation().copy();
    ModelSphere sphere = new ModelSphere(getId(), getLibraryName(), getPosition().copy(), rotation,
        m_radius, getMass());
    sphere.addExtraParameters(getExtraParameters());
    return sphere;
  }
  
  /* (non-Javadoc)
   * @see com.bbn.morph.model.ModelBody#clone()
   */
  public ModelBody clone() {
    ModelRotation rotation = (getRotation() == null)?null:getRotation().copy();
    ModelSphere sphere = new ModelSphere(getId(), getLibraryName(), getPosition().copy(), 
        rotation, m_radius, getMass());
    sphere.setParentObject(getParentObject());
    sphere.addExtraParameters(getExtraParameters());
    return sphere;
  }
  
  @Override
  public Measurement getAttributeValue(String attribute) {
    if (attribute.equals("radius"))
      return m_radius.copy();
    else
      // try our super class:
      return super.getAttributeValue(attribute);
  }
  
  @Override
  public Set<RobotAttributeValues> getAttributeValues() {
    Set<RobotAttributeValues> rav = super.getAttributeValues();
    rav.add(new RobotAttributeValues(getId()+".radius", m_radius.copy()));
    return rav;
  }
  
  @Override
  public void updateRobotAttributeFromDeltas(RobotAttributeValues robotAttributeValues,
      List<ModelJoint> joints) {
    // TODO joints not handled yet
    // handle what we can handle:
    String attribute = robotAttributeValues.getAttribute();
    Measurement delta = robotAttributeValues.getMeasurement();
    
    if (attribute.equals("radius"))
      m_radius = m_radius.add(delta);
    else
      // try our super class:
      super.updateRobotAttributeFromDeltas(robotAttributeValues, joints);
  }
  
  /**
   * Creates a {@link ModelSphere} from the given XML element
   * @param element the XML element representing a {@link ModelSphere}
   * @param id the id
   * @param libraryName the library name for this component or an empty string if none
   * @param position the position of the sphere
   * @param rotation the rotation of the sphere, can be null
   * @param mass the mass
   * @param tagNames the set of tag names that have not been handled (tags removed as they are handled)
   * @return the {@link ModelSphere} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelBody createFromXML(Element element, String id, String libraryName,
      ModelPosition position, ModelRotation rotation, 
      Measurement mass, Set<String> tagNames) throws InvalidModelException {
    
    try {
      NodeList radiusNodes = element.getElementsByTagName("radius");
      if (radiusNodes.getLength() != 1)
        throw new InvalidModelException("Did not have one and only one radius");
      double radius = Double.parseDouble(((Element)radiusNodes.item(0)).getTextContent());
      Measurement radiusM = new Measurement(Units.getMeters(), radius);
      
      tagNames.remove("radius");
      
      return new ModelSphere(id, libraryName, position, rotation, radiusM, mass);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }
  
  @Override
  public Element toXMLElement(Document doc) {
    Element baseElement = super.toXMLElement(doc);
    baseElement.setAttribute("type", TYPE_NAME);
    
    Element radius = doc.createElement("radius");
    radius.setTextContent(Double.toString(m_radius.getStandardValue()));
    baseElement.appendChild(radius);
    
    return baseElement;
  }
}
