/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Represents a transform in our model
 * @author aadler
 *
 */
public class ModelTransform extends ModelObject {

  private List<ModelObject> m_transformTargets; // OR is this a list??
  private double m_xShift;
  private double m_yShift;
  private double m_zShift;
  
  /**
   * Creates a transform
   * @param xShift the x shift
   * @param yShift the y shift
   * @param zShift the z shift
   */
  public ModelTransform(double xShift, double yShift, double zShift) {
    m_xShift = xShift;
    m_yShift = yShift;
    m_zShift = zShift;
    m_transformTargets = new ArrayList<ModelObject>();
  }
  
  @Override
  public ModelObject copy() {
    return new ModelTransform(getXShift(), getYShift(), getZShift());
  }
  
  /**
   * @param mo the model object to add to the transform
   */
  public void addTransformTarget(ModelObject mo) {
    m_transformTargets.add(mo);
    mo.setParentObject(this);
    //FIXME sorting?
  }
  

  /**
   * @param modelObjects add all of the model objects to the transform
   */
  public void addTransformTargets(List<ModelObject> modelObjects) {
    for (ModelObject modelObject : modelObjects) {
      addTransformTarget(modelObject);
    }
    // FIXME sorting?
  }
  
  /**
   * @param modelObject the model object to remove from the transform
   */
  public void removeTransformTarget(ModelObject modelObject) {
	  m_transformTargets.remove(modelObject);
	  modelObject.setParentObject(null);
  }

  
  /**
   * Creates a {@link ModelTransform} from the given XML element
   * Not that this is NOT recursive into child elements
   * @param element the XML element representing a {@link ModelTransform}
   * @return the {@link ModelTransform} encoded by the XML
   * @throws InvalidModelException if the XML is ill-formed
   */
  public static ModelTransform createFromXML(Element element) throws InvalidModelException {
    if (!element.hasAttribute("x") || !element.hasAttribute("y") ||
        !element.hasAttribute("z"))
      throw new InvalidModelException("Missing required transform attribute");
    
    try {
      double x = Double.parseDouble(element.getAttribute("x"));
      double y = Double.parseDouble(element.getAttribute("y"));
      double z = Double.parseDouble(element.getAttribute("z"));
      
      return new ModelTransform(x, y, z);
    }
    catch (NumberFormatException e) {
      throw new InvalidModelException(e);
    }
  }

  /**
   * @return a copy of the list of the transform targets 
   */
  public List<ModelObject> getTransformTargets() {
    //FIXME is this a list or a set??
    return new ArrayList<ModelObject>(m_transformTargets);
  }
  
  /**
   * @return the x-direction shift of the transform
   */
  public double getXShift() {
	  return m_xShift;
  }
  
  /**
   * @return the y-direction shift of the transform
   */
  public double getYShift() {
	  return m_yShift;
  }
  
  /**
   * @return the z-direction shift of the transform
   */
  public double getZShift() {
	  return m_zShift;
  }
  
  public void setXShift(double x) {
	  m_xShift = x;
  }
  
  public void setYShift(double y) {
	  m_yShift = y;
  }
  
  public void setZShift(double z) {
	  m_zShift = z;
  }
  
  @Override
  public Element toXMLElement(Document doc) {
    // note that this IS recursive
    Element transform = doc.createElement("transform");
    transform.setAttribute("x", Double.toString(m_xShift));
    transform.setAttribute("y", Double.toString(m_yShift));
    transform.setAttribute("z", Double.toString(m_zShift));
    
    // now recurse:
    for (ModelObject modelObject : m_transformTargets) {
      transform.appendChild(modelObject.toXMLElement(doc));
    }
    
    return transform;
  }
}
