/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.model;

/**
 * A class representing a unit of measurement. A unit has a {@link UnitNames} and
 * a prefix representing a quantification modification.
 * @author kmcguire
 * @author aadler
 *
 */
public class Units{
	
	private UnitNames unitName;
	private UnitPrefixes unitPrefix;
	

	private static final Units METERS = new Units(UnitPrefixes.NONE, UnitNames.METERS);
	private static final Units GRAMS = new Units(UnitPrefixes.NONE, UnitNames.GRAMS);
	private static final Units KILOGRAMS = new Units(UnitPrefixes.KILO, UnitNames.GRAMS);
	private static final Units RADIANS = new Units(UnitPrefixes.NONE, UnitNames.RADIANS);
	//private static final Units NONE = new Units("", "");
	
	
	/**
	 * Units stores a UnitPrefix(ex. UnitPrefixes.CENTI) and a UnitName (ex. UnitNames.METERS)
	 * 
	 * @param prefix
	 * @param name
	 * @author kmcguire
	 */
	public Units(UnitPrefixes prefix, UnitNames name) {
		unitPrefix = prefix;
		unitName = name;
	}
	
	/**
	 * @return a copy of the {@link Units}
	 */
	public Units copy() {
	  return new Units(unitPrefix, unitName);
	}
	
	/**
	 * Convert a {@link String} to a {@link Units}
	 * @param name the name to get the {@link Units} for
	 * @return the {@link Units} that matches the specified {@link String}
	 */
	public static Units getUnit(String name) {
	  if (name.equals(METERS.getBaseUnit()))
	    return getMeters();
	  if (name.equals(GRAMS.getfullName()))
		  return getGrams();
	  if (name.equals(KILOGRAMS.getBaseUnit()))
		  return getKilograms();
	  if (name.equals(RADIANS.getBaseUnit()))
		  return getRadians();
	  return getNone();
	}
	
	/**
	 * @return {@link Units} for meters
	 */
	public static Units getMeters() {
	  return new Units(UnitPrefixes.NONE, UnitNames.METERS);
	}
	
  /**
   * @return {@link Units} for grams
   */
	public static Units getGrams() {
	  return new Units(UnitPrefixes.NONE, UnitNames.GRAMS);
	}
	
  /**
   * @return {@link Units} for kilograms
   */
	public static Units getKilograms() {
	  return new Units(UnitPrefixes.KILO, UnitNames.GRAMS);
	}
	
  /**
   * @return {@link Units} for radians
   */
	public static Units getRadians() {
	  return new Units(UnitPrefixes.NONE, UnitNames.RADIANS);
	}
	
  /**
   * @return {@link Units} representing the lack of a unit
   */
	public static Units getNone() {
	  return new Units(UnitPrefixes.NONE, UnitNames.NONE);
	}
	
	/**
	 * @return returns standard MKS units. Iff the unit is grams, returns the prefix kilo. 
	 * Otherwise, the {@link UnitPrefixes.NONE} is used.
	 */
	public Units getStandardizedUnits() {
	  UnitPrefixes prefix = UnitPrefixes.NONE;
	  if (getBaseUnit() == UnitNames.GRAMS) {
	    prefix = UnitPrefixes.KILO;
	  }

		return new Units(prefix, getBaseUnit());
	}
	
	/**
	 * @return returns the {@link UnitNames} of the measurement without the prefix.
	 */
	public UnitNames getBaseUnit(){
		return unitName;
	}

	/**
	 * @return the full name of the {@link Units} as a {@link String}
	 */
	public String getfullName() {
		String fullName = unitPrefix.toString() + unitName.toString();
		return fullName;
	}
	
	/**
	 * @return the {@link UnitPrefixes} for the {@link Units}
	 */
	public UnitPrefixes getPrefix() {
		return unitPrefix;
	}

	/**
	 * @return returns the factor by which a prefixed value can be converted back to the standard value through multiplication.
	 */
	public double getConvert(){
		return unitPrefix.getConvert();
	}
	
	@Override
	public String toString() {
	  return getfullName();
	}
}