/* Panel for displaying the robot and environment
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.ui;

import java.awt.BorderLayout;
import java.awt.Component;

import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import java.awt.Dimension;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;


import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLProfile;
import javax.media.opengl.awt.GLCanvas;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JPanel;

import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.model.DesiredPhysicalModel;
import com.bbn.morph.model.PhysicalModel;
import com.jogamp.opengl.util.AnimatorBase;
import com.jogamp.opengl.util.FPSAnimator;

public class BotPanel extends JPanel implements ItemListener {
  private static final long serialVersionUID = 1L;
  private AnimatorBase m_animator;
  private GLCanvas m_canvas;
  private JCheckBox envButton; // Displays environment or not
  private JCheckBox desButton; // Displays desired robot/env or not
  private UIRobotDraw m_gfx;
  private MouseTracker m_mouseTracker;


  public BotPanel(CompositeModel composite) {
    PhysicalModel model = composite.getRobotModel();
    DesiredPhysicalModel desired = composite.getDesiredRobotModel();
    PhysicalModel envModel = composite.getEnvironmentModel();
    DesiredPhysicalModel envDesired = composite.getDesiredEnvironmentModel();
    setLayout(new BorderLayout());
    m_canvas = new GLCanvas(createGLCapabilites());
    add(m_canvas, BorderLayout.CENTER);

    m_mouseTracker = new MouseTracker();
    m_gfx = new UIRobotDraw(m_mouseTracker, model, envModel, desired, envDesired);
    m_canvas.addGLEventListener(m_gfx);

    m_animator = new FPSAnimator(80);
    m_animator.add(m_canvas);
    
    JPanel checkBoxPanel = new JPanel();
    checkBoxPanel.setLayout(new BoxLayout(checkBoxPanel, BoxLayout.Y_AXIS));
    add(checkBoxPanel, BorderLayout.SOUTH);
    
    envButton = new JCheckBox("Show environment");
    envButton.addItemListener(this);
    envButton.setAlignmentX(Component.LEFT_ALIGNMENT);
    checkBoxPanel.add(envButton);
    
    desButton = new JCheckBox("Show desired robot and environment");
    desButton.addItemListener(this);
    desButton.setAlignmentX(Component.LEFT_ALIGNMENT);
    checkBoxPanel.add(desButton);
    

    
    // mouse events that we care about:
    m_canvas.addMouseListener(new MouseAdapter() {
      @Override
      public void mouseDragged(MouseEvent me) {
        m_mouseTracker.drag(me);
      }
      
      @Override
      public void mousePressed(MouseEvent me) {
        m_mouseTracker.buttonDown(me);
      }
      
      @Override
      public void mouseReleased(MouseEvent me) {
        m_mouseTracker.buttonUp(me);
      }
    });
    
    // keyboard events that we care about:
    m_canvas.addKeyListener(new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        switch (e.getKeyCode()) {
        case KeyEvent.VK_UP:
          m_mouseTracker.onJoyMove(0, 2);
          break;
        case KeyEvent.VK_DOWN:
          m_mouseTracker.onJoyMove(0, -2);
          break;
        case KeyEvent.VK_LEFT:
          m_mouseTracker.onJoyMove(2, 0);
          break;
        case KeyEvent.VK_RIGHT:
          m_mouseTracker.onJoyMove(-2, 0);
          break;
        case KeyEvent.VK_PAGE_UP:
          m_mouseTracker.onKeyZoom(5);
          break;
        case KeyEvent.VK_PAGE_DOWN:
          m_mouseTracker.onKeyZoom(-5);
          break;
        case KeyEvent.VK_Z:
          m_mouseTracker.reset();
          break;
        default:
          super.keyPressed(e);
        }
      }
    });
  }
  
  public void itemStateChanged(ItemEvent e) {
    Object source = e.getItemSelectable();

    if(source == envButton) {
      if(e.getStateChange() == ItemEvent.SELECTED) {
        //m_canvas.addGLEventListener(gfx2);
        m_gfx.switchIsEnvironmentOn(true);
      }

      if(e.getStateChange() == ItemEvent.DESELECTED) {
        //m_canvas.removeGLEventListener(gfx2);
        m_gfx.switchIsEnvironmentOn(false);
      }
    }
	  if(source == desButton) {
		  if(e.getStateChange() == ItemEvent.SELECTED) {
			  m_gfx.switchIsDesiredOn(true);
		  }
		  
		  if(e.getStateChange() == ItemEvent.DESELECTED) {
			  m_gfx.switchIsDesiredOn(false);
		  }
	  }
  }

  @Override
  public Dimension getMinimumSize() {
    // This is a fix that gets around a bug in JOGL that won't allow
    // the GLCanvas to be shrunk.
    // See: http://download.java.net/media/jogl/builds/archive/jsr-231-1.1.0/jogl-1.1.0-windows-i586/Userguide.html
    return new Dimension(0, 0);
  }
  
  private GLCapabilities createGLCapabilites() {
    GLProfile glp = GLProfile.getDefault();
    GLCapabilities capabilities = new GLCapabilities(glp);
    capabilities.setHardwareAccelerated(true);

    // try to enable 2x anti aliasing - should be supported on most hardware
    capabilities.setNumSamples(2);
    capabilities.setSampleBuffers(true);

    return capabilities;
  }

  public void start(){
    m_animator.start();
  }

  public void stop() {
    m_animator.stop();
    System.exit(0);
  }
}
