/* Interface for displaying/editing model parameters
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.ui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.ArrayList;
import java.util.EventObject;
import java.util.HashMap;
import java.util.List;
import javax.swing.AbstractCellEditor;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.SwingWorker;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellEditor;

import com.bbn.morph.controller.Controller;
import com.bbn.morph.controller.DiscreteUpdateController;
import com.bbn.morph.model.CompositeModel;
import com.bbn.morph.model.DesiredPhysicalModel;
import com.bbn.morph.model.Measurement;
import com.bbn.morph.model.ModelListener;
import com.bbn.morph.model.PhysicalModel;
import com.bbn.morph.model.RobotAttributeValues;
import com.bbn.morph.model.Units;
import com.bbn.morph.view.AttributeValueView;
import com.bbn.morph.view.RobotAttribute;

/**
 * Display parameters that the user can update in a {@link JPanel} extension.
 * @author aadler
 *
 */
public class ParameterPanel extends JPanel implements ActionListener {

  private static final long serialVersionUID = 1L;
  private PhysicalModel m_robotModel;
  private DesiredPhysicalModel m_desiredRobotModel;
  private PhysicalModel m_environmentModel;
  private DesiredPhysicalModel m_desiredEnvironmentModel;
  private Controller m_controller;
  private DiscreteUpdateController m_discreteController;
  private ProtoSimulator m_protoSimulator;
  private AboutDialog m_dialog = null;
  
  private final static int MAX_ITERATIONS = 100;
  private JButton m_iterateButton;
  private JButton m_equilibriumButton;
  private JButton m_simulationButton;
  private JButton m_discreteOneButton;
  private JButton m_discreteAllButton;
  
  public ParameterPanel(CompositeModel compositeModel, Controller controller, DiscreteUpdateController discreteController) {
    m_robotModel = compositeModel.getRobotModel();
    m_desiredRobotModel = compositeModel.getDesiredRobotModel();
    m_environmentModel = compositeModel.getEnvironmentModel();
    m_desiredEnvironmentModel = compositeModel.getDesiredEnvironmentModel();
    m_controller = controller;
    m_protoSimulator = new ProtoSimulator(compositeModel);
    m_discreteController = discreteController;
    
    setLayout(new BoxLayout(this, BoxLayout.PAGE_AXIS));
    add(new ModelParameterPanel(m_robotModel, m_desiredRobotModel, m_controller, "Robot Model"));
    add(new ModelParameterPanel(m_environmentModel, m_desiredEnvironmentModel, m_controller, "Environment Model"));
    
    m_iterateButton = new JButton("Run Iteration");
    m_iterateButton.setActionCommand("iterate");
    m_iterateButton.addActionListener(this);
    add(m_iterateButton);

    m_equilibriumButton = new JButton("Run to Equilibrium");
    m_equilibriumButton.setActionCommand("equilibrium");
    m_equilibriumButton.addActionListener(this);
    add(m_equilibriumButton);
    
    m_discreteOneButton = new JButton("Choose one component");
    m_discreteOneButton.setActionCommand("discreteChooseOne");
    m_discreteOneButton.addActionListener(this);
    add(m_discreteOneButton);
    
    m_discreteAllButton = new JButton("Choose all components");
    m_discreteAllButton.setActionCommand("discreteChooseAll");
    m_discreteAllButton.addActionListener(this);
    add(m_discreteAllButton);
//TODO: Commented out for switching to ROSS. Fix it to display ROSS simulation     
//    m_simulationButton = new JButton("Show Simulation");
//    m_simulationButton.setActionCommand("simulation");
//    m_simulationButton.addActionListener(this);
//    add(m_simulationButton);
  }
  
  private void setButtonState(boolean state) {
    m_iterateButton.setEnabled(state);
    m_equilibriumButton.setEnabled(state);
    //m_simulationButton.setEnabled(state);
    m_discreteOneButton.setEnabled(state);
    m_discreteAllButton.setEnabled(state);
  }
  
  private class OneIteration extends SwingWorker<Void, Void> {
    @Override
    protected Void doInBackground() throws Exception {
      try {
        m_controller.runIteration();
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      
      return null;
    }
    
    @Override
    protected void done() {
      setButtonState(true);
    }
  }
  
  private class IterationLoop extends SwingWorker<Void, Void> {
    @Override
    protected Void doInBackground() throws Exception {
      try {
        m_controller.runToEquilibrium(MAX_ITERATIONS);
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      return null;
    }
    
    @Override
    protected void done() {
      setButtonState(true);
    }
  }
  
  private class DiscreteIterationLoop extends SwingWorker<Void, Void> {
    @Override
    protected Void doInBackground() throws Exception {
      try {
        m_discreteController.chooseAllComponents(MAX_ITERATIONS);
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      return null;
    }
    
    @Override
    protected void done() {
      setButtonState(true);
    }
  }

  private class DiscreteIteration extends SwingWorker<Void, Void> {
    @Override
    protected Void doInBackground() throws Exception {
      try {
        m_discreteController.chooseOneComponent();
      }
      catch (Exception e) {
        e.printStackTrace();
      }
      return null;
    }
    
    @Override
    protected void done() {
      setButtonState(true);
    }
  }

  private class Simulation extends SwingWorker<Void, Void> {
    @Override
    protected Void doInBackground() throws Exception {
      m_protoSimulator.attemptProtoCall();
      return null;
    }
    
    @Override
    protected void done() {
      setButtonState(true);
    }
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    if (e.getActionCommand().equals("iterate")) {
      setButtonState(false);
      OneIteration process = new OneIteration();
      try {
        process.execute();
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
    else if (e.getActionCommand().equals("equilibrium")) {
      setButtonState(false);
      IterationLoop process = new IterationLoop();
      try {
        process.execute();
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
    else if (e.getActionCommand().equals("discreteChooseAll")) {
      setButtonState(false);
      DiscreteIterationLoop process = new DiscreteIterationLoop();
      try {
        process.execute();
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
    else if (e.getActionCommand().equals("discreteChooseOne")) {
      setButtonState(false);
      DiscreteIteration process = new DiscreteIteration();
      try {
        process.execute();
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
    else if (e.getActionCommand().equals("simulation")) {
      setButtonState(false);
      Simulation process = new Simulation();
      try {
        process.execute();
      }
      catch (Exception ex) {
        ex.printStackTrace();
      }
    }
    else
      System.err.println("Unknown command: "+e.getActionCommand());
  }
  
  private class ModelParameterPanel extends JPanel implements ModelListener, WindowListener {
    
    private static final long serialVersionUID = 1L;
    private PhysicalModel m_model;
    private DesiredPhysicalModel m_desiredModel;
    
    private HashMap<RobotAttribute, ActualDesired> m_parameterMap;
    private List<ActualDesired> m_parameterList;
    private JTable m_table;
    private ParameterTableModel m_parameterTableModel;
    private Controller m_controller;

    public ModelParameterPanel(PhysicalModel model, DesiredPhysicalModel desiredModel, Controller controller, String title) {
      // TODO we probably want to wrap the two models and the controller in a view
      m_model = model;
      m_desiredModel = desiredModel;
      m_model.addModelListener(this);
      m_controller = controller;
      
      m_parameterMap = new HashMap<RobotAttribute, ActualDesired>();
      m_parameterList = new ArrayList<ActualDesired>();
      
      m_parameterTableModel = new ParameterTableModel();
      m_table = new JTable(m_parameterTableModel);
      m_table.setPreferredScrollableViewportSize(new Dimension(200, 200));
      m_table.setDefaultEditor(Measurement.class, new MeasurementEditor(this));
      m_table.setRowHeight(20);
      JScrollPane scrollPane = new JScrollPane(m_table);
      
      setLayout(new BorderLayout());
      add(scrollPane, BorderLayout.CENTER);
      add(new JLabel(title), BorderLayout.NORTH);
      
      refresh();
    }
    
    public void refresh() {
      // refresh the model info
      AttributeValueView attributeValueView = m_model.getAttributeValueView();
      for (RobotAttribute ra : attributeValueView.keySet()) {
        Measurement newValue = attributeValueView.get(ra);
        if (m_parameterMap.containsKey(ra)) {
          ActualDesired actualDesired = m_parameterMap.get(ra);
          actualDesired.setActual(newValue);
          // clear the old desired value:
          actualDesired.clearDesired();
        }
        else {
          ActualDesired ad = new ActualDesired(ra, newValue);
          m_parameterMap.put(ra, ad);
          m_parameterList.add(ad);
        }
      }
      
      // refresh the desired info
      for (RobotAttributeValues rav : m_desiredModel.getDesiredChanges()) {
        RobotAttribute ra = new RobotAttribute(rav.getAttribute());
        Measurement value = rav.getMeasurement();
        m_parameterMap.get(ra).setDesired(value);
      }
      
      m_parameterTableModel.fireTableDataChanged();
    }
    
    public void removeDesired(int rowIndex) {
      ActualDesired actualDesired = m_parameterList.get(rowIndex);
      m_controller.removeDesiredModel(m_desiredModel, actualDesired.getAttribute());
      refresh();
    }

    @Override
    public void modelChanged(EventObject event) {
      if (event.getSource() == m_model) {
        refresh();
      }
      else 
        System.err.println("unknown event source");
    }

    /**
     * @param rowIndex
     * @return return the unit for the attribute in the specified row.
     */
    private Units getUnit(int rowIndex){
      return m_parameterList.get(rowIndex).getActual().getUnits();
    } 
    
    private class ParameterTableModel extends AbstractTableModel {
      private static final long serialVersionUID = 1L;
      private String[] m_columnNames = {"Parameter", "Value", "Units", "Desired Value"};
      
      @Override
      public int getColumnCount() {
        return m_columnNames.length;
      }
      
      @Override
      public int getRowCount() {
        return m_parameterList.size();
      }
      
      @Override
      public Object getValueAt(int rowIndex, int columnIndex) {
        ActualDesired actualDesired = m_parameterList.get(rowIndex);
        switch (columnIndex) {
        case 0:
          return actualDesired.getAttribute().getName();
        case 1:
          return actualDesired.getActual().getNeaterMeasurement().getPrefixValue();
        case 2:
          return actualDesired.getActual().getNeaterMeasurement().getUnits().getfullName();
        case 3:
          Measurement desired = actualDesired.getDesired();
          if(desired!=null) {
            desired = desired.getNeaterMeasurement();
            return (desired.getPrefixValue() + " " + desired.getUnits().getfullName());
          }
        }
        return null;
      }
      
      @Override
      public String getColumnName(int column) {
        return m_columnNames[column];
      }
      
      @Override
      public Class<?> getColumnClass(int columnIndex) {
        switch(columnIndex) {
        case 0:
          return String.class;
        case 1:
          return Double.class;
        case 2:
          return String.class;
        case 3:
          return Measurement.class;
        }
        return super.getColumnClass(columnIndex);
      }

      @Override
      public boolean isCellEditable(int rowIndex, int columnIndex) {
        if (columnIndex < 3)
          return false;
        else
          return true;
      }
      
      @Override
      public void setValueAt(Object aValue, int rowIndex, int columnIndex) {
        if (columnIndex == 3)
          if (aValue instanceof String) {
            String aStr = (String)aValue;
            if (!aStr.isEmpty()) {
              ActualDesired actualDesired = m_parameterList.get(rowIndex);
              RobotAttribute attribute = actualDesired.getAttribute();
              Measurement newValue = Measurement.getMeasurement(aStr, getUnit(rowIndex));
              m_controller.changeDesiredModel(m_desiredModel, m_model, attribute, newValue);
              refresh();
            }
          }
      }
    }

	@Override
	public void windowActivated(WindowEvent e) {
	}

	@Override
	public void windowClosed(WindowEvent e) {
		m_dialog = null;
	}

	@Override
	public void windowClosing(WindowEvent e) {
	}

	@Override
	public void windowDeactivated(WindowEvent e) {
	}

	@Override
	public void windowDeiconified(WindowEvent e) {
	}

	@Override
	public void windowIconified(WindowEvent e) {
	}

	@Override
	public void windowOpened(WindowEvent e) {
	}
  }
  
  private class MeasurementEditor extends AbstractCellEditor implements TableCellEditor{

	  /**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	// This is the component that will handle the editing of the cell value
	private JTextField m_textField;
	private JButton m_deleteButton;
	private JComponent m_component;
  private ModelParameterPanel m_mpp;
  private int m_rowIndex = -1;
	  
	  public MeasurementEditor(ModelParameterPanel mpp) {
	    super();
	    
	    m_mpp = mpp;
	    
	    m_textField = new JTextField();
	    m_deleteButton = new JButton(new ImageIcon("resources/delete.png"));
	    m_deleteButton.setPreferredSize(new Dimension(15, 16));
	    m_deleteButton.setBorderPainted(false);
	    m_deleteButton.addActionListener(new ActionListener() {
        
        @Override
        public void actionPerformed(ActionEvent e) {
          m_textField.setText("");
          stopCellEditing();
          m_mpp.removeDesired(m_rowIndex);
        }
      });
	    m_component = new JPanel(new BorderLayout());
	    m_component.setBackground(Color.WHITE);

      m_component.add(m_textField, BorderLayout.CENTER);
      m_component.add(m_deleteButton, BorderLayout.LINE_END);
    }

	  // This method is called when a cell value is edited by the user.
	  public Component getTableCellEditorComponent(JTable table, Object value,
			  boolean isSelected, int rowIndex, int vColIndex) {
		  // 'value' is value contained in the cell located at (rowIndex, vColIndex)

		  if (isSelected) {
			  // cell (and perhaps other cells) are selected
		  }

		  // Configure the component with the specified value
		  m_textField.setText((String)value);
		  m_rowIndex = rowIndex;

		  // Return the configured component
		  return m_component;
	  }

	  // This method is called when editing is completed.
	  // It must return the new value to be stored in the cell.
	  public Object getCellEditorValue() {
		  return m_textField.getText();
	  }
	  
	  private void popUpDialog() {
		  if(m_dialog == null){
			  m_dialog = new AboutDialog(new JFrame(), "Error: Unit Mismatch", "Units do not match expected units of type " + m_mpp.getUnit(m_rowIndex).getBaseUnit() + ".");
			  m_dialog.setLocationRelativeTo(null);
			  m_dialog.setAlwaysOnTop(true);
			  m_dialog.setVisible(true);
			  m_dialog.addWindowListener((java.awt.event.WindowListener) m_mpp);
		  }
	  }
	  
	@Override
	public boolean stopCellEditing() {
		String s = (String)getCellEditorValue();
		
		Units unit = m_mpp.getUnit(m_rowIndex);
		Measurement measurement = Measurement.getMeasurement(s, unit);
		
		if(measurement == null && !s.isEmpty()){
			popUpDialog();
			return false;
		}
		return super.stopCellEditing();
	
	  
  }
  }
  
  private class ActualDesired {
    
    private Measurement m_actual = null;
    private Measurement m_desired = null;
    private RobotAttribute m_ra;
    
    public ActualDesired(RobotAttribute ra, Measurement actual) {
      m_ra = ra;
      m_actual = actual;
    }
    
    public RobotAttribute getAttribute() {
      return m_ra;
    }
    
    public Measurement getActual() {
      return m_actual; 
    }
    
    public void setActual(Measurement newValue) {
      m_actual = newValue;
    }
    
    /**
     * @return returns the desired measurement or null if there is no desired measurement.
     */
    public Measurement getDesired() {
      return m_desired;
    }
    
    public void clearDesired() {
      m_desired = null;
    }
    
    public void setDesired(Measurement value) {
      m_desired = value;
    }

  }
  
  

}
