/* 
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.view;

import com.bbn.morph.executer.Pair;
import com.bbn.morph.executer.StressInputType;
import com.bbn.morph.executer.StressInputs;


/**Implements a symmetric stress function with five linear regions identified by
 *  one  point x0 and two radiuses r0 & r1 . 
 * Ideal for representing the stress on systems that have a target value in which case 
 * the target should be x0.
 * 
 * 
 * 
 * -----(x0-r1,1)			 (x0+r1,1)-------
 *        \                      /   
 *         \					/
 *         (x0-r0,0)--------(x0+r0 ,0)
 *   
 *         
 * 
 * Stress value in [0,1] and MaxDouble if in non-viable space
 * Stress Direction is the slope (- for left side and + for right side)
 * 
 * @author fyaman
 *
 */

public class SymmetricFivePieceLinearStressFunction implements StressFunction{
    StressInputType inputLabel;
    double center = 0.0; // for now the center is assumed to be zero
    double r0;
    double r1;
    StressFunction left;
    StressFunction right;
	
	public SymmetricFivePieceLinearStressFunction(StressInputType inputLabel, double target, double r0, double r1) {
		super();
		this.inputLabel = inputLabel;
		this.r0=r0;
		this.r1=r1;
		left = new ThreePieceLinearStressFunction(inputLabel, new Pair (target-r1,1), new Pair (target-r0,0) );
		right = new ThreePieceLinearStressFunction(inputLabel, new Pair (target+r0,0), new Pair (target+r1,1) );
	}

	public StressInputType getInputLabel() {
		return inputLabel;
	}

	@Override
	public Double computeStress(StressInputs inputs) {
		double in= inputs.get(inputLabel);
		// If close enough to the center return 0 stress
		if (Math.abs(center - in) <= r0 )
			return 0.0;
		if (in < center)
			return left.computeStress(inputs);
		return right.computeStress(inputs);
		
	}

	@Override
	public String description() {
		String desc="SymmetricFivePieceStressFunction(" + center+ ","+ r0+ "," + r1+")"; 
		return desc;
	}

	
}
