/*
Copyright (C) 2011, Raytheon BBN Technologies and contributors listed
in the AUTHORS file in the MADV distribution's top directory.

This file is part of MADV, and is distributed under the terms of the
GNU General Public License, with a linking exception, as described in
the file LICENSE in the MADV distribution's top directory. */
package com.bbn.morph.view.test;
import junit.framework.Assert;

import org.junit.Test;

import com.bbn.morph.executer.Feature;
import com.bbn.morph.executer.Pair;
import com.bbn.morph.executer.StressInputType;
import com.bbn.morph.executer.StressInputs;
import com.bbn.morph.view.SymmetricFivePieceLinearStressFunction;
import com.bbn.morph.view.ThreePieceLinearStressFunction;
import com.bbn.morph.view.StressFunction;

public class StressFunctionTest {

	public StressInputs genInput(String FBname, StressInputType inputLabel, double val){
		StressInputs inputs = new StressInputs(new Feature(FBname));
		inputs.put(inputLabel, val);
		return inputs;
	}
	
	@Test
	public void testIncreasingStressFunction(){
		Pair first =new Pair(3,1);
		Pair second =new Pair(5,3);
		StressInputType inputLabel= StressInputType.TIME;
		StressFunction stressF= new ThreePieceLinearStressFunction(inputLabel, first, second);
		System.out.println("Testing:"+ stressF.description());
		
		StressInputs inputs1 = genInput("testFB", inputLabel, 2);
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);
		StressInputs inputs2 = genInput("testFB", inputLabel, 7);
		Assert.assertEquals(stressF.computeStress(inputs2),Double.MAX_VALUE);	
		StressInputs inputs3 = genInput("testFB", inputLabel, 4);
		Assert.assertEquals(stressF.computeStress(inputs3),-0.5);
		
	
		
	};
	
	@Test
	public void testDecreasingStressFunction(){
		Pair second =new Pair(5,1);
		Pair first =new Pair(3,3);
		StressInputType inputLabel= StressInputType.TIME;
		StressFunction stressF= new ThreePieceLinearStressFunction(inputLabel, first, second);
		System.out.println("Testing:"+ stressF.description());
		
		StressInputs inputs1 = genInput("testFB", inputLabel, 2);
		Assert.assertEquals(stressF.computeStress(inputs1),Double.MAX_VALUE);
		StressInputs inputs2 = genInput("testFB", inputLabel, 7);
		Assert.assertEquals(stressF.computeStress(inputs2),0.0);	
		
		StressInputs inputs3 = genInput("testFB", inputLabel, 4);
		Assert.assertEquals(stressF.computeStress(inputs3),0.5);	
	};

	
	@Test
	public void testSymetricStressFunction(){
		StressInputType inputLabel= StressInputType.TIME;
		double center = 0;
		double r0 = 1;
		double r1 = 5;
		
		StressFunction stressF= new SymmetricFivePieceLinearStressFunction(inputLabel, center, r0, r1);
		System.out.println("Testing:"+ stressF.description());
		
		
		// Test center
		StressInputs inputs1 = genInput("testFB", inputLabel, center);
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);
		// Test a point within inner circle
		inputs1 = genInput("testFB", inputLabel, center+ (r0/2));
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);	
		// Test inner circle boundary
		inputs1 = genInput("testFB", inputLabel, center+ r0);
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);	
		
		// Test right side 
		// Test a point within outer circle
		inputs1 = genInput("testFB", inputLabel, center+ r0+ (r1- r0)/2);
		Assert.assertEquals(stressF.computeStress(inputs1),-0.5);	
		// Test outer circle boundary
		inputs1 = genInput("testFB", inputLabel, center+ r1);
		Assert.assertEquals(stressF.computeStress(inputs1),Double.MAX_VALUE);	
		
		
		// Test left side 
		// Test a point within outer circle
		inputs1 = genInput("testFB", inputLabel, center-r0- (r1- r0)/2);
		Assert.assertEquals(stressF.computeStress(inputs1),0.5);	
		// Test outer circle boundary
		inputs1 = genInput("testFB", inputLabel, center- r1);
		Assert.assertEquals(stressF.computeStress(inputs1),Double.MAX_VALUE);		
	};
	
	
	@Test
	public void testSymetricStressFunctionFor0InnerRadius(){
		StressInputType inputLabel= StressInputType.TIME;
		double center = 0;
		double r0 = 0;
		double r1 = 5;
		
		StressFunction stressF= new SymmetricFivePieceLinearStressFunction(inputLabel, center, r0, r1);
		System.out.println("Testing:"+ stressF.description());
		
		
		// Test center
		StressInputs inputs1 = genInput("testFB", inputLabel, center);
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);
		// Test a point within inner circle
		inputs1 = genInput("testFB", inputLabel, center+ (r0/2));
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);	
		// Test inner circle boundary
		inputs1 = genInput("testFB", inputLabel, center+ r0);
		Assert.assertEquals(stressF.computeStress(inputs1),0.0);	
		
		// Test right side 
		// Test a point within outer circle
		inputs1 = genInput("testFB", inputLabel, center+ r0+ (r1- r0)/2);
		Assert.assertEquals(stressF.computeStress(inputs1),-0.5);	
		// Test outer circle boundary
		inputs1 = genInput("testFB", inputLabel, center+ r1);
		Assert.assertEquals(stressF.computeStress(inputs1),Double.MAX_VALUE);	
		
		
		// Test left side 
		// Test a point within outer circle
		inputs1 = genInput("testFB", inputLabel, center-r0- (r1- r0)/2);
		Assert.assertEquals(stressF.computeStress(inputs1),0.5);	
		// Test outer circle boundary
		inputs1 = genInput("testFB", inputLabel, center- r1);
		Assert.assertEquals(stressF.computeStress(inputs1),Double.MAX_VALUE);		
	};
	
	
	
	public static  void main(String[] args) {
		
	}

	
}
