;; This is a topological representation of blueprints

;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;;;;;;;;;;;;;;;;;
;;;; BASE CLASSES: ELEMENTS AND LINKS
;;;;;;;;;;;;;;;;;;;

;; Element classes:
;; - topological cell (may have lower-dimension cells as boundaries)
;; - parameter (must attach to cell being measured)
;; - value (numerical realization of a parameter)
;; - relation (links parameters) [mathematical or FB]

;; Linkage classes:
;; - boundary: links K-cell to (K-1)-cell
;; - measurement: links parameter to K-cells
;; - assignment: links parameter to value
;; - derivation: links values together
;; - constraint: links relations



;; just a set of all elements, links
(defvar elements (make-hash-table))
(defvar links (make-hash-table))
;; these are caches for optimization
(defvar relation-list '()) (defvar parameter-list '())
(defun clear-topology () 
  (clrhash elements) (clrhash links) 
  (setf relation-list nil parameter-list nil)
  'cleared)
(defun all-elements () (hash-keys elements))
(defun all-links () (hash-keys links))
(defun all-parameters () parameter-list)
(defun all-relations () relation-list)

(defclass element () 
  ((links :initform (make-hash-table :values nil))
   (marks :initform (make-hash-table))
   (name :initform nil :initarg :name :accessor name)))
(defmethod initialize-instance :after ((e element) &rest args)
  (puthash-key e elements)
  (when (typep e 'relation) (push e relation-list))
  (when (typep e 'parameter) (push e parameter-list)))
(defmethod delete-element ((e element))
  (assert (with-slots (links) e (= (hash-table-count links) 0))
      (e) "Cannot delete element ~A while it still has links" e)
  (remhash e elements)
  (when (typep e 'relation) (setf relation-list (remove e relation-list)))
  (when (typep e 'parameter) (setf parameter-list (remove e parameter-list))))

(defmethod name-it ((name string) (e element))
  (setf (name e) name))
;; Note: right now marking is binary; we may want to allow n-ary marks too
(defmethod mark ((e element) marker &optional (value t))
  (with-slots (marks) e (setf (gethash marker marks) value)))
(defmethod unmark ((e element) marker)
  (with-slots (marks) e (remhash marker marks)))
(defmethod mark? ((e element) marker)
  (with-slots (marks) e (gethash marker marks)))
(defun clear-mark-class (class)
  (dolist (e (all-elements)) (unmark e class)))

(defclass link () ())
(defgeneric head ((l link))) (defgeneric tail ((l link)))
(defmethod initialize-instance :after ((l link) &rest args)
  (puthash-key l links)
  (with-slots (links) (head l) (puthash-key l links))
  (with-slots (links) (tail l) (puthash-key l links)))
(defmethod switch-link ((l link) old-e new-e)
  (when (eq (head l) old-e) (setf (head l) new-e))
  (when (eq (tail l) old-e) (setf (tail l) old-e))
  (with-slots (links) old-e (remhash l links))
  (with-slots (links) new-e (puthash-key l links)))
(defmethod delete-link ((l link))
  (remhash l links)
  (with-slots (links) (head l) (remhash l links))
  (with-slots (links) (tail l) (remhash l links)))

(defun links-of-type (elt type elt-is-head)
  (let ((key (if elt-is-head #'head #'tail)))
    (with-slots (links) elt
      (let ((bl (remove-if-not (ppf typep ?x type) (hash-keys links))))
	(remove-if-not (pcf eq elt)  bl :key key)))))
(defmethod links (elt)
  (with-slots (links) elt (hash-keys links)))

;;;;;;;;;;;;;;;;;;;
;;;; TOPOLOGY
;;;;;;;;;;;;;;;;;;;

;; We will represent the topology of a design as a CW-complex.
;; A CW-complex is a topological collection with the following
;; properties:
;; 1) It contains one or more cells - closed polytopes of dimension k
;; 2) All boundaries of all cells are contained within the collection

(defclass cell (element) 
  ((dimension :initarg :dim :accessor dimension)))
(defun make-cell (dim &rest boundaries)
  (apply #'make-named-cell nil dim boundaries))
(defun make-named-cell (name dim &rest boundaries)
  (assert (>= dim 0))
  (let ((c (make-instance 'cell :dim dim :name name)))
    (dolist (b boundaries) (make-boundary c b))
    c))
(defgeneric find-cells (pattern &optional extrema))
(defmethod print-object ((c cell) stream)
  (format stream "<~A-Cell~@[ ~A~] ~A>"
	  (dimension c) (name c) (mapcar (cmp name tail) (coordinates c))))

(defclass boundary (link)
  ((bounded :initarg :bounded :accessor tail) ; cell of dimension k
   (bound :initarg :bound :accessor head))) ; cell of dimension k-1
(defmethod make-boundary ((bounded cell) (bound cell))
  (assert (= (dimension bounded) (1+ (dimension bound))))
  (make-instance 'boundary :bounded bounded :bound bound))
(defmethod print-object ((b boundary) stream)
  (with-slots (bounded bound) b
    (format stream "<Boundary of ~A includes ~A>" bounded bound)))

(defmethod boundaries ((c cell)) 
  (mapcar #'head (links-of-type c 'boundary nil)))
(defmethod ensure-boundary ((bounded cell) (bound cell))
  (or (find bound (boundaries bounded))
      (make-boundary bounded bound)))

;; closure of boundaries (i.e. all boundaries, boundaries of boundaries, etc...
(defmethod boundaries* ((c cell))
  (let ((fullset '()) (lastset (list c)))
    (dotimes (i (dimension c))
      (let ((nextset (remove-duplicates (mapcan #'boundaries lastset))))
	(setf lastset nextset)
	(setf fullset (append fullset nextset))))
    fullset))

;; Note that this may generalize in non-intuitive ways for non-rectilinear spaces
(defclass derived-parallel (link)
  ((parent :initarg :parent :accessor tail)
   (child :initarg :child :accessor head)))
(defmethod make-parallel ((parent cell) (child cell))
  (assert (= (dimension parent) (dimension child)))
  (make-instance 'derived-parallel :parent parent :child child))
(defmethod ensure-parallel (parent child)
  (or (find parent (links-of-type child 'derived-parallel t) :key #'tail)
      (make-parallel parent child)))
(defmethod print-object ((p derived-parallel) stream)
  (with-slots (parent child) p
    (format stream "<~A is constructed parallel to ~A>" child parent)))

(defclass subspace (derived-parallel)
  ((parent :initarg :parent :accessor tail)
   (child :initarg :child :accessor head)))
(defmethod make-subspace ((parent cell) (child cell))
  (assert (= (dimension parent) (dimension child)))
  (make-instance 'subspace :parent parent :child child))
(defmethod print-object ((s subspace) stream)
  (with-slots (parent child) s
    (format stream "<~A is subspace of ~A>" child parent)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; PARAMETERS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;; NOTE: denominator not yet handled
(defclass parameter (element)
  ;; units: elementary = symbol, composite = (numerator-list . denominator-list)
  ((units :initarg :units :accessor units)
   (deltas :initform nil :accessor deltas)
   ;; this slot is true if there is precisely one relation attached
   (singleton :initform nil :accessor singleton)
   ;; this slot holds the parameter value if there is precisely one
   (single-value :initform nil :accessor single-value)))
(defmethod print-object ((p parameter) stream)
  (with-slots (name) p (format stream "<Parameter ~A>" name)))
(defun make-parameter (name units &rest measurements)
  (assert (not (find-parameter name)) (name) "Parameter ~A already exists" name)
  (let ((p (make-instance 'parameter :units units :name name)))
    (dolist (m measurements) (ensure-measure p m))
    p))
(defun find-parameter (name) 
  (if (typep name 'parameter)
      name
    (find-if (lambda (e) (and (typep e 'parameter) (eqv name (name e)))) 
	     (all-elements))))
(defun ensure-parameter (name units)
  (let ((p (find-parameter name)))
    (assert (or (not p) (equal units (units p))))
    (or p (make-parameter name units))))
      
(defmethod composite? ((p parameter))
  (with-slots (units) p (consp units)))

;; A dimension is a parameter of geometric magnitude (e.g., length,
;; area, volume).  This needs to be distinguished from generic
;; parameters so that it can be handled differently by development
;; operators: dimensions are often modified, while plain parameters
;; are not affected.
(defclass dimension (parameter) ())
(defun ensure-dimension (name units)
  (let ((d (find-parameter name)))
    (assert (or (not d) (and (typep d 'dimension) (equal units (units d)))))
    (or d (make-instance 'dimension :units units :name name))))

(defmethod make-parallel ((parent parameter) (child parameter))
  (make-instance 'derived-parallel :parent parent :child child))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; COORDINATE SYSTEMS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; A coordinate system is a lattice of measurements on cells,
;; generally associated with some set of parameters as well
(defclass coordinate-system (element)
  ((lattice-extrema)			; pair: (minimum, maximum)
   (spans :initform (make-hash-table :test #'equal)))) ; hash of (- . +) -> span


;;;; Components of coordinate lattices:
;; Note that since every lattice-element belongs to precisely one
;; coordinate system, we are specializing away from the usual "link"
;; approach to associating them, in order to reduce the number of links
(defclass lattice-element (element)	; placeholder join for coord & span
  ((coordinate-system :initarg :cs :accessor lattice)))
(defclass lattice-coordinate (lattice-element)
  ((precedes :initform nil)		; list of spans, nil = maximum
   (succeeds :initform nil)))		; list of spans, nil = minimum
(defmethod print-object ((lc lattice-coordinate) stream)
  (format stream "<lattice-coordinate ~A>" (name lc)))
(defclass lattice-span (lattice-element)
  ((lower :initarg :lower :accessor lower) 
   (upper :initarg :upper :accessor upper)))
(defmethod print-object ((ls lattice-span) stream)
  (format stream "<lattice-span ~A to ~A>" (name (lower ls)) (name (upper ls))))

;;;; Construction and refinement of coordinate systems
;; Generate a new unrefined coordinate system, consisting only of a
;; minimum, a maximum, and the span between them
(defmethod make-coordinate-system ((name symbol))
  (let ((cs (make-instance 'coordinate-system :name name)))
    (let* ((n-min (format nil "~A-" (name cs)))
	   (min (make-instance 'lattice-coordinate :cs cs :name n-min))
	   (n-max (format nil "~A+" (name cs)))
	   (max (make-instance 'lattice-coordinate :cs cs :name n-max)))
      (setf (slot-value cs 'lattice-extrema) (cons min max))
      (get-span min max)
      cs)))
;; When making from a parameter, the parameter measures the base-span
(defmethod make-coordinate-system ((p parameter))
  (let ((cs (make-coordinate-system (name p))))
    (ensure-measure p (base-span cs))
    cs))

;; Refinements are named parent.#/alpha, where # means it is the
;; nth successor of the min and alpha counts up for num-subs
(defmethod refine-span ((ls lattice-span) num-subs)
  (assert (and (integerp num-subs) (plusp num-subs)))
  (if (= num-subs 1)					 
      (list ls)				  ; one-subspan = no-op
    (let ((intermediates '())
	  (names (refinement-names (lower ls) num-subs))
	  (cs (lattice ls)))
      (dotimes (i (1- num-subs)) 
	(push (make-instance 'lattice-coordinate :cs cs :name (pop names)) 
	      intermediates))
      (with-slots (lower upper) ls
	(mapcar (ppf make-instance 'lattice-span :cs cs :lower ?x :upper ?y
		     :name (span-name ?x ?y))
		(cons lower intermediates)
		(append intermediates (list upper)))))))

;;;; Accessors for coordinate system elements
(defmethod get-span ((lower lattice-coordinate) (upper lattice-coordinate))
  (assert (eq (lattice lower) (lattice upper)))
  (let ((cs (lattice lower)))
    (with-slots (spans) cs
      (ensure (gethash (cons lower upper) spans) 
	      (make-instance 'lattice-span :cs cs :lower lower :upper upper
			     :name (span-name lower upper))))))
(defmethod base-span ((cs coordinate-system))
  (let ((pair (slot-value cs 'lattice-extrema)))
    (get-span (car pair) (cdr pair))))
(defmethod coordinate-spans ((cs coordinate-system))
  (with-slots (spans) cs (hash-values spans)))
(defmethod coordinate-values ((cs coordinate-system))
  (let ((set (coordinate-spans cs)))
    (remove-duplicates (append (mapcar #'lower set) (mapcar #'upper set)))))

;;;; Comparison of coordinate system elements
(defmethod immediately-precedes? 
    ((lower lattice-coordinate) (upper lattice-span))
  (eq (slot-value upper 'lower) lower))
(defmethod immediately-precedes? 
    ((lower lattice-span) (upper lattice-coordinate))
  (eq (slot-value lower 'upper) upper))
(defmethod immediately-precedes?
    ((lower lattice-element) (upper lattice-element))
  nil) ;; hitting this case means both elements were the same class


;; Coordinate-system helper functions
(defmethod initialize-instance :after ((ls lattice-span) &rest args)
  (with-slots (lower upper) ls
    (assert (eq (lattice lower) (lattice upper)))
    (setf (gethash (cons lower upper) (slot-value (lattice ls) 'spans)) ls) 
    (with-slots (succeeds) upper (pushnew ls succeeds))
    (with-slots (precedes) lower (pushnew ls precedes))))
(defun span-name (lower upper)
  (assert (eq (lattice lower) (lattice upper)))
  (format nil "~A::~A" (name lower) (name upper)))
(defun refinement-names (lower num-subs)
  (when (> num-subs 26)
    (warn "More than 26 sub-coordinate names; using non-alphabetic symbols"))
  (flet ((strip-minus (str)
	   (if (eql #\- (elt str (- (length str) 1)))
	       (subseq str 0 (- (length str) 1))
	     str)))
    (let ((num (length (slot-value lower 'precedes)))
	  (lowname (strip-minus (name lower))))
      (mapcar 
       (lambda (n) (format nil "~A.~A~A" lowname num (code-char (+ 65 n))))
       (reverse (count-list (- num-subs 2)))))))
    

;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; COORDINATE-VALUES LINK LATTICE ELEMENTS TO CELLS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; This is the other way to link measuring parameters to cells.  It
;; represents the values of a directional measurement.  When this is a
;; lattice-value, it represents a uniform value over the space.  When
;; it is a lattice-span, it represents a graduated value and must
;; connect between some pair of boundaries.
(defclass coordinate-value (link)
  ((coordinate :initarg :coordinate :accessor tail) ; lattice-element
   (cell :initarg :cell :accessor head))) ;

(defclass uniform-coordinate (coordinate-value) ()) ; lattice-coordinate
(defmethod print-object ((c uniform-coordinate) stream)
  (format stream "<Uniform ~A>" (name (tail c))))

(defclass graduated-coordinate (coordinate-value) ; lattice-span
  ((bound+ :initarg :bound+ :accessor bound+) ; boundary link
   (bound- :initarg :bound- :accessor bound-))) ; boundary link
(defmethod print-object ((g graduated-coordinate) stream)
  (format stream "<Gradient ~A>" (name (tail g))))


;; Constructors of coordinate values
(defmethod make-coordinate-value ((c cell) (v lattice-coordinate)
				  &optional bound- bound+)
  (assert (not (or bound- bound+)) () "No bounds for uniform coordinates")
  (make-instance 'uniform-coordinate :cell c :coordinate v))
(defmethod make-coordinate-value ((c cell) (v lattice-span)
				  &optional bound- bound+)
  (assert (and bound- bound+) () "Graduated coordinate needs bounds")
  (make-instance 'graduated-coordinate :cell c :coordinate v
		 :bound+ bound+ :bound- bound-))
(defmethod ensure-coordinate-value ((c cell) (v lattice-element)
				    &optional bound- bound+)
  (flet ((same (cv)
	   (and (eq v (tail cv))
		(if (typep cv 'graduated-coordinate)
		    (and (eq (bound+ cv) bound+) (eq (bound- cv) bound-))
		  t))))
    (let ((existing (links-of-type c 'coordinate-value t)))
      (or (find-if #'same existing)
	  (make-coordinate-value c v bound- bound+)))))


;; Accessors and comparators
(defmethod coordinates ((c cell))
  (links-of-type c 'coordinate-value t))
(defmethod span-values ((c cell))
  (remove-if-not (ppf typep ?x 'lattice-span) (coordinates c) :key #'value))

(defmethod same-coordinate ((c1 coordinate-value) (c2 coordinate-value))
  nil)					; they must be different types
(defmethod same-coordinate ((c1 uniform-coordinate) (c2 uniform-coordinate))
  (eq (tail c1) (tail c2)))
(defmethod same-coordinate ((c1 graduated-coordinate) (c2 graduated-coordinate))
  (and (eq (tail c1) (tail c2))
       (eq (bound+ c1) (bound+ c2))
       (eq (bound- c1) (bound- c2))))
(defmethod base-span? ((c coordinate-value))
  (eq (tail c) (base-span (lattice (tail c)))))
(defmethod extremum? ((c coordinate-value))
  (let ((base (base-span (lattice (tail c)))))
    (or (eq (tail c) (upper base)) (eq (tail c) (lower base)))))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; MEASUREMENT LINKING PARAMETERS TO CELLS OR COORDINATE-SPANS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Measurements are often, but not always, geometric. They are always,
;; however, non-directional --- directional measures are handled in
;; parallel by coordinate systems.

(defclass measurement (link)
  ((parameter :initarg :parameter :accessor tail)
   (object :initarg :object :accessor head)))
(defmethod print-object ((m measurement) stream)
  (format stream "<Measurement ~A of ~A>" (name (tail m)) (head m)))

;; Find the parameter (if any) used to measure certain units of a cell
; (defun find-measure (units cell)
;   (let ((mset (remove-if (ppf typep ?x 'ordinal-value) (measurements cell))))
;     (let ((pmatch (remove-if-not (lambda (p) (multiset-eqv (units p) units))
; 				 (mapcar #'tail mset))))
;       (assert (<= (length pmatch) 1) (units cell pmatch) 
; 	"Multiple measures of ~A in ~A: ~A" units cell pmatch)
;       (first pmatch))))
(defmethod ensure-measure ((p parameter) (obj element))
  (or (find p (links-of-type obj 'measurement t) :key #'tail)
      (make-instance 'measurement :parameter p :object obj)))

;; What are all of the measurement links on this object?
(defmethod measurements ((obj element))
  (links-of-type obj 'measurement t))
;; Which cells does this parameter measure?
(defmethod measures ((p parameter)) 
  (mapcar #'head (links-of-type p 'measurement nil)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; VALUES OF PARAMETERS
;;;; (including assignment and derivation relations)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defclass value (element)
  ((number :initarg :n :accessor value)))
(defmethod make-value (parameter n &rest derivation)
  (let ((v (make-instance 'value :n n)))
    (make-assignment v parameter)
    (dolist (d derivation)
      (make-derivation v d))
    v))
(defmethod print-object ((v value) stream)
  (with-slots (number) v (format stream "<Value ~A>" number)))

(defclass assignment (link) 
  ((parameter :initarg :parameter :accessor head)
   (value :initarg :value :accessor tail)))
(defmethod make-assignment ((v value) (p parameter))
  (let ((a (make-instance 'assignment :parameter p :value v)))
    (setf (single-value p)
      (when (= (length (links-of-type p 'assignment t)) 1) v))
    a))
(defmethod print-object ((a assignment) stream)
  (with-slots (parameter value) a
    (format stream "<Assignment ~A to ~A>" value (name parameter))))
(defmethod get-values ((p parameter))
  (if (single-value p)
      (list (single-value p))
    (mapcar #'tail (links-of-type p 'assignment t))))

;; ensure that p is assigned to precisely value
(defmethod assert-value ((p parameter) value)
  (let ((vlist (get-values p)))
    (assert (<= (length vlist) 1))
    (if vlist 
	(setf (value (first vlist)) value)
      (make-value p value))))

;; Right now, constraints generally don't know how to handle multiple
;; values, so we've got a method that works with the assumption of single
(defmethod get-single-value ((p parameter))
  (let ((vlist (get-values p)))
    (assert (<= (length vlist) 1))
    (when vlist (value (first vlist)))))


(defclass derivation (link) 
  ((derived :initarg :d :accessor head)
   (source :initarg :s :accessor tail)))
(defmethod make-derivation ((derived value) (source value))
  (make-instance 'derivation :d derived :s source))
(defmethod print-object ((d derivation) stream)
  (with-slots (derived source) d
    (format stream "<~A derived from ~A>" derived source)))

;; find values of derivation
(defmethod derived-from ((v value))
  (mapcar #'tail (links-of-type v 'derivation t)))
;; find parameters of derivation
(defmethod derived-from ((p parameter))
  (let* ((value-sets (mapcar #'derived-from (get-values p)))
	 (values (remove-duplicates (apply #'append value-sets)))
	 (assignments (mapcar (ppf links-of-type ?x 'assignment nil) values)))
    (remove-duplicates (mapcar #'head (apply #'append assignments)))))


;; Attach Values
(defun parameter-values (pv-list)
  (dolist (pv pv-list) 
    (let ((p (find-parameter (first pv))) (v (second pv)))
      (assert p)
      (assert-value p v))))
  


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; RELATIONS 
;;;; (including functional blueprints)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass relation (element)
  ((enabled :initform t :accessor enabled) ; should this relation be ignored?
   (stress :initform 0 :accessor stress))) ; current computed stress
(defun make-relation (&rest parameters)
  (let ((r (make-instance 'relation)))
    (dolist (p parameters) (make-constraint r p))
    r))
;; there might be multiple; if so, return list
(defun find-relation (name) 
  (let ((found (remove-if-not (lambda (e) (and (typep e 'relation) 
					       (eqv name (name e))))
			      (all-elements))))
    (if (= (length found) 1) (first found) found)))
(defmethod resolve ((r relation))
  (error "Don't know how to resolve ~A" r))
(defmethod print-object ((r relation) stream)
  (format stream "<Relation ~A>" (type-of r)))
(defmethod abs-stress ((r relation)) (abs (stress r)))

(defmethod relations ((p parameter))
  (mapcar #'tail (links-of-type p 'constraint t)))
(defmethod parameters ((r relation))
  (mapcar #'head (links-of-type r 'constraint nil)))
(defmethod request-change ((p parameter) (src relation) direction stress)
  (cond ((eq direction '+) (push (cons '+ (abs stress)) (deltas p)))
	((eq direction '-) (push (cons '- (abs stress)) (deltas p)))
	(t (error "Don't know how to change in direction ~A" direction))))

(defclass constraint (link)
  ((parameter :initarg :parameter :accessor head)
   (relation :initarg :relation :accessor tail)))
(defmethod make-constraint ((r relation) (p parameter))
  (let ((c (make-instance 'constraint :relation r :parameter p)))
    (setf (singleton p) (= (length (links-of-type p 'constraint t)) 1))
    c))
(defmethod print-object ((c constraint) stream)
  (with-slots (relation parameter) c
    (with-slots (name) parameter
      (format stream "<Constraint of ~A by ~A>" name relation))))

;;;; Unit Identity relation:
;; Takes one composite unit and its definition in terms of primitive units
;; Note: doesn't yet support inverse units (e.g. Hertz = 1/s, velocity = m/s)
(defclass unit-identity (relation) 
  ((composite :initarg :composite :accessor composite-unit)
   (elementary :initarg :elementary :accessor elementary-units)))

(defmethod find-elementary-parameters ((p parameter))
  (if (not (composite? p))
      (list p)
    (let* ((relations (mapcar #'tail (links-of-type p 'constraint t)))
	   (cmp-id (find-if (ppf typep ? 'unit-identity) relations)))
      (remove p (mapcar #'head (links-of-type cmp-id 'constraint nil))))))

(defvar compress-names t)
(defparameter name-separators "-_")
;; find the last separator between two names, return end of newsuffix on match
(defun shared-prefix (base newsuffix)
  (let* ((diff (mismatch base newsuffix))
	 (lastseparator (position-if (lambda (c) (find c name-separators))
				     (subseq newsuffix 0 diff) :from-end t)))
    (when lastseparator
      (subseq newsuffix (1+ lastseparator)))))

(defun composite-parameter-name (elementaries)
  (if compress-names
      (let ((cmpname "") (curblock ""))
	(dolist (name (sort (mapcar (cmp symbol-name name) elementaries) #'lt))
	  (let ((suffix (shared-prefix curblock name)))
	    (if suffix 
		(setf curblock (format nil "~A~A" curblock suffix))
	      (progn (setf cmpname (format nil "~A~A" cmpname curblock))
		     (setf curblock name)))))
	(setf cmpname "~A~A" cmpname curblock)
	(intern cmpname))
    (intern (format nil "~{~A~}" (sort (mapcar #'name elementaries) #'lt)))))

(defun find-composite-parameter (elementaries)
  (find-parameter (composite-parameter-name elementaries)))
  
;; composite-parameter needs to check and see whether these particular sources
;; have even been composited before
;; If all sources are dimensions, the result is also a dimension
(defun composite-parameter (&rest sources)
  (let ((elementary '()))
    (dolist (s sources)
      (if (composite? s)
	  (let ((defns (remove-if-not (ppf typep ?x 'unit-identity) (relations s))))
	    (assert (= (length defns) 1)
		(s defns) "Wrong number of definitions of ~A: ~A" s defns)
	    (dolist (e (elementary-units (first defns))) (push e elementary)))
	(push s elementary)))
    (or (find-composite-parameter elementary) ; check for reusable
	(let* ((is-dimension 
		(not (position-if-not (ppf typep ? 'dimension) elementary)))
	       (compname (composite-parameter-name elementary))
	       (newunits (mapcar #'units elementary))
	       (newp (if is-dimension 
			 (ensure-dimension compname newunits)
		       (ensure-parameter compname newunits)))
	       (id (make-instance 'unit-identity 
		     :name (format nil "Unit ~A" compname)
		     :composite newp :elementary elementary)))
	  (make-constraint id newp)
	  (dolist (e elementary) (make-constraint id e))
	  newp))))

; ;; should only be called by other composite functions
; (defun make-new-composite-parameter (compname elementary)
;   (let ((newp (make-parameter compname (mapcar #'units elementary)))
; 	(id (make-instance 'unit-identity :name compname)))
;     (make-constraint id newp)
;     (dolist (e elementary) (make-constraint id e))
;     newp))

; (defun named-composite-parameter (compname &rest sources)
;   (or (find-parameter compname)
;       (let ((elementary '()))
; 	(dolist (s sources)
; 	  (if (composite? s)
; 	      (let ((defns (remove-if-not (ppf typep ?x 'unit-identity) (relations s))))
; 		(assert (= (length defns) 1)
; 		    (s defns) "Wrong number of definitions of ~A: ~A" s defns)
; 	    (dolist (e (elementary-units (first defns))) (push e elementary)))
; 	(push s elementary)))
;     (or (find-composite-parameter elementary) ; check for reusable
; 	(let* ((compname (intern (format nil "~{~A~}" (mapcar #'name elementary))))
;   )

  

(defun composite-parameters* (source-set)
  (mapset #'composite-parameter 
	  (remove-if (lambda (s) (< (length s) 2)) (subsets source-set))))

;;;; Upper Bound relation
;;; Ensure that lower parameter value is no bigger than upper value
(defclass upper-bound (relation) 
  ((upper :initarg :upper) (lower :initarg :lower)))
(defmethod make-upper-bound ((upper parameter) (lower parameter))
  (or					; avoid duplication: reuse if possible
   (find-if (lambda (e) (and (typep e 'upper-bound)
			     (eq upper (slot-value e 'upper))
			     (eq lower (slot-value e 'lower))))
	    (all-elements))
   (let ((ub (make-instance 'upper-bound :upper upper :lower lower)))
     (make-constraint ub upper) (make-constraint ub lower)
     ub)))
(defmethod make-upper-bound ((upper symbol) (lower symbol))
  (let ((upp (find-parameter upper)) (lowp (find-parameter lower)))
    (assert (and upp lowp))
    (make-upper-bound upp lowp)))

;;;; Summation relation
;;; Ensure that sub parameter values
(defclass summation (relation) 
  ((total :initarg :total) (summands :initarg :summands)))
(defmethod make-summation ((total parameter) &rest summands)
  (or					; avoid duplication: reuse if possible
   (find-if (lambda (e) (and (typep e 'summation)
			     (eq total (slot-value e 'total))
			     (not (set-exclusive-or summands (slot-value e 'summands)))))
	    (all-elements))
   (let ((sum (make-instance 'summation :total total :summands summands)))
     (make-constraint sum total)
     (mapcar (pcf make-constraint sum) summands)
     sum)))
(defmethod make-summation ((total symbol) &rest summands)
  (let ((pset (mapcar #'find-parameter (cons total summands))))
    (assert (not (position nil pset)))
    (apply #'make-summation pset)))

;;;; Functional Blueprint relation
;; Just a relation subclass that displays differently
;; All content needs to be added by people making subclasses
(defclass functional-blueprint (relation) ())
(defmethod print-object ((fb functional-blueprint) stream)
  (format stream "<~A Functional Blueprint>" (type-of fb)))


;;;;;;
;; Relations that represent equalities rather than constraints:
(defmethod equality-relation ((r relation)) nil)
(defmethod equality-relation ((r summation)) t)
(defmethod equality-relation ((r unit-identity)) t)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; DEBUGGING FUNCTIONS FOR EXAMINING TOPOLOGY
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun compute-topology-stats ()
  (let ((all-l (all-links)) (all-e (all-elements)) (stats '()))
    ;; Statistics of objects:
    (push (list 'objects (length all-e)) stats)
    (push (list 'cells (count-if (ppf typep ? 'cell) all-e)) stats)
    (push (list '0-cells (length (find-cells '((dimension ?c 0))))) stats)
    (push (list '1-cells (length (find-cells '((dimension ?c 1))))) stats)
    (push (list '2-cells (length (find-cells '((dimension ?c 2))))) stats)
    (push (list '3-cells (length (find-cells '((dimension ?c 3))))) stats)
    (push (list 'parameters (count-if (ppf typep ? 'parameter) all-e)) stats)
    (push (list 'dimensions (count-if (ppf typep ? 'dimension) all-e)) stats)
    (push (list 'relations (count-if (ppf typep ? 'relation) all-e)) stats)
    (push (list 'values (count-if (ppf typep ? 'value) all-e)) stats)
    ;; Statistics of links:
    (push (list 'links (length all-l)) stats)
    (push (list 'boundaries (count-if (ppf typep ? 'boundary) all-l)) stats)
    (push (list 'derived-parallels (count-if (ppf typep ? 'derived-parallel) all-l)) stats)
    (push (list 'subspaces (count-if (ppf typep ? 'subspace) all-l)) stats)
    (push (list 'measurements (count-if (ppf typep ? 'measurement) all-l)) stats)
    (push (list 'coordinate-values (count-if (ppf typep ? 'coordinate-value) all-l)) stats)
    (push (list 'graduated (count-if (ppf typep ? 'graduated-coordinate) all-l)) stats)
    (push (list 'uniform (count-if (ppf typep ? 'uniform-coordinate) all-l)) stats)
    (push (list 'constraints (count-if (ppf typep ? 'constraint) all-l)) stats)
    (push (list 'assignments (count-if (ppf typep ? 'assignment) all-l)) stats)
    (push (list 'derivations (count-if (ppf typep ? 'derivation) all-l)) stats)
;; Return stat assoc
    stats))

(defun print-topology-stats ()
  (let ((stats (compute-topology-stats)))
    ;; Statistics of objects:
    (format t "# Objects = ~A~%" (cadr (assoc 'objects stats)))
    (format t "    # cells = ~A (~A 0D, ~A 1D, ~A 2D, ~A 3D)~%"
	    (cadr (assoc 'cells stats)) (cadr (assoc '0-cells stats))
	    (cadr (assoc '1-cells stats)) (cadr (assoc '2-cells stats))
	    (cadr (assoc '3-cells stats)))
    (format t "    # parameters = ~A (~A dimensions)~%"
	    (cadr (assoc 'parameters stats)) (cadr (assoc 'dimensions stats)))
    (format t "    # relations = ~A~%" (cadr (assoc 'relations stats)))
    (format t "    # values = ~A~%" (cadr (assoc 'values stats)))
    ;; Statistics of links:
    (format t "# Links = ~A~%" (cadr (assoc 'links stats)))
    (format t "    # boundary = ~A, # derived-parallel = ~A (~A subspace)~%" 
	    (cadr (assoc 'boundaries stats))
	    (cadr (assoc 'derived-parallels stats))
	    (cadr (assoc 'subspaces stats)))
    (format t "    # measurement = ~A~%"
	    (cadr (assoc 'measurements stats)))
    (format t "    # coordinate = ~A (~A graduated, ~A uniform))~%" 
	    (cadr (assoc 'coordinate-values stats))
	    (cadr (assoc 'graduated stats))
	    (cadr (assoc 'uniform stats)))
    (format t "    # constraint = ~A~%" (cadr (assoc 'constraints stats)))
    (format t "    # assignment = ~A, # derivation = ~A~%"
	    (cadr (assoc 'assignments stats)) (cadr (assoc 'derivations stats)))
    ))
