;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EVOLVING VALUES VIA CONSTRAINTS
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Two-stage resolution:
;; 1. All parameters attempt to adjust their values
;; 2. All relations assert changes to those values
(defvar current-max-stress 0)
(defvar current-max-abs-stress 0)
(defun resolve-all ()
  (dolist (p (all-parameters)) (resolve p))
  (dolist (r (all-relations)) (resolve r))
  (setf current-max-stress (max-net-stress) current-max-abs-stress (max-stress))
  (record-log)
  )

;; just like resolution, but it only runs parameter completions
;; also clears stress and marks afterward
(defun propagate-all ()
  (dolist (r (all-relations)) (resolve r) (setf (stress r) 0))
  (dolist (p (all-parameters)) (setf (deltas p) nil)))

  


(defun report-stress ()
  (mapcar (lambda (r) (list r (stress r))) (all-relations)))

(defun total-stress ()
  (apply #'+ (mapcar (cmp abs second) (report-stress))))
(defun max-stress ()
  (apply #'max (mapcar (cmp abs second) (report-stress))))
(defun max-net-stress ()
  (apply #'max (mapcar #'stress (all-parameters))))

(defun find-stressed-relations (&optional only-nonviable)
  (let* ((relations (remove-if-not (ppf typep ?e 'relation) (all-elements)))
	 (report (remove-if-not 
		  (lambda (r) (> (abs (stress r)) (if only-nonviable 1 0)))
		  relations)))
    report))
(defun problem-summary (&optional only-nonviable)
  (let ((set (find-stressed-relations only-nonviable)))
    (format t "~:{~A (~,3F):~:{ ~A = ~,3E~}~%~}"
	    (mapcar 
	     (lambda (r)
	       (list (type-of r) (stress r) 
		     (mapcar 
		      (lambda (p) (list (name p) (get-single-value p)))
		      (mapcar #'head (links-of-type r 'constraint nil)))))
	     (sort set #'> :key (cmp abs stress))))))
(defun find-stressed-parameters (&optional only-nonviable)
  (let* ((report (remove-if-not 
		  (lambda (p) (> (abs (stress p)) (if only-nonviable 1 0)))
		  (all-parameters))))
    report))


;; Total stress on parameter is max of all the requested change stresses
;; times sum of changes over abs of changes
(defmethod stress ((p parameter))
  (if (not (deltas p))
      0
    (let* ((changereqs (mapcar #'delta-to-changereq (deltas p)))
	   (absreqs (mapcar #'abs changereqs)))
      (* (apply #'max absreqs) 
	 (/ (apply #'+ changereqs) (apply #'+ absreqs))))))
(defmethod abs-stress ((p parameter))
  (apply #'+ (mapcar (cmp abs delta-to-changereq) (deltas p))))
;; are all stresses asking for changes in the same direction?
;; also returns true for unstressed
(defmethod stress-unopposed? ((p parameter))
  (let ((changereqs (mapcar #'delta-to-changereq (deltas p))))
    (or (not (position-if #'plusp changereqs))
	(not (position-if #'minusp changereqs)))))

(defun delta-to-changereq (d)
  (cond ((eq (car d) '+) (cdr d))
	((eq (car d) '-) (- (cdr d)))
	(t (error "Delta type must be + or -: ~A" (car d)))))
(defun viability-threat (stress)
  (let ((safe-stress (min 1 (max (abs stress) 0))))
    (assert (< safe-stress 1) () "Constraint has become nonviable!")
    (/ 1 (- 1 safe-stress))))
(defun sign (n) (if (minusp n) -1 (if (plusp n) 1 0)))
; ;; stress ratio option
; (defmethod compute-step-factor ((p parameter))
;   (let* ((s-global current-max-stress)
; 	 (s-local (stress p))
; 	 (step-factor (if (> s-global 0) (/ s-local s-global) 0)))
;     step-factor))
; ;; viability threat option
; (defmethod compute-step-factor ((p parameter))
;   (let* ((t-global (viability-threat current-max-stress))
; 	 (t-local (viability-threat (stress p)))
; 	 (step-factor (if (> t-global 0) (/ t-local t-global) 0)))
;     step-factor))
; ;; count deltas option
; (defmethod compute-step-factor ((p parameter))
;   (if (plusp (length (deltas p)))
;       (let ((scale (/ (- (count '+ (deltas p) :key #'car) 
; 			 (count '- (deltas p) :key #'car)) 
; 		      (length (deltas p)))))
; 	scale)
;     0))
;; stress ratio, plus minimum
(defmethod compute-step-factor ((p parameter))
  (let* ((s-global current-max-abs-stress)
	 (s-local (stress p))
	 (step-factor (if (> s-global 0) (/ s-local s-global) 0)))
    ;;(* (sign step-factor) (max 0.1 (abs step-factor)))))
    (* 1 (sign step-factor) (abs step-factor))))

(defparameter step-size 0.005)		; base 0.5% stretch per step
(defparameter singleton-step-size 0.1)  ; base 10% per step for "easy" targets
(defmethod resolve ((p parameter))
  (let* ((v (get-values p))
	 (sf (compute-step-factor p)))
    (assert (<= (length v) 1))		; Can only handle single assignments
    (setf v (first v))
    (when (and v (> (abs sf) 0))
      (let* ((maxstep step-size);(if (singleton p) singleton-step-size step-size))
	     (step (* sf maxstep)))
	(when (> (abs step) maxstep)
	  (setf step (* step (/ maxstep (abs step)))))
	(setf (value v) (* (value v) (+ 1 step)))))
    (setf (deltas p) nil)))
;     (when (and v (> (abs t-local) 0))
;       (setf (value v) (* (value v) (+ 1 (* step-factor step-size)))))
;     (setf (deltas p) nil)))

(defmethod propagate-priority-on-derivations ((p parameter))
  (unless (mark? p 'priority)
    (let* ((sources (derived-from p))
	   (src-priority
	    (remove nil (mapcar (ppf mark? ?x 'priority) sources))))
      (when src-priority
	(mark p 'priority (+ 2 (apply #'min src-priority)))))))

(defvar stress-epsilon 0.00001)
(defmethod resolve :around ((r relation))
  (when (enabled r)
    (let ((v (call-next-method r)))
      (unless (or (< (abs (stress r)) stress-epsilon) (mark? r 'priority))
	(let* ((src-priority 
		(remove nil
			(mapcar (ppf mark? ?x 'priority) (parameters r))))
	       (min-priority (if src-priority (apply #'min src-priority) 0)))
	  ;; also should put in the derivation link!
	  (when src-priority 
	    ;;(format t "Marking priority of ~A from neighbors to ~A, based on ~A neighbors, stress ~A~%" (pretty-name r) (1+ min-priority) (length src-priority) (stress r))
	    (mark r 'priority (1+ min-priority)))))
      (mapcar #'propagate-priority-on-derivations (parameters r))
      v)))

(defmethod resolve :around ((p parameter))
  (let* ((stress (abs (stress p)))	; will be zeroed by call-next-method
	 (v (call-next-method p)))
    (unless (or (< stress stress-epsilon) (mark? p 'priority))
      (let* ((src-priority 
	      (remove nil
		      (mapcar (ppf mark? ?x 'priority) (relations p))))
	     (min-priority (if src-priority (apply #'min src-priority) 0)))
	;; also should put in the derivation link!
	(when src-priority 
	  ;;(format t "Marking priority of ~A from neighbors to ~A, based on ~A neighbors, stress ~A~%" (pretty-name p) (1+ min-priority) (length src-priority) stress)
	  (mark p 'priority (1+ min-priority)))))
    v))

(defparameter use-priorities nil)
(defmethod request-change :around ((p parameter) (src relation) dir stress)
  (let* ((p-pri (mark? p 'priority))
	 (s-pri (mark? src 'priority)))
    (when (and 
	   (not (mark? p 'specification)) ; don't change specs
	   (or (not use-priorities)
	       (or (and s-pri (not p-pri)) (and p-pri s-pri (< s-pri p-pri)))))
      (call-next-method p src dir stress))))

    

;;; Old update model:
; (defmethod resolve ((p parameter))
;   (let* ((v (get-values p)))
;     (assert (<= (length v) 1))		; Can only handle single assignments
;     (setf v (first v))
;     (when (and v (plusp (length (deltas p))))
;       (let ((scale (/ (- (count '+ (deltas p) :key #'first) 
; 			 (count '- (deltas p) :key #'first)) 
; 		      (length (deltas p)))))
; 	(setf (value v) (* (value v) (+ 1 (* scale step-size))))))
;     (setf (deltas p) nil)))

(defmethod riser (r p) nil)
;(defmethod riser ((r unit-identity) (p parameter))
;  (eq (composite-unit r) p))

(defparameter stretch-tolerance 1.5)	; viable up to 50% difference
(defmethod resolve ((u unit-identity))
  (let ((cu (composite-unit u)) (eu-set (elementary-units u)))
    (let ((cv (get-values cu)) (ev (mapcar #'get-values eu-set)))
      ;; Don't know how to handle it if there are multiple values assigned
      (assert (<= (length cv) 1))
      (assert (= 0 (count-if (lambda (v) (> (length v) 1)) ev)))
      (setf cv (first cv)) (setf ev (mapcar #'first ev))
      ;; Assert constraint
      (setf (stress u) 0)		; default to zero stress
      (cond 
       ;; No gaps: check to see whether they match
       ((and cv (= 0 (count nil ev)))	; nothing missing
	(let ((cvn (value cv)) (evn (apply #'* (mapcar #'value ev))))
	  (if (= cvn evn) 
	      t				; all OK
	    ;; Request shift to bring things closer into alignment
	    (let ((cvdir (if (< cvn evn) '+ '-)) (evdir (if (< cvn evn) '- '+))
		  (stress
		   (max 0 (- (abs (log (/ cvn evn) stretch-tolerance)) 0.05))))
	      (when (> (abs stress) 0)
		(setf (stress u) stress)
		(request-change cu u cvdir stress)
		(mapcar (ppf request-change ? u evdir stress) eu-set))))))
       ;; One gap: fill in gap from asserted elements
       ((and (eq cv nil) (= 0 (count nil ev)))
	(apply #'make-value cu (apply #'* (mapcar #'value ev)) ev))
       ((and cv (= 1 (count nil ev)))
	(let ((missing (elt eu-set (position nil ev)))
	      (newval (apply #'/ (value cv) (mapcar #'value (remove nil ev)))))
	  (apply #'make-value missing newval (cons cv (remove nil ev)))))
       ;; Two or more gaps: do nothing, remaining satisfied
       (t t)))))


(defparameter warning-threshold 0.9)	; warn when within 90%
(defmethod resolve ((ub upper-bound))
  (with-slots (upper lower) ub
    (let ((uv (get-values upper)) (lv (get-values lower)))
      ;; Don't know how to handle it if there are multiple values assigned
      (assert (<= (length uv) 1)) (assert (<= (length lv) 1))
      (setf uv (first uv)) (setf lv (first lv))
      (setf (stress ub) 0)
      (when (and uv lv)
	(assert (and (>= (value lv) 0) (>= (value uv) 0)))
	(let ((ratio (/ (value lv) (value uv))))
	  (when (> ratio warning-threshold)
	    (setf (stress ub)
	      (/ (- ratio warning-threshold) (- 1 warning-threshold)))
	    (request-change upper ub '+ (stress ub))
	    (request-change lower ub '- (stress ub))))))))
(defmethod riser ((ub upper-bound) (p parameter))
  (with-slots (upper) ub (eq p upper)))

(defmethod resolve ((s summation))
  (with-slots (total summands) s
    (let ((tv (get-values total)) (sv (mapcar #'get-values summands)))
      ;; Don't know how to handle it if there are multiple values assigned
      (assert (<= (length tv) 1))
      (assert (= 0 (count-if (lambda (v) (> (length v) 1)) sv)))
      (setf tv (first tv)) (setf sv (mapcar #'first sv))
      ;; Assert constraint
      (setf (stress s) 0)		; default to zero stress
      (cond 
       ;; No gaps: check to see whether they match
       ((and tv (= 0 (count nil sv)))	; nothing missing
	(let ((tvn (value tv)) (svn (apply #'+ (mapcar #'value sv))))
	  (if (= tvn svn) 
	      t				; all OK
	    ;; Request shift to bring things closer into alignment
	    (let ((tvdir (if (< tvn svn) '+ '-)) (svdir (if (< tvn svn) '- '+))
		  (stress 
		   (max 0 (- (abs (log (/ tvn svn) stretch-tolerance)) 0.05))))
	      (when (> stress 0)
		(setf (stress s) stress)
		(request-change total s tvdir stress)
		(let* ((svs (mapcar #'value sv))
		       (sumtotal (apply #'+ svs))
		       (rel-stress (mapcar (ppf * stress (/ sumtotal)) svs)))
		  (mapcar (ppf request-change ?x s svdir ?y) summands rel-stress)))))))
       ;; One gap: fill in gap from asserted elements
       ((and (eq tv nil) (= 0 (count nil sv)))
	(apply #'make-value total (apply #'+ (mapcar #'value sv)) sv))
       ((and tv (= 1 (count nil sv)))
	(let ((missing (elt summands (position nil sv)))
	      (newval (apply #'- (value tv) (mapcar #'value (remove nil sv)))))
	  (apply #'make-value missing newval (cons tv (remove nil sv)))))
       ;; Two or more gaps: do nothing, remaining satisfied
       (t t)))))


(defclass modular-relation (relation)
  ((varsets :initarg :varsets)		; list of lists of variable for fn calls
   (risers :initarg :risers)		; variables that go up with + stress
   (fallers :initarg :fallers)		; variables that go down with + stress
   (stress-fn :initarg :stress-fn)	; function for computing stress
   ;; functions for filling in one or n missing variables
   (completion-fn :initform nil :initarg :cmp-fn) 
   (n-completion-fn :initform nil :initarg :n-cmp-fn)))
(defmethod riser ((r modular-relation) (p parameter))
  (with-slots (risers) r (tp (position p risers))))

(defmethod initialize-instance :after ((r modular-relation) &rest args)
  (with-slots (varsets) r
    (dolist (p (flatten varsets))
      (make-constraint r p))))

(defun make-modular-relation (varsets risers fallers stress-fn 
			      &optional completion-fn n-completion-fn)
  (let* ((paramsets (mapcar (lambda (l) (mapcar #'find-parameter l)) varsets))
	 (allparams (apply #'append paramsets))
	 (p-riser (mapcar #'find-parameter risers))
	 (p-faller (mapcar #'find-parameter fallers)))
    (assert (not (position nil allparams)))
    (assert (and (subsetp p-riser allparams) (subsetp p-faller allparams)))
    (make-instance 'modular-relation
      :varsets paramsets :risers p-riser :fallers p-faller
      :stress-fn stress-fn :cmp-fn completion-fn :n-cmp-fn n-completion-fn)))

(defmethod resolve ((r modular-relation))
  (with-slots (varsets risers fallers stress-fn completion-fn n-completion-fn) r
    (setf (stress r) 0)			; default to zero stress
    (let* ((valsets (get-variable-values varsets))
	   (num-missing (count-missing-values valsets)))
      (cond ((= num-missing 0)
	     (setf (stress r) (apply stress-fn r valsets))
	     (cond ((= (stress r) 0) 'do-nothing)
		   ((> (stress r) 0) (request-changes risers fallers r (stress r)))
		   ((< (stress r) 0) (request-changes fallers risers r (stress r)))))
	    ((and (= num-missing 1) completion-fn)
	     (let ((cv (apply completion-fn r valsets)))
	       (attach-missing-value cv varsets valsets)))
	    ((and (< num-missing (num-variables varsets)) n-completion-fn)
	     (error "n completion resolution not yet implemented")
	     (apply n-completion-fn r valsets))
	    (t t))))) ;; do nothing: too much information is missing

(defun get-variable-values (varsets)
  (mapcar (lambda (set) (mapcar #'get-single-value set)) varsets))
(defun count-missing-values (valsets)
  (apply #'+ (mapcar (pcf count nil) valsets)))
(defun num-variables (varsets)
  (length (flatten varsets)))
(defun request-changes (plus-set minus-set requester stress)
  (mapcar (ppf request-change ?x requester '+ (abs stress)) plus-set)
  (mapcar (ppf request-change ?x requester '- (abs stress)) minus-set))
(defun attach-missing-value (v varsets valsets)
  (let* ((pairs (mapcar #'cons (apply #'append valsets) 
			(apply #'append varsets)))
	 (var (cdr (assoc nil pairs))))
    (assert var)
    (make-value var v)))		; ignoring derivation for now

;; makes a stress function by composing an evaluation function with a 
;; linear stress function
(defun eval-stress (eval-fn stress-fn)
  (lambda (&rest args) (funcall stress-fn (apply eval-fn args))))
    
(defun linear-stress-fn (nonviable- stress- stress+ nonviable+)
  (lambda (v)
    (cond ((< v nonviable-) -1)
	  ((< v stress-) (/ (- v stress-) (- stress- nonviable-)))
	  ((< v stress+) 0)
	  ((< v nonviable+) (/ (- v stress+) (- nonviable+ stress+)))
	  (t 1))))

(defun positive-linear-stress-fn (stress+ nonviable+)
  (lambda (v)
    (cond ((< v stress+) 0)
	  ((< v nonviable+) (/ (- v stress+) (- nonviable+ stress+)))
	  (t 1))))

;;; IDENTITY-RELATION
;; An identity relation should always have value = 1
;; stress rises from 0 to 1 over tolerance -- max-tolerance
(defclass identity-relation (modular-relation)
  ((constant :initarg :constant)
   (constant-units :initarg :constant-units :initform '())))

(defun identity-eval (relation numerator-v denominator-v)
  (with-slots (constant) relation
    (* constant (/ (apply #'* numerator-v) (apply #'* denominator-v)))))

(defun identity-completion (relation numerator-v denominator-v)
  (with-slots (constant) relation
    (let ((subv (* constant (/ (apply #'* (remove nil numerator-v))
			       (apply #'* (remove nil denominator-v))))))
      (if (position nil numerator-v) (/ subv) subv))))

;; defaults to stress on the range of 1% to 50% error
(defun make-identity-relation (constant numerator-var denominator-var
			       &optional (tolerance 1.01) (max-tolerance 1.5))
  (let* ((n-param (mapcar #'find-parameter numerator-var))
	 (d-param (mapcar #'find-parameter denominator-var)))
    (assert (not (position nil (append n-param d-param))))
    (let ((lsf (linear-stress-fn (/ 1 max-tolerance) (/ 1 tolerance)
				 tolerance max-tolerance)))
      (make-instance 'identity-relation :varsets (list n-param d-param)
		     :risers d-param :fallers n-param
		     :stress-fn (eval-stress #'identity-eval lsf)
		     :cmp-fn #'identity-completion
		     :constant (car constant)
		     :constant-units (cdr constant)))))


;;; COMPOSITE-SUM-BOUND-RELATION
;; An identity relation should always have value = 1
(defclass composite-sum-bound (modular-relation)
  ((constants :initarg :constants)))

(defun composite-sum-bound-eval (relation bounded-v bound-v)
  (with-slots (constants) relation
    (/ (apply #'+ (mapcar #'* (first constants) (append bounded-v '(1))))
       (apply #'+ (mapcar #'* (second constants) (append bound-v '(1)))))))

(defun ensure-sb-constants-length (constants bounded bound)
  (let ((lc1 (length (first constants))) (lc2 (length (second constants)))
	(lb1 (length bounded)) (lb2 (length bound)))
    (assert (or (= lc1 lb1) (= lc1 (1+ lb1))))
    (assert (or (= lc2 lb2) (= lc2 (1+ lb2))))
    (when (< lb1 lc1) (rplacd (last (first constants)) '(0)))
    (when (< lb2 lc2) (rplacd (last (second constants)) '(0)))))
;; defaults to stress on the range of +/- 5%
;; Maintains: sum(c1*bounded_var) < sum(c2*bound_var)
(defun make-sum-bound-relation (constants bounded-var bound-var &optional (tolerance -0.05) (max-tolerance 0.05))
  (let* ((bounded-param (mapcar #'find-parameter bounded-var))
	 (bound-param (mapcar #'find-parameter bound-var)))
    (ensure-sb-constants-length constants bounded-var bound-var)
    (assert (not (position nil (append bounded-param bound-param))))
    (let ((plsf (positive-linear-stress-fn (+ 1 tolerance) (+ 1 max-tolerance))))
      (make-instance 'composite-sum-bound
	:varsets (list bounded-param bound-param)
	:risers bound-param :fallers bounded-param
	:stress-fn (eval-stress #'composite-sum-bound-eval plsf)
	:constants constants))))

;;; KLUDGE QUANTIZED COMPONENT
;; Right now, this supports only the components that are
;; fixed to a single value, and keeps them fixed there
(defclass quantized-component (relation)
  ((library :initarg :library :accessor library)
   (parameter-tuple :initarg :params :accessor params)
   (stress-fn :initarg :sfn)))

(defun make-quantized-component (library &rest parameter-list)
  (let* ((pl (mapcar #'find-parameter parameter-list)))
    (assert (not (position nil pl)))
    (let ((r (make-instance 'quantized-component :library library :params pl
			    ;; 1% to 50% error
			    :sfn (linear-stress-fn (/ 1.5) (/ 1.01) 1.01 1.5))))
      (dolist (p pl) (make-constraint r p))
      r)))

(defmethod resolve ((q quantized-component)) 
  (setf (stress q) 0) 
  (with-slots (library parameter-tuple stress-fn) q
    (let* ((values (mapcar #'get-single-value parameter-tuple)))
      (dolist (pv (mapcar #'list parameter-tuple values library))
	(let ((p (first pv)) (v (second pv)) (ref (third pv)))
	  (when v
	    (let ((pstress (funcall stress-fn (/ v ref))))
	      (setf (stress q) (max (stress q) (abs pstress)))
	      (cond ((= pstress 0) 'do-nothing)
		    ((> pstress 0) (request-change p q '- pstress))
		    ((< pstress 0) (request-change p q '+ pstress))))))))))

;;; PLACEHOLDERS
;; These are used where we know a set of parameters are tied by a 
;; functional blueprint, but that functional blueprint is not yet
;; actually implemented
(defclass placeholder-functional-blueprint (relation) ())
(defun make-placeholder-fb (annotation &rest parameter-list)
  (declare (ignore annotation))
  (let* ((pl (mapcar #'find-parameter parameter-list)))
    (assert (not (position nil pl)))
    (let ((r (make-instance 'placeholder-functional-blueprint)))
      (dolist (p pl) (make-constraint r p))
      r)))
  
(defmethod resolve ((p placeholder-functional-blueprint)) 
  (setf (stress p) 0) 
  (pprint "Not resolving placeholder blueprints"))


;;; USER BLUEPRINTS
(defclass user-blueprint (modular-relation)
  ((constant :initarg :constant)
   (goal :initarg :goal)))

(defun goal-stress (relation values)
  (with-slots (constant goal) relation
    (let ((diff (/ (apply #'* values) goal)))
      (cond ((< diff 0.99) constant)
	    ((> diff 1.01) (- constant))
	    (t 0)))))

; (defun add-user-target (variable value &optional (stress 0.9))
;   (let ((vp (find-parameter variable)))
;     (assert vp)
;     (make-instance 'user-blueprint :varsets (list (list vp))
; 		   :risers (list vp) :fallers nil
; 		   :stress-fn #'goal-stress :constant stress :goal value)))
;; kinda kludgey, since we're just throwing a random number on the top stress
(defun add-user-target (variable value &optional (pull 1.1))
  (let* ((current (parameter-value variable))
	 (u (units variable))
	 (r (if (> value current)
		(make-identity-relation (cons value u) 
					'() (list variable) 1.01 pull)
	      (make-identity-relation (cons (/ value) u)
				      (list variable) '() 1.01 pull))))
    (name-it "user-target" r) 
    r))

;;;;;
;; Easy UI for running sims
(defvar design-name "blueprint")

(defvar initial-values '())
(defvar value-log '())
(defvar parameter-stress-log '())
(defvar relation-stress-log '())
(defvar log-length 0)

(defun clear-log () 
  (setf log-length 0)
  (setf value-log nil parameter-stress-log nil relation-stress-log nil))
(defun record-log ()
  (let ((pset (all-parameters)) (rset (all-relations)))
    (incf log-length)
    (push (mapcar (lambda (r) (list r (stress r))) rset) relation-stress-log)
    (push (mapcar (lambda (p) (list p (stress p))) pset) parameter-stress-log)
    (push (mapcar (lambda (p) (list p (get-single-value p))) pset) value-log)
    'logged))
(defun window-values (log n)
  (let ((subset (subseq log 0 (min n log-length))))
    (reduce (pcf merge-sorted-alist-pair #'append) subset)))


(defun all-parameter-values ()
  (let ((params (remove-if-not (ppf typep ?e 'parameter) (all-elements))))
    (sort (mapcar (lambda (p) (list (name p) (get-single-value p))) params)
	  #'lt :key #'first)))

(defun one-step ()
  (dotimes (i 1) (resolve-all))
  (output-graphviz (format nil "~A-with-values.dot" design-name) (/ 11 8.5)))
(defun ten-steps ()
  (dotimes (i 10) (resolve-all))
  (output-graphviz (format nil "~A-with-values.dot" design-name) (/ 11 8.5)))
(defun startup-propagate ()
  (clear-log)
  (dotimes (i 10) (propagate-all))
  (setf initial-values (all-parameter-values))
  (output-graphviz (format nil "~A-with-values.dot" design-name) (/ 11 8.5)))

(defun make-stress-graph (steps &optional (report-every 10))
  (startup-propagate)
  (let ((stress-chart '()))
    (dotimes (i steps) 
      (resolve-all) 
      (when (= (mod i report-every) 0) (format t "~,2F " (total-stress)))
      (when (= (mod (1+ i) (* 15 report-every)) 0) (format t "~%" (total-stress)))
      (push (total-stress) stress-chart))
    (output-graphviz (format nil "~A-with-values.dot" design-name) (/ 11 8.5))
    (reverse stress-chart)))

(defun value-changes ()
  (flet ((vdiff (va vb) 
	   (assert (eqv (first va) (first vb))) 
	   (list (first va) (/ (second va) (second vb))))
	 (ratio-from-one (v) (if (> v 1) v (/ v))))
    (sort (mapcar #'vdiff (all-parameter-values) initial-values)
	  #'> :key (cmp ratio-from-one second))))

(defun parameter-value (name)
  (get-single-value (find-parameter name)))


;; handle simulation / caching
(defparameter sim-cache-length 10)
(defvar num-simulations-run 0)
(defmethod resolve :around ((r modular-relation))
  (if (not (mark? r 'simulation))
      (call-next-method r)
    (let ((cache (mark? r 'cached-stress)))
      (if (or (not cache) (>= (car cache) sim-cache-length))
	  (progn
	    (call-next-method r)
	    (incf num-simulations-run)
	    (mark r 'cached-stress (cons 0 (stress r))))
	(with-slots (risers fallers) r
	  (setf (stress r) (cdr cache))
	  (mark r 'cached-stress (cons (1+ (car cache)) (cdr cache)))
	  (cond ((= (stress r) 0) 'do-nothing)
		((> (stress r) 0) (request-changes risers fallers r (stress r)))
		((< (stress r) 0) (request-changes fallers risers r (stress r))))
	  )))))
