;; This file contains developmental operators for generating 
;; a representation incrementally via shape transformation rules

;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SEED GENERATION
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Create an orthotope (n-dimensional rectangle) by a sequence of extrusions
(defun extrude-orthotope (units &rest dims)
  (let ((construct (make-cell 0)))
    (dolist (d dims)
      (let ((d-real (ensure-dimension d units)))
	(setf construct (extrude construct d-real))))
    construct))

;; Specialized versions for 2D and 3D
(defun extrude-rectangle (&optional (d1 'x) (d2 'y))
  (extrude-orthotope 'meters d1 d2))
(defun extrude-box (&optional (d1 'x) (d2 'y) (d3 'z))
  (extrude-orthotope 'meters d1 d2 d3))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; EXTRUDE OPERATION
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Create a k+1-cell of length p connecting the k-cell to a copy of itself
;; All relations are carried along

;; Take a set of faces and extrude with length equal to parameter p
(defmethod extrude ((l list) (d dimension))
  (let* ((newcoord (make-coordinate-system d))
	 (-face (remove-duplicates (apply #'append l (mapcar #'boundaries* l))))
	 (+face (mapcar (lambda (c) (make-cell (dimension c))) -face))
	 (connectors (mapcar (lambda (c) (make-cell (1+ (dimension c)))) -face))
	 (sets (mapcar #'list -face connectors +face))
	 (param-map (make-hash-table)))
    ;; First, copy existing parameter, measurement, and boundary links
    ;; from -cell to +cell and connector
    (dolist (s sets)
      (let ((-cell (first s)))
	(dolist (c (coordinates -cell)) (copy-coordinate-value c s sets))
	(dolist (m (measurements -cell)) (copy-measurement m s d param-map))
	(dolist (b (boundaries -cell)) (copy-boundaries b s sets))))
    ;; Then add new links (needs to be after so they aren't copied)
    (dolist (s sets)
      (let ((-cell (first s)) (connector (second s)) (+cell (third s)))
	;; Add links for the new parameter
	(ensure-coordinate-value -cell (lower (base-span newcoord)))
	(ensure-coordinate-value +cell (upper (base-span newcoord)))
	(ensure-coordinate-value connector (base-span newcoord) -cell +cell)
	(when (= (dimension connector) 1) (ensure-measure d connector))
	;; Create new boundary links
	(make-boundary connector -cell)
	(make-boundary connector +cell)))
    ;; return a list of new k+1 cells derived from the input list
    (subseq connectors 0 (length l))))

;; Single-cell syntactic sugar
(defmethod extrude ((c cell) (d dimension))
  (first (extrude (list c) d)))

(defmethod copy-measurement ((m measurement) set newdim param-map)
  (let ((connector (second set)) (+cell (third set)))
    (ensure-measure (tail m) +cell)
    (if (typep (tail m) 'dimension)
	(let ((newp (ensure (gethash (tail m) param-map)
			    (composite-parameter newdim (tail m)))))
	  (ensure-measure newp connector))
      (ensure-measure (tail m) connector))))

(defmethod copy-coordinate-value ((v uniform-coordinate) copyset sets)
  (declare (ignore sets))
  (dolist (target (cdr copyset))
    (make-coordinate-value target (tail v))))
(defmethod copy-coordinate-value ((v graduated-coordinate) copyset sets)
  (let ((+set (assoc (bound+ v) sets))
	(-set (assoc (bound- v) sets)))
    (dolist (args (cdr (mapcar #'list copyset +set -set)))
      (make-coordinate-value (first args) (tail v) (second args) 
			     (third args)))))

(defmethod copy-boundaries ((bound cell) copyset sets)
  (let ((equivs (assoc bound sets)))
    (make-boundary (second copyset) (second equivs))
    (make-boundary (third copyset) (third equivs))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; CONNECT OPERATION
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Connect is like extrude, but only the center cell needs creation
;; We visualize it as an extrusion from the -cell to the +cell
;; The connection may become twisted along the way
; (defmethod connect ((c1 cell) (c2 cell) (d dimension))
;   (assert (= (dimension c1) (dimension c2)))
;   (let* ((-face (remove-duplicates (cons c1 (boundaries* c1))))
; 	 (+face (remove-duplicates (cons c2 (boundaries* c2))))
; 	 (connectors (mapcar (lambda (c) (make-cell (1+ (dimension c)))) -face))
; 	 (sets (mapcar #'list -face connectors +face))
; 	 (dim-set (base-span-set d))
; 	 (param-map (make-hash-table)))
;     (assert (= (length -face) (length +face)))
;     (dolist (s sets)
;       (let ((-cell (first s)) (connector (second s)) (+cell (third s)))
; 	;; Copy existing parameter links from -cell to connector
; 	(dolist (m (measurements -cell))
; 	  (cond ((typep m 'coordinate-value)
; 		 (make-coordinate-value +cell (tail m) (value m))
; 		 (make-coordinate-value connector (tail m) (value m)))
; 		(t						     ; just a measurement
; 		 (let ((type (type-of m)))
; 		   (make-instance type :parameter (tail m) :cell +cell)
; 		   (let ((newp (ensure (gethash (tail m) param-map)
; 				       (composite-parameter d (tail m)))))
; 		     (make-instance type :parameter newp :cell connector))))))
; 	;; Copy boundaries from -cell to +cell and connector
; 	(dolist (b (boundaries -cell))
; 	  (let ((equivs (assoc b sets)))
; 	    (make-boundary connector (second equivs))
; 	    (make-boundary +cell (third equivs))))))
;     (dolist (s sets)
;       (let ((-cell (first s)) (connector (second s)) (+cell (third s)))
; 	;; Add links for the new parameter
; 	(mapcar (ppf make-coordinate-value ?x d ?y) s dim-set)
; 	(when (= (dimension connector) 1)
; 	  (make-instance 'coordinate-span :parameter d :cell connector))
; 	;; Create new boundary links
; 	(make-boundary connector -cell)
; 	(make-boundary connector +cell)))
;     ;; TODO: perhaps this should be returning a list of size = size l?
;     (first connectors)))		; return a new k+1-cell



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; SUBSPACE OPERATION
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Temporarily disabled, waiting time to finish cleaning up
; (defmethod subspace ((parent cell) (anchor cell) coord-map)
;   (when (find anchor (boundaries* parent)) ; fail if anchor is detached
;     (let ((cmap (make-coordinate-map coord-map)))
;       ;; Duplicate or share cells to make child
;       ;; Use coordinate map to mark parameters, measurements, and boundaries
;       (dolist (c (coordinates parent))
; 	(dolist (newc (lookuc m in map))
; 	  (copy-coordinate-value newc par/child sets)))
;       (dolist (m (measurements parent))
; 	(dolist (newm (lookup m in map))
; 	  (ensure-measure newm child)))
;       (dolist (b (boundaries parent))
; 	(ensure-boundary child (assoc (head b) sets)))
;       ;; mark parent with gap values
;       )))


; ;; Create a hashtable that maps 
; (defun make-coordinate-map (cell anchor coord-map)
;   ;; Begin by finding the base sections
;   (let ((roots (mapcar #'tail (coordinates anchor)))
; 	(spans (remove-if-not (ppf typep ? 'graduated-coordinate)
; 			      (coordinates cell))))
;     ;; For each coordinate system mentioned in the map, do a direct mapping
;     (dolist (entry coord-map)
;       (let ((
;     ;; For each other coordinate, map it based on the mapped coordinate 
;     ;; with matching spans
;     ))

;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;; OLD Version

;; Create-subspace-dimensions:
;; names is a list of three items: (lower-gap span upper-gap)
;; In the case where a lower-gap or upper-gap does not exist, it is nil
;; In the case where it exists, it is given its name or ? for anonymous
(defmethod create-and-map-subspace-dimension ((ht hash-table) (ls lattice-span) 
					       names unit)
  (assert (and (listp names) (= (length names) 3)))
  (destructuring-bind (lower-gap span upper-gap) names
    (let* ((olddim (dimension ls))
	   (newdim (ensure-dimension span unit))
	   (newspan (base-span newdim))
	   (d-lower (when (and lower-gap (not (eq lower-gap '?)))
		      (ensure-dimension lower-gap unit)))
	   (d-upper (when (and upper-gap (not (eq upper-gap '?)))
		      (if (eq upper-gap lower-gap)
			  d-lower
			(ensure-dimension upper-gap unit)))))
      ;; Ensure that parallel links exist:
      (ensure-parallel olddim newdim)
      (when d-lower (ensure-parallel olddim d-lower))
      (when d-upper (ensure-parallel olddim d-upper))
      ;; Map span to new dimension's base span
      (push newspan (gethash ls ht))
      (push (lower newspan) (gethash (lower ls) ht))
      (push (upper newspan) (gethash (upper ls) ht))
      ;; Map dimension to new dimension, and create relations
      (push newdim (gethash olddim ht))
      ;; Create bounding relations on parameters
      (unless (eq newdim olddim)
	(if (or (eq lower-gap '?) (eq upper-gap '?))
	    (make-upper-bound olddim newdim)
	  (apply #'make-summation olddim 
		 (remove nil (list newdim d-lower d-upper)))))
      ;; Return complements
      (cons d-lower d-upper))))


(defmethod map-span-for-subspace ((ht hash-table) (ls lattice-span) names)
  (destructuring-bind (lower-gap span upper-gap) names
    (let ((newspans (refine-span ls (length (remove nil names))))
	  (keyspan nil))
      (cond ((and lower-gap upper-gap) (setf keyspan (second newspans)))
	    (lower-gap (setf keyspan (second newspans)))
	    (upper-gap (setf keyspan (first newspans)))
	    (t (setf keyspan (first newspans))))
      (push keyspan (gethash ls ht))
      (push (lower keyspan) (gethash (lower ls) ht))
      (push (upper keyspan) (gethash (upper ls) ht)))))
    

; ;; Given a hashtable and a specification for how this dimension will be
; ;; refined, populate the hashtable 
; ;; Return is a pair of (lower . upper) new gap dimensions
(defmethod add-to-lattice-subspace-map ((ht hash-table) (span lattice-span)
					other-dim-spans names unit)
  ;; map all the other dimensions
  (mapcar (ppf map-span-for-subspace ht ? names) (cons span other-dim-spans))
  ;; create and map the new dimension and the two gap dimensions
  (create-and-map-subspace-dimension ht span names unit))


;; Find all of the dimensions for whom a boundary of this cell is a base span
(defun gather-root-dimensions (cell)
  (let* ((1d (remove 1 (cons cell (boundaries* cell))
		     :test-not #'= :key #'dimension))
	 (root-spans 
	  (remove-if-not (ppf typep ?x 'coordinate-span)
			 (apply #'append (mapcar #'measurements 1d)))))
    (remove-duplicates (mapcar #'tail root-spans))))

(defun separate-root-span (dim span-coords)
  (let ((root-coord (find dim span-coords :key (cmp dimension lower))))
    (assert root-coord)
    (cons root-coord (remove root-coord span-coords))))

;; Returns a pair of root span, other spans
(defun gather-all-spans-along-dimension (cell dim)
  (flet ((check-if-same-dim (line)
	   (find dim (mapcar #'tail (span-values line)))))
    (let* ((1d (remove 1 (cons cell (boundaries* cell)) 
		       :test-not #'= :key #'dimension))
	   (samed (remove-if-not #'check-if-same-dim 1d))
	   (span-coords (remove-duplicates (mapcar #'value (apply #'append (mapcar #'span-values samed))))))
      (separate-root-span dim span-coords))))

(defun subspace-composite-parameters (ht dims)
  (let ((newdims (apply #'append (mapcar (ppf gethash ? ht) dims))))
    (mapcar (lambda (old new) 
	      (unless (eq old new) (ensure-parallel old new))
	      (push new (gethash old ht)))
	    (composite-parameters* dims)
	    (composite-parameters* newdims))))

(defun map-uniform-values (ht cell)
  (let* ((m-sets (mapcar (ppf links-of-type ? 'coordinate-value t)
			(cons cell (boundaries* cell))))
	 (uniform (fold (ppf intersection ?x ?y :test #'same-coordinate) 
			(cdr m-sets) (car m-sets))))
    (dolist (v uniform)
      (push (value v) (gethash (value v) ht)))))

(defun make-parameter-maps (cell name-maps)
  ;; Note: should check to ensure there's no bogus maps
  (let ((dims (gather-root-dimensions cell))
	(ht (make-hash-table))
	(cht (make-hash-table)))
    (dolist (d dims)
      (let* ((dimnames (or (cdr (find (name d) name-maps :key #'car))
			   (list nil (name d) nil))) ; missing = no difference
	     (spanpair (gather-all-spans-along-dimension cell d))
	     (cpl (add-to-lattice-subspace-map ht (car spanpair) (cdr spanpair)
					       dimnames (units d))))
	(setf (gethash (car spanpair) cht) cpl)))
    (subspace-composite-parameters ht dims)
    (map-uniform-values ht cell)
    (cons ht cht)))

(defun duplicate-or-share (parameter-map cell)
  (flet ((same-element (p) (find (value p) (gethash (value p) parameter-map))))
    (let* ((params (remove-if-not (ppf typep ?x 'coordinate-value)
				  (measurements cell)))
	   (new-cell (if (find-if-not #'same-element params)
			 (make-cell (dimension cell))
		       cell)))
      new-cell)))

(defun ensure-measurements (parameter-map cell daughter cell-map)
  ;; Note that daughter may be the same as cell; we still map measures
  (dolist (m (measurements cell))
    (cond ((typep m 'coordinate-span)
	   (let ((newdims (gethash (tail m) parameter-map)))
	     (assert (= (length newdims) 1))
	     (ensure-measure 'coordinate-span (car newdims) daughter)))
	  ((typep m 'uniform-coordinate)
	   (let ((newvs (gethash (value m) parameter-map)))
	     (dolist (newv newvs)
	       (ensure-coordinate-value (dimension newv) daughter newv))))
	  ((typep m 'graduated-coordinate)
	   (let ((newvs (gethash (value m) parameter-map)))
	     (dolist (newv newvs)
	       (warn "Should ensure graduated-coordinate ~A, but don't currently know how to" newv)
	       )))
	       ;;(let* ((bset (boundaries daughter))
		;;      (bound+ (find (second newv
		;;      )
		;; (ensure-coordinate-value (dimension newv) daughter (first newv)
		;;			  bound+ bound-)))))
	  (t (error "unknown measurement: ~A" m)))))

;; A cell can be determined to be a subspace if there are no relevant
;; different coordinate values
(defun is-subspace (parent child param-map)
  (let* ((m (measurements parent))
	 (base-dims (remove-duplicates 
		     (apply #'append 
			    (mapcar
			     (cmp find-elementary-parameters tail)
			     (remove-if-not (ppf typep ? 'coordinate-span) 
					    m)))))
	 (all-dims (append base-dims (composite-parameters* base-dims)))
	 (vals (remove-if-not (ppf typep ? 'coordinate-value) m)))
    (not
     (find-if
      (lambda (v) (and (not (find (tail v) all-dims)) 
		       (not (find (value v) (gethash (value v) param-map)))))
      vals))))

(defmethod subspace ((c cell) name-maps)
  ;; First, handle the manipulation of the dimensions and their lattice coords
  (destructuring-bind (param-map . gap-map) (make-parameter-maps c name-maps)
    ;; Next, create the parent/child associations
    (let* ((parents (cons c (boundaries* c)))
	   (children (mapcar (pcf duplicate-or-share param-map) parents))
	   (sets (mapcar #'list parents children)))
      (dolist (s sets)
	(let ((parent (first s)) (child (second s)))
	  (unless (eq parent child)
	    ;; Copy boundaries from parent to child
	    (dolist (b (boundaries parent))
	      (destructuring-bind (bparent bchild) (assoc b sets) 
		(make-boundary child bchild))))))
      ;; Finally, copy all the remaining relations
      (dolist (s sets)
	(let ((parent (first s)) (child (second s)))
	  (ensure-measurements param-map parent child sets)))
      (dolist (s sets)
	(let ((parent (first s)) (child (second s)))
	  (unless (eq parent child)
	    ;; Create subspace and parallel links
	    (when (> (dimension parent) 0)
	      (if (is-subspace parent child param-map)
		  (make-subspace parent child)
		(make-parallel parent child))))))
      (first children))))

; ;; What's still missing from subspace:
; ;; - linking of complementary parameters
; ;; - optional coordinate alignment based on an anchor
; ;; - applicability to sets ---> do it by anchor containment
; ;;   For each anchor, make one subspace per cell it is a boundary of
; ;; Next operations needed:
; ;; - construction of symmetric coordinate systems
; ;;   (or at least something that will allow "subspace-from X" like chair legs)
