;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;; Diagnostic routines (some of which are wrappers for
;; existing material)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; DESCRIPTION UTILITIES  ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod pretty-name ((s string) &optional (to-space "_-"))
  (substitute-if #\Space (lambda (c) (find c to-space)) (string-capitalize s)))
(defmethod pretty-name ((e element) &optional (to-space "_-")) 
  (pretty-name (name e) to-space))
(defmethod pretty-name ((s symbol) &optional (to-space "_-")) 
  (pretty-name (symbol-name s) to-space))

;;;; Significant figure printing
;; (which isn't done quite right by the existing utilities)
;; print with a certain number of significant figures and not too many zeros
(defun sigfig (value &optional (n 3) (zeros 3))
  (setf n (- n 1))			; kludge: fencepost errors adding 1 sigfig
  (if (> (abs value) 1)
      (let ((place (log (abs value) 10)))
	(cond ((< place n) (format nil (format nil "~~,~SF" (ceiling (- n place))) value))
	      ((< place (+ n zeros 1)) 
	       (let ((divisor (expt 10 (floor (- place n)))))
		 (format nil "~S" (* divisor (round value divisor)))))
	      (t (format nil (format nil "~~,~SE" n) value))))
    (if (zerop value)
	(format nil (format nil "~~,~SF" n) 0)
      (let ((place (- (log (abs value) 10))))
	(if (<= place zeros) 
	    (format nil (format nil "~~,~SF" (ceiling (+ n place))) value)
	  (format nil (format nil "~~,~SE" n) value))))))



;;;; Units
;; return (unit . exponent)
(defmethod units ((s symbol)) (symbol-name s))

(defun strip-unit-exponent (name)
  (let ((expts (count #\^ name)))
    (if (= expts 1)
	(cons (subseq name 0 (position #\^ name)) 
	      (parse-integer (subseq name (1+ (position #\^ name)))))
      (cons name 1))))

(defun compress-unit-terms (units)
  (let ((stripped (mapcar #'strip-unit-exponent 
			  (mapcar (ppf pretty-name ?x "_") (flatten units)))))
    (mapcar (lambda (uname)
	      (let* ((match (remove uname stripped :key #'car :test-not #'eqv))
		     (pow (apply #'+ (mapcar #'cdr match))))
		(if (zerop pow) ""
		  (format nil "~A~@[^~A~]" uname (when (> pow 1) pow)))))
	    (mapcar #'car (remove-duplicates stripped :test #'eqv :key #'car)))))

(defun unit-string (params &optional alt-params)
  (apply #'format nil "~@{~A~#[~; ~]~}" 
	 (compress-unit-terms (mapcar #'units (or params alt-params)))))


;;;; Products
;; find square, cube, etc. terms and compress them:
(defun compress-product-terms (params)
  (mapcar (lambda (p) 
	    (let ((pow (count p params)))
	      (format nil "~A~@[^~A~]" (pretty-name p) (when (> pow 1) pow))))
	  (remove-duplicates params)))
	  
(defun product-string (params)
  (assert (> (length params) 0) () "Can't describe a product of nothing")
  (apply #'format nil "~#[nothing~;~A~;the product of ~A and ~A~:;~
                       the product of ~@{~#[~;and ~]~A~^, ~}~]"
	 (compress-product-terms params)))

;;;; Sums
(defun compress-sum-terms (params)
  (mapcar (lambda (p) 
	    (let ((n (count p params)))
	      (format nil "~@[~A * ~]~A" (when (> n 1) n) (pretty-name p))))
	  (remove-duplicates params)))
	  
(defun sum-string (params)
  (assert (> (length params) 0) () "Can't describe a sum of nothing")
  (apply #'format nil "~#[nothing~;~A~;the sum of ~A and ~A~:;~
                       the sum of ~@{~#[~;and ~]~A~^, ~}~]"
	 (compress-sum-terms params)))

;;;; Sets
(defun param-set-string (params)
  (assert (> (length params) 0) () "Can't describe a set of nothing")
  (apply #'format nil "~#[nothing~;~A~;~A and ~A~:;~@{~#[~;and ~]~A~^, ~}~]"
	 (mapcar #'pretty-name params)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; EXPLAINING CONFLICTS ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defmethod explain-conflict :around ((r relation) &optional (stream t))
  (unless (zerop (stress r)) (call-next-method)))

(defmethod explain-conflict ((r relation) &optional (stream t))
  (format stream "Don't know how to explain conflict for ~A of type ~A~%" 
	  (pretty-name r) (type-of r)))

(defmethod explain-conflict ((r identity-relation) &optional (stream t))
  (with-slots (risers fallers constant constant-units) r
    (let* ((d-val (apply #'* (mapcar #'get-single-value risers)))
	   (n-val (* constant (apply #'* (mapcar #'get-single-value fallers))))
	   (dir (< d-val n-val)))
      (format stream "~A: ~A is too ~:[large~;small~] at ~A ~A,~%  ~
                      Should be equal to ~@[~A times~] ~A, which is ~A ~A~%"
	      (pretty-name r) (product-string risers) dir 
	      (sigfig d-val) (unit-string risers)
	      (unless (= 1 constant) (sigfig constant)) (product-string fallers)
	      (sigfig n-val) (unit-string (append constant-units fallers))))))

(defmethod explain-conflict ((r unit-identity) &optional (stream t))
  (with-slots (composite elementary) r
    (let* ((c-val (get-single-value composite))
	   (e-val (apply #'* (mapcar #'get-single-value elementary)))
	   (dir (< c-val e-val)))
      (format stream "~A: ~A is too ~:[large~;small~] at ~A ~A,~%  ~
                      Should be equal to ~A, which is ~A ~A~%"
	      (pretty-name r) (pretty-name composite) dir 
	      (sigfig c-val) (unit-string (list composite))
	      (product-string elementary)
	      (sigfig e-val) (unit-string elementary)))))


(defmethod explain-conflict ((r summation) &optional (stream t))
  (with-slots (total summands) r
    (let* ((t-val (get-single-value total) )
	   (s-val (apply #'+ (mapcar #'get-single-value summands)))
	   (dir (< t-val s-val)))
      (format stream "~A: ~A is too ~:[large~;small~] at ~A ~A,~%  ~
                      Should be equal to ~A, which is ~A ~A~%"
	      (pretty-name r) (pretty-name total) dir 
	      (sigfig t-val) (unit-string (list total))
	      (sum-string summands) 
	      (sigfig s-val) (unit-string (list (first summands)))))))

(defun const-sum-string (params constants)
  (assert (> (length constants) 0) () "Can't describe a sum of nothing")
  (let* ((nonzero (remove 0 (mapcar #'cons params constants) :key #'cdr))
	 (const-sum-terms 
	  (mapcar (lambda (pc) 
		    (format nil "~@[~A ~]~A" 
			    (unless (= (cdr pc) 1) (sigfig (cdr pc))) 
			    (pretty-name (car pc))))
		  nonzero)))
    (if (not params) 
	(format nil "a constant")
      (apply #'format nil "~#[nothing~;~A~;the sum of ~A and ~A~:;~
                       the sum of ~@{~#[~;and ~]~A~^, ~}~]~@[ plus a constant~]"
	     const-sum-terms (> (length constants) (length params))))))

(defmethod explain-conflict ((r composite-sum-bound) &optional (stream t))
  (with-slots (risers fallers constants) r
    (let* ((bound-v (append (mapcar #'get-single-value risers) '(1)))
	   (bounded-v (append (mapcar #'get-single-value fallers) '(1)))
	   (h-val (apply #'+ (mapcar #'* bound-v (second constants))))
	   (l-val (apply #'+ (mapcar #'* bounded-v (first constants)))))
      (format stream "~A: ~A is too large, at ~A ~A,~%  ~
                      Should be significantly smaller than ~A, which is ~A ~A~%"
	      (pretty-name r) (const-sum-string fallers (first constants))
	      (sigfig l-val) (unit-string fallers risers)
	      (const-sum-string risers (second constants))
	      (sigfig h-val) (unit-string risers fallers)))))

(defmethod explain-conflict ((r modular-relation) &optional (stream t))
  (with-slots (risers fallers) r
    (format stream "~A: ~A are too large, relative to ~A~%"
	    (pretty-name r) 
	    (param-set-string fallers) (param-set-string risers))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; EXPLAINING SUGGESTIONS ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defmethod explain-suggestion ((p parameter) &optional (stream t))
  (format stream 
	  "  ~:[Increase~;Decrease~] ~A from its current value of ~A ~A~%"
	  (minusp (stress p)) (pretty-name p) (sigfig (get-single-value p))
	  (unit-string (list p))))

(defvar change-epsilon 0.001)  ;; don't mention any change less than 0.1%
(defun get-parameter-change (p)
  (let ((old (mark? p 'original-value)) (new (get-single-value p)))
    (assert (and old new))
    (list p old new (if (> new old) (/ new old) (/ old new)))))
(defun composite-parameter? (p)
  (find p (remove-if-not (ppf typep ?x 'unit-identity) (all-relations))
	:key #'composite-unit))
(defmethod explain-solution (&optional (compressed nil) (stream t))
  (let* ((changesets (mapcar #'get-parameter-change (all-parameters)))
	 (changed (remove-if (ppf < ?x (+ 1 change-epsilon)) changesets :key #'fourth)))
    (format stream "Found a possible solution:~%")
    (dolist (pset (sort changed #'> :key #'fourth))
      (destructuring-bind (p old new ratio) pset
	(unless (and compressed (composite-parameter? p))
	  (format stream "  ~:[Increase~;Decrease~] ~A from ~A to ~A ~A~%"
		  (> old new) (pretty-name p) (sigfig old) (sigfig new) 
		  (unit-string (list p))))
	))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; MANAGING PRIORITIES ;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun marked-values (pvs)
  (parameter-values pvs)
  (dolist (pv pvs) 
    (let ((p (find-parameter (first pv))))
      (mark p 'priority 0)
      (mark p 'specification))))
(defun prioritize-stressed ()
  (dolist (e (all-elements)) 
    (when (and (or (typep e 'parameter) (typep e 'relation)) 
	       (not (zerop (stress e))))
      (mark e 'priority 0))))
(defun prioritize-random-stressed ()
  (let ((stressed (remove-if #'zerop (append (all-parameters) (all-relations)) 
			     :key #'stress)))
    (when stressed
      (mark (choose-random stressed #'abs-stress) 'priority 0))))

;; Method #1: prioritize two-way pulls
; (defun shift-resolution-focus (&optional (spec-only nil))
;   (clear-mark-class 'priority)
;   (let ((specs (remove-if-not (ppf mark? ?x 'specification) (all-parameters)))
; 	(conflicts (remove-if #'stress-unopposed? (all-parameters))))
;     (dolist (p (append specs (if spec-only nil conflicts)))
;       (mark p 'priority 0))))
;; Method #2: prioritize randomly
; (defun shift-resolution-focus (&optional (spec-only nil))
;   (clear-mark-class 'priority)
;   (let ((specs (remove-if-not (ppf mark? ?x 'specification) (all-parameters))))
;     (dolist (p specs) (mark p 'priority 0))
;     (unless spec-only (prioritize-random-stressed))))
;; Method #3: prioritize *only* randomly
(defun shift-resolution-focus (&optional (spec-only nil))
  (clear-mark-class 'priority)
  (unless spec-only (prioritize-random-stressed)))


(defun record-original-values ()
  (dolist (p (all-parameters))
    (let ((v (get-values p)))
      (when (= (length v) 1)
	(mark p 'original-value (value (first v)))))))


;; return stressed relations, sorted by descending stress
(defun current-conflicts (&optional (include-warnings nil))
  (sort (find-stressed-relations (not include-warnings))
	#'> :key (cmp abs stress)))

(defun diagnose (&optional (include-warnings nil))
  (startup-propagate)
  (dolist (r (all-relations)) (resolve r))
  (setf current-max-stress (max-net-stress) current-max-abs-stress (max-stress))
  (current-conflicts include-warnings))

;; Note: this will ignore "fixed" values (e.g., priority 0 parameters)
(defun find-nonconflicting-deltas (stressed-relations)
  (let* ((param-sets (mapcar #'parameters stressed-relations))
	 (params (remove-duplicates (apply #'append param-sets))))
    (remove-if-not 
     (lambda (p) (and (not (zerop (stress p))) (stress-unopposed? p)))
     params)))

(defun human-readable-conflict-report (problems &optional (stream t))
  (explain-current-priorities)
  (if (not problems)
      (format stream "No problems detected in current design.~%")
    (let ((suggestions (find-nonconflicting-deltas problems)))
      (format stream "~%Detected ~A problems in current design ~
                      (in descending order of importance):~%"
	      (length problems))
      (dolist (relation problems)
	(explain-conflict relation stream))
      (when suggestions
	(format stream "~%Suggestions to fix design problems:~%")
	(dolist (parameter suggestions)
	  (explain-suggestion parameter stream))))))

(defun explain-current-priorities (&optional (stream t))
  (let* ((pr (append (all-parameters) (all-relations)))
	 (pri (remove-if-not (lambda (p) (and p (zerop p))) pr 
			     :key (ppf mark? ?x 'priority)))
	 (specs (remove-if-not (ppf mark? ?x 'specification) pr)))
    (format stream "Assuming fixed value for ~A~%" (param-set-string specs))
    (if pri
	(format stream "Focusing resolution on ~A~%" (param-set-string pri))
      (format stream "No value being prioritized~%"))))

(defparameter min-freq 0.02) ;; is there a low-high crossing every 1/min-freq or less?
(defparameter osc-margin 5)
(defparameter min-step-size 1e-5)
(defparameter autoadjust t)
(defun mean (list) (/ (fold #'+ list 0) (length list)))
(defun diagnose-oscillation (total-stress-chart)
  (let* ((smax (apply #'max total-stress-chart))
	 (smin (apply #'min total-stress-chart))
	 (smean (mean total-stress-chart))
	 (margin (min 0.2 (* osc-margin step-size)))
	 (highthresh (* (+ 1 margin) smean))
	 (lowthresh (* (- 1 margin) smean)))
    (when (and autoadjust (> smin 0) 
	       (or (> (/ smax smin) 1.1) 
		   (> (- smax smin) (* 5 step-size))))
      ;; start in high state, no crossings
      (let ((crossings 0) (state t) (all-down t) (slast smax))
	(dolist (snext total-stress-chart)
	  (cond ((and (< snext lowthresh) state)
		 (setf state nil))
		((and (> snext highthresh) (not state))
		 (setf state t) (incf crossings)))
	  (when (> snext slast) (setf all-down nil))
	  (setf slast snext)
	  )
	(format t "Found ~A crossings in ~A steps: freq = ~A~%"
		crossings (length total-stress-chart) 
		(sigfig (/ crossings (length total-stress-chart))))
	;(pprint total-stress-chart)
	(cond ((> (/ crossings (length total-stress-chart)) min-freq)
	       (when (> step-size min-step-size)
		 (setf step-size (* step-size 0.5))
		 (format t "Oscillation detected: reduced step size to ~A~%" 
			 (sigfig step-size))))
	      ((and (= crossings 0) (> (first total-stress-chart) (car (last total-stress-chart))) (< step-size 0.1))
	       (setf step-size (* step-size 2))
	       (format t "Fast smooth descent: increased step size to ~A~%"
		       (sigfig step-size))))
	))))

;; assumes diagnosis has already been run, producing non-zero stress
(defparameter batch-size 150)
(defparameter batch-diff 1) ; threshold = diffsize * step-size
(defun run-batches-until-stuck (&optional max-sets)
  (let ((old-stress (total-stress)))
    (unless (zerop old-stress)
      (let* ((chart (make-stress-graph batch-size))
	     (new-stress (total-stress))
	     (diff (min (/ (abs (- new-stress old-stress)) old-stress)
			(abs (- new-stress old-stress)))))
	(diagnose-oscillation chart)
	(let ((rest (when (and (> diff (* batch-diff step-size)) 
			       (or (not max-sets) (plusp max-sets)))
		      (run-batches-until-stuck (when max-sets (1- max-sets))))))
	  (merge-charts (cons (list (length chart)) chart) rest))))))

(defun merge-charts (chart1 chart2)
  (cons (append (car chart1) (car chart2))
	(append (cdr chart1) (cdr chart2))))

(defun run-with-shifting-priorities (&optional (max-shifts 100))
  (unless (or (zerop (total-stress)) (not (plusp max-shifts)))
    (shift-resolution-focus)
    (explain-current-priorities)
    (let* ((chart (run-batches-until-stuck 20))
	   (rest (run-with-shifting-priorities (1- max-shifts))))
      (merge-charts chart rest))))

(defun diagnose-and-repair (&optional (stream t))
  (let ((problems (diagnose t)))
    (human-readable-conflict-report problems stream)
    (when problems
      (format stream "~%Attempting to find solution to conflicts.~%")
      (record-original-values)
      (let ((chart (run-with-shifting-priorities))
	    (newproblems (diagnose t)))
	(if newproblems
	    (progn (format stream "~%Could not find solution.~%"))
	  (explain-solution t stream))
	chart))))

(defun list-to-matlab-variable (variable list &optional (stream t))
  (format stream "~A = [" variable)
  (dotimes (i (length list))
    (format stream "~A " (sigfig (pop list)))
    (when (= (mod (1+ i) 15) 0) (format stream " ...~%  ")))
  (format stream "];~%"))

(defun output-diagnose-and-repair-data
    (chart prefix &optional (filename "stressdata") header)
  (with-open-file (out filename :direction :output :if-exists :new-version)
    (when header (format out header))
    (list-to-matlab-variable (format nil "~Astress" prefix) (cdr chart) out)
    (list-to-matlab-variable "rawshifts" (car chart) out)
    (format out "~Ashifts = cumsum(rawshifts);~%" prefix)))

(defun output-diagnose-and-repair-graph
    (chart prefix &optional (filename "stressplot") (directory "./") header)
  (let ((plotname (format nil "~A~A.m" directory filename))
	(dataname (format nil "~A~A_data.m" directory filename)))
    (with-open-file (out plotname :direction :output :if-exists :new-version)
      (when header (format out header))
      (output-diagnose-and-repair-data chart prefix dataname header)
      (let ((v_st (format nil "~Astress" prefix))
	    (v_sh (format nil "~Ashifts" prefix)))
	(format out "~A_data;~%" filename)   ; load data
	(format out "h = figure;~%")
	(format out "plot(1:numel(~A),~A,'b-'); hold on;~%" v_st v_st)
	(format out
		"for i=1:numel(~A), plot([~A(i) ~A(i)],[0 max(~A)],'r:'); end~%"
		v_sh v_sh v_sh v_st)
	(format out "xlabel('Iteration'); ylabel('Total absolute stress');~%")
	)))
  'done)

(defun output-current-design-as-vector 
    (&optional (filename "designvector.m") (variable "params") header)
  (with-open-file (out filename :direction :output :if-exists :new-version)
    (when header (format out header))
    (format out "%~{ ~A~}~%" (mapcar #'first (all-parameter-values)))
    (let ((sorted-values (mapcar #'second (all-parameter-values))))
      (list-to-matlab-variable variable sorted-values out)))
  'done)
