;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;; Design Parameters
(clear-topology)
(setf design-name "minidroid")

;; Densities in kg/m^3
(defun g/cc->kg/m^3 (n) (* n 1000))	; water is 1 g/cc, 1000 kg/m^3
(defun in->m (n) (/ (* n 2.54) 100))
(defparameter abs-density (g/cc->kg/m^3 1.04))
(defparameter steel-density (g/cc->kg/m^3 7.75))
(defparameter rubber-density (g/cc->kg/m^3 1.1))
;; battery specs extrapolated from Nokia BL-5C
;; Ref: http://en.wikipedia.org/wiki/Energy_density
(defparameter li-battery-density (g/cc->kg/m^3 1.74))
(defparameter li-battery-energy 7e5)	; Joules/kg
(defparameter g0 9.80665)		; standard gravity
(defun kg-cm->N-m (torque) (* g0 0.01 torque))

;; Core bot
(extrude-box 'bot_body_length_x 'bot_body_length_y 'bot_body_length_z)
(ensure-parameter 'bot_body_mass 'kilograms)
;; mass/volume relation more complicated - defined below

(ensure-parameter 'case_rounding_diameter 'meters) ; diameter of circle
(ensure-parameter 'case_rounding_surface_loss 'meters^2)
;; Square = 4D, Circle = pi D: loss area = difference*bot_body_length_y
(name-it "mass/volume"
 (make-identity-relation	; mass/volume relation
  `(,(- pi 4))
  '(case_rounding_diameter bot_body_length_y)
  '(case_rounding_surface_loss)))
(make-upper-bound 'bot_body_length_z 'case_rounding_diameter)
(make-upper-bound 'bot_body_length_x 'case_rounding_diameter)
 

(ensure-parameter 'structure_mass 'kilograms)
(ensure-parameter 'structure_thickness 'meters)
(ensure-parameter 'structure_surface 'meters^2)
;; surface defined from body minus case-rounding
(name-it "surface area"
 (make-summation				; surface is cover of body
  'structure_surface
  'bot_body_length_xy 'bot_body_length_xy
  'bot_body_length_xz 'bot_body_length_xz
  'bot_body_length_yz 'bot_body_length_yz
  'case_rounding_surface_loss))
(name-it "mass/volume"
 (make-identity-relation	; mass/volume relation
  `(,(* 1.1 abs-density) kilograms meters^-3) ; surface + 10% mounting volume
  '(structure_surface structure_thickness)
  '(structure_mass)))

;; Flipper
(extrude-box 'flipper_length_x 'flipper_length_y 'flipper_length_z)
(ensure-parameter 'flipper_mass 'kilograms)
(ensure-parameter 'flipper_rotational_speed 'radians/second)
(name-it "mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,abs-density kilograms meters^-3)
  '(flipper_length_xyz)
  '(flipper_mass)))

(ensure-parameter 'flipper_axle_y 'meters)
(ensure-parameter 'flipper_axle_diameter 'meters)
(ensure-parameter 'flipper_axle_mass 'kilograms)
(ensure-parameter 'flipper_axle_torque_limit 'Newton-Meters)
(ensure-parameter 'flipper_axle_gear_diameter 'meters) ; ignoring its mass
(name-it "mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,steel-density kilograms meters^-3)
  '(flipper_axle_y flipper_axle_diameter flipper_axle_diameter)
  '(flipper_axle_mass)))

(extrude-box 'flipper_motor_x 'flipper_motor_y 'flipper_motor_z)
(ensure-parameter 'flipper_motor_mass 'kilograms)
(ensure-parameter 'flipper_motor_torque 'Newton-Meters)
(ensure-parameter 'flipper_motor_power 'amps)
(ensure-parameter 'flipper_motor_max_speed 'radians/second)
(ensure-parameter 'flipper_motor_gear_multiplier 'number)

(extrude-box 'flipper_encoder_x 'flipper_encoder_y 'flipper_encoder_z)
(ensure-parameter 'flipper_encoder_mass 'kilograms)
(ensure-parameter 'flipper_encoder_power 'amps)

;; Wheels
(ensure-parameter 'wheel_y 'meters)
(ensure-parameter 'wheel_diameter 'meters)
(ensure-parameter 'wheel_mass 'kilograms)
(ensure-parameter 'wheel_separation 'meters) ; distance from wheel to wheel
(ensure-parameter 'wheel_clearance 'meters)
(name-it "mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,abs-density kilograms meters^-3)
  '(wheel_y wheel_diameter wheel_diameter)
  '(wheel_mass)))

(ensure-parameter 'track_thickness 'meters)
(ensure-parameter 'track_mass 'kilograms)
(ensure-parameter 'track_length 'meters)
(make-summation 'track_length		; square approximation of length
		'wheel_diameter 'wheel_diameter 'wheel_diameter
		'wheel_diameter 'wheel_diameter 'wheel_diameter
		'wheel_separation 'wheel_separation)
(name-it "mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,abs-density  kilograms meters^-3)
  '(wheel_y track_length track_thickness)
  '(track_mass)))


(extrude-box 'wheel_motor_x 'wheel_motor_y 'wheel_motor_z)
(ensure-parameter 'wheel_motor_mass 'kilograms)
(ensure-parameter 'wheel_motor_torque 'Newton-Meters)
(ensure-parameter 'wheel_motor_power 'amps)
(ensure-parameter 'wheel_motor_max_speed 'radians/second)

;; Electronics
(extrude-box 'battery_x 'battery_y 'battery_z)
(ensure-parameter 'battery_mass 'kilograms)
(ensure-parameter 'battery_energy 'amp-hours)
(name-it "mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,(* 2 li-battery-density) kilograms meters^-3) ; 2 battery packs
  '(battery_xyz)
  '(battery_mass)))
(name-it "energy-density"
 (make-identity-relation
  `(,(* li-battery-energy 0.8) amp-hours kilograms^-1) ; assume 80% efficient use of space
  '(battery_mass) '(battery_energy)))

(extrude-box 'CPU_x 'CPU_y 'CPU_z)
(ensure-parameter 'CPU_mass 'kilograms)
(ensure-parameter 'CPU_power 'amps)
;; CPU vastly overspecc'ed, so computational properties not included

(extrude-box 'proximity_sensor_x 'proximity_sensor_y 'proximity_sensor_z)
(ensure-parameter 'proximity_sensor_mass 'kilograms)
(ensure-parameter 'proximity_sensor_power 'amps)
(ensure-parameter 'proximity_sensor_range 'meters) ; unused

(extrude-box 'power_supply_board_x 'power_supply_board_y 'power_supply_board_z)
(ensure-parameter 'power_supply_board_mass 'kilograms)

(extrude-box 'on_off_button_x 'on_off_button_y 'on_off_button_z)
(ensure-parameter 'on_off_button_mass 'kilograms)

;; Finally: total body mass is the sum of all component masses
(name-it "total-mass"
 (make-summation 'bot_body_mass
		 'structure_mass 
		 'flipper_mass 'flipper_mass ;; x2
		 'flipper_axle_mass 'flipper_motor_mass 'flipper_encoder_mass
		 'wheel_mass 'wheel_mass 'wheel_mass 'wheel_mass ; x4
		 'track_mass 'track_mass ; x2
		 'wheel_motor_mass 'wheel_motor_mass
		 'battery_mass 'CPU_mass 'proximity_sensor_mass
		 'power_supply_board_mass 'on_off_button_mass))



;;;; Functional blueprints:

;; Quantized components:
(defparameter servo-library '()) ; kludged for now
(defparameter encoder-library '(11e-3 20e-3 11e-3 10e-3 10e-3))
(defparameter CPU-library '(68.6e-3 53.3e-3 15e-3 10e-3 40e-3))
(defparameter proximity-sensor-library '(14e-3 30e-3 13e-3 40e-2 33e-3 3.6e-3))
(defparameter dc-converter-library '(28e-3 28e-3 5e-3 10e-3))
(defparameter button-library '(18e-3 18e-3 7.62e-3 5e-3))

; (make-quantized-component 
;  servo-library 
;  'flipper_motor_x 'flipper_motor_y 'flipper_motor_z 'flipper_motor_mass 
;  'flipper_motor_torque 'flipper_motor_power 'flipper_motor_max_speed)

; (make-quantized-component 
;  servo-library 
;  'wheel_motor_x 'wheel_motor_y 'wheel_motor_z 'wheel_motor_mass 
;  'wheel_motor_torque 'wheel_motor_power 'wheel_motor_max_speed)

;; motor kludge: make everything proportional to torque except speed,
;; which is constant
;; Might also want to switch torque to proportional to volume, not lengths
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ (* 39.88e-3 36.83e-3 19.81e-3) 1.216))
  '(flipper_motor_torque) '(flipper_motor_xyz)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ 61.8e-3 1.216)) '(flipper_motor_torque) '(flipper_motor_mass)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ 1.5 1.216)) '(flipper_motor_torque) '(flipper_motor_power)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ (/ (/ pi 3) 0.13))) '(flipper_motor_max_speed) '()))
;; maximum motor size available, based on hitec servo data
(name-it "quantized bound: flipper motor torque max"
 (make-sum-bound-relation
  `((1) (,(kg-cm->N-m 38)))		; bound of ~3.7
  '(flipper_motor_torque)
  '()))
(name-it "component: flipper motor"
  (bind-into-component
   ;; bounding relations
   '("quantized bound: flipper motor torque max")
   ;; scalable dimensions: implicitly includes all derived units
   '(flipper_motor_x flipper_motor_y flipper_motor_z) 
   ;; cumulative parameters
   '(flipper_motor_mass flipper_motor_power flipper_motor_torque)
   ;; non-cumulative parameters
   '(flipper_motor_max_speed)))

(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ (* 39.88e-3 36.83e-3 19.81e-3) 1.216))
  '(wheel_motor_torque) '(wheel_motor_xyz)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ 61.8e-3 1.216)) '(wheel_motor_torque) '(wheel_motor_mass)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ 1.5 1.216)) '(wheel_motor_torque) '(wheel_motor_power)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ (/ (/ pi 3) 0.13))) '(wheel_motor_max_speed) '()))
(name-it "quantized bound: wheel motor torque max"
 (make-sum-bound-relation
  `((1) (,(kg-cm->N-m 38)))		; bound of ~3.7
  '(wheel_motor_torque)
  '()))
(name-it "component: wheel motor"
  (bind-into-component
   ;; bounding relations
   '("quantized bound: wheel motor torque max")
   ;; scalable dimensions: implicitly includes all derived units
   '(wheel_motor_x wheel_motor_y wheel_motor_z) 
   ;; cumulative parameters
   '(wheel_motor_mass wheel_motor_power wheel_motor_torque)
   ;; non-cumulative parameters
   '(wheel_motor_max_speed)))

(name-it "quantized: encoder"
 (make-quantized-component 
  encoder-library 'flipper_encoder_x 'flipper_encoder_y 'flipper_encoder_z
  'flipper_encoder_power 'flipper_encoder_mass))

(name-it "quantized: CPU"
 (make-quantized-component 
  CPU-library 'CPU_x 'CPU_y 'CPU_z 'CPU_power 'CPU_mass))

(name-it "quantized: range-sensor"
 (make-quantized-component 
  proximity-sensor-library 
  'proximity_sensor_x 'proximity_sensor_y 'proximity_sensor_z
  'proximity_sensor_range 'proximity_sensor_power 'proximity_sensor_mass))

(name-it "quantized: DC converter"
 (make-quantized-component
  dc-converter-library
  'power_supply_board_x 'power_supply_board_y 'power_supply_board_z 
  'power_supply_board_mass))

(name-it "quantized: microswitch"
 (make-quantized-component
  button-library
  'on_off_button_x 'on_off_button_y 'on_off_button_z 'on_off_button_mass))


;; Closed-form relations:
;; Flipper rotatonal speed
;; flipper_motor / multiplier = flipper_speed
(name-it "rotational-speed"
 (make-identity-relation 
  `(1)
  '(flipper_motor_max_speed)
  '(flipper_rotational_speed flipper_motor_gear_multiplier)))
;; flipper gear effect on torque
;; torque-mult = gear-diameter / k(motor gear), where motor-diam = 0.5in
(name-it "gearing"
 (make-identity-relation
  `(,(/ (in->m 0.5)))
  '(flipper_axle_gear_diameter) '(flipper_motor_gear_multiplier)))


;; Sufficient axle diameter
;; [ref: http://en.wikipedia.org/wiki/Torsion_constant]
;; Radians = Torque*Length/JG
;; J = pi D^4 / 32 
;; G = 79.3 GPa [modulus of rigidity for steel]
;; Starting assumption: use a safety margin of a 0.01-radian twist
(name-it "axle-strength"
 (make-identity-relation
  `(,(/ 1 (* 0.01 (/ pi 32) 79.3e9)) Newton-Meters meters^-3)
  '(flipper_axle_torque_limit flipper_axle_y)
  '(flipper_axle_diameter flipper_axle_diameter flipper_axle_diameter
    flipper_axle_diameter)))
(make-upper-bound 'flipper_axle_torque_limit 'flipper_motor_torque)

;; Sane treads
(name-it "tread-proportion"
 (make-identity-relation 
  `(1.03) '(wheel_diameter) '(wheel_y)
  1.1 1.5))
(name-it "tread-proportion"
 (make-identity-relation 
  `(0.059) '(wheel_diameter) '(track_thickness)
  1.1 1.5))

;; sane flipper: base off root proportions
(name-it "flipper-proportion"
 (make-identity-relation 
  `(0.086) '(flipper_length_x) '(flipper_length_y)
  1.1 1.5))
(name-it "flipper-proportion"
 (make-identity-relation 
  `(0.18) '(flipper_length_x) '(flipper_length_z)
  1.1 1.5))

;; Endurance:
;; battery >= (CPU+prox)*(wait+drive) + (2x motor + k flipper)*drive
(defparameter wait-time (* 4 3600))	; 4 hours
(defparameter drive-time (* 0.5 3600))	; 1 hour, full speed
(defparameter pct-flipper 0.1)		; 10% flipper use
(defparameter voltage 6)		; 6V for both electronics and motors
(name-it "endurance"
 (make-sum-bound-relation
  `((,(* (+ wait-time drive-time) voltage)
     ,(* (+ wait-time drive-time) voltage)
     ,(* (+ wait-time drive-time) voltage)
     ,(* 2 drive-time voltage)
     ,(* 0.1 drive-time voltage))
    (1))
  '(CPU_power flipper_encoder_power proximity_sensor_power
    wheel_motor_power flipper_motor_power)
  '(battery_energy)
  -0.05 1.2))
; -3e3 3e3)				; +/- 3 kJ margin
;; assume: wheel_motor_draw = wheel_motor_power
(mark (find-relation "endurance") 'simulation)


;; Motor sufficiency
;(make-placeholder-fb
; "drive straight simulation; tested by driving time"
; 'wheel_motor_torque 'wheel_diameter 'wheel_motor_max_speed 'bot_body_mass)
;; Source: ftp://ftp.rta.nato.int/pubfulltext/RTO/MP/RTO-MP-014/$MP-014-01.pdf
;;    requirement 0.9 force/weight per side (for hard maneuvers)
;; so:  wheel_torque/(1/2 wheel_diam) >= 0.9 g body_mass
;;  Stressed tolerance is up to 0.6 weight (0.3 per side)
;; kludge equation: 0.9 g body_mass 0.5 wheel_diam < wheel_torque
(defun driving-eval (relation vars)
  (declare (ignore relation))
  (let ((torque (first vars)) (diam (second vars)) (mass (third vars)))
    (/ (* 0.45 g0 mass diam) torque)))

(name-it "flat-driving"
 (make-modular-relation
  '((wheel_motor_torque wheel_diameter bot_body_mass wheel_motor_max_speed))
  '(wheel_motor_torque wheel_diameter)
  '(bot_body_mass)
  (let ((plsf (positive-linear-stress-fn 0.9 3.0)))
    (eval-stress #'driving-eval plsf))))
(mark (find-relation "flat-driving") 'simulation)

;(make-placeholder-fb
; "drive up incline simulation; tested by driving time"
; 'wheel_motor_torque 'wheel_diameter 'wheel_motor_max_speed 'bot_body_mass)

;; Step climbing
(ensure-parameter 'step_height 'meters)
;(make-placeholder-fb
; "step climbing simulation"
; 'flipper_length_x 'bot_body_length_x 'step_height)
;; Angles stress range is 0.1 to 0.5 radians
;; kludge equation: asin(step_height/[body_length+flipper_length]) < angle
(defun step-climbing-eval (relation vars)
  (declare (ignore relation))
  (let ((flipper_x (first vars)) (body_x (second vars)) (step (third vars)))
    (let ((tmp (asin (/ step (+ body_x flipper_x)))))
      (if (realp tmp) tmp 1))))

(name-it "step-climbing"
 (make-modular-relation
  '((flipper_length_x bot_body_length_x step_height))
  '(flipper_length_x bot_body_length_x)
  '(step_height)
  (let ((plsf (positive-linear-stress-fn 0.35 0.5)))
    (eval-stress #'step-climbing-eval plsf))))
(mark (find-relation "step-climbing") 'simulation)

;; flipper must be under center of gravity, so at least 60% of bot length
;; 0.6 length <= flipper_x
(name-it "flipper-leverage"
 (make-sum-bound-relation 
  '((0.6) (1))
  '(bot_body_length_x)
  '(flipper_length_x)))
(mark (find-relation "flipper-leverage") 'simulation)

;; Flip self over
;(make-placeholder-fb
; "flip-self simulation"
; 'flipper_rotational_speed 'bot_body_mass)
;; kludge equation: 0.5*body_x*g*bot_mass < flipper_torque

(defun flipping-eval (relation vars)
  (destructuring-bind (torque body_x mass multiplier speed) vars
    (declare (ignore relation speed))
    (/ (* 0.5 body_x g0 mass) (* multiplier torque))))

(name-it "self-righting"
 (make-modular-relation
  '((flipper_motor_torque bot_body_length_x bot_body_mass 
     flipper_motor_gear_multiplier flipper_rotational_speed))
  '(flipper_motor_torque)
  '(bot_body_length_x bot_body_mass)
  (let ((plsf (positive-linear-stress-fn 0.7 1.0)))
    (eval-stress #'flipping-eval plsf))))
(mark (find-relation "self-righting") 'simulation)



;; sidebays between wheels:
;; battery_x + 2*structure-thick < wheel_sep
(name-it "battery-in-sidebay"
 (make-sum-bound-relation
  '((1 2) (1))
  '(battery_x structure_thickness)
  '(wheel_separation)))

;; Structural assumptions:
;; all major axis lengths < 1.5 meter
(name-it "material-limit"
 (make-sum-bound-relation '((1) (1.0)) '(bot_body_length_x) '() 0 0.05))
(name-it "material-limit"
 (make-sum-bound-relation '((1 2) (1.0)) '(bot_body_length_y wheel_y) '() 0 0.05))
(name-it "material-limit"
 (make-sum-bound-relation '((1) (1.0)) '(bot_body_length_z) '() 0 0.05))
;; mass < 100 kg
(name-it "mass-limit"
 (make-sum-bound-relation
  '((1) (100))
  '(bot_body_mass) '()))
; -1 1)
;; structure thickness at least 2 mm
(name-it "minimal-structure"
 (make-sum-bound-relation 
  '((2e-3) (1)) '() '(structure_thickness)))
;; wheel clearance at least 0.2 in
(name-it "minimal-clearance"
 (make-sum-bound-relation 
  `((,(in->m 0.2)) (1)) '() '(wheel_clearance))) ; -0.001 0.001)

;; turning ratio (tracked vehicle ideal)
;; proportional to around 1.5 X vs. 1 Y
(name-it "turning-ratio"
 (make-identity-relation
  `(,(/ 1 1.5)) '(bot_body_length_x) '(bot_body_length_y)
  1.5 2))
 

;; Component layout:
;; (including ground clearance FBs)
;; X:
;; body_x + 2*clearance < wheel_sep + 2*wheel_diameter
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1 2))
  '(bot_body_length_x wheel_clearance)
  '(wheel_separation wheel_diameter)))
;; wheels + wheelsep < body_x - case_rounding [wheels attach to body]
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 1) (1))
  '(wheel_separation wheel_diameter case_rounding_diameter)
  '(bot_body_length_x)))
;; flipper_length_x < body_x
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1) (1))
  '(flipper_length_x) 
  '(bot_body_length_x)))
;; CPU_x + encoder_x + wheel_motor_x + 4*structure_thickness < body_x
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 1 4) (1))
  '(CPU_x flipper_encoder_x wheel_motor_x structure_thickness)
  '(bot_body_length_x)))
;; CPU_x + encoder_x + range_sensor_x + 4*structure_thickness < body_x
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 1 4) (1))
  '(CPU_x flipper_encoder_x proximity_sensor_x structure_thickness)
  '(bot_body_length_x)))
;; case_rounding + wheel_motor_x + power_board_x + flipper_motor_x + flipper_axle/2 + flipper_gear_diameter/2 + 4 thickness < body_x
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 1 1 0.5 0.5 4) (1))
  '(case_rounding_diameter wheel_motor_x power_supply_board_x flipper_motor_x 
    flipper_axle_diameter flipper_axle_gear_diameter structure_thickness)
  '(bot_body_length_x)))
;; case_rounding/2 < flipper_gear_diameter/2 + thickness
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((0.5) (0.5 1))
  '(case_rounding_diameter)
  '(flipper_axle_gear_diameter structure_thickness)))

;; Y:
;; CPU_y + power_board_y + 2*battery + 5*structure_thick < body_y + 2*wheel_y
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 2 5) (1 2))
  '(CPU_y power_supply_board_y battery_y structure_thickness)
  '(bot_body_length_y wheel_y)))
;; CPU_y + flipper_motor_y + 2*battery + 5*structure_thick < body_y + 2*wheel_y
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 2 5) (1 2))
  '(CPU_y flipper_motor_y battery_y structure_thickness)
  '(bot_body_length_y wheel_y)))
;; flipper_axle_y = body_y + 2*wheel_y
(name-it "layout-geometry"
 (make-summation 'flipper_axle_y 'bot_body_length_y 'wheel_y 'wheel_y))
;; 2*flipper_motor_y + range_sensor_y + 2*structure < body_y
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((2 1 2) (1))
  '(flipper_motor_y proximity_sensor_y structure_thickness)
  '(bot_body_length_y)))
;; flipper_motor_y + CPU_y < body_y
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1) (1))
  '(flipper_motor_y CPU_y)
  '(bot_body_length_y)))

;; Z:
;; CPU_z < body/2 + flipper_axle_diam/2 [port out back]
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1) (0.5 0.5))
  '(CPU_z) 
  '(bot_body_length_z flipper_axle_diameter)))
;; power_button_z + encoder_z + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 1 2) (1))
  '(on_off_button_z flipper_encoder_z structure_thickness)
  '(bot_body_length_z)))
;; batteries_z + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(battery_z structure_thickness)
  '(bot_body_length_z)))
;; wheel_motor_z + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(wheel_motor_z structure_thickness)
  '(bot_body_length_z)))
;; flipper_motor_z + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(flipper_motor_z structure_thickness)
  '(bot_body_length_z)))
;; range_sensor_z + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(proximity_sensor_z structure_thickness)
  '(bot_body_length_z)))
;; power_board_z + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(power_supply_board_z structure_thickness)
  '(bot_body_length_z)))
;; flipper_axle_diam + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(flipper_axle_diameter structure_thickness)
  '(bot_body_length_z)))
;; flipper_axle_diam + 2* thick < wheel_diameter
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(flipper_axle_diameter structure_thickness) 
  '(wheel_diameter)))
;; flipper_axle_diam < flipper_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1) (1))
  '(flipper_axle_diameter) 
  '(flipper_length_z)))
;; flipper_gear_diameter + 2*thick < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(flipper_axle_gear_diameter structure_thickness)
  '(bot_body_length_z)))
;; flipper_z < body_z
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1) (1))
  '(flipper_length_z) 
  '(bot_body_length_z)))
;; body_z + 2*clearance < wheel_diameter
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1 2) (1))
  '(bot_body_length_z wheel_clearance) 
  '(wheel_diameter)))

;; Deferred
;center of mass
;CG location vs. flipper_x
;turning speed simulation
;; Optional:
;man-portable: m < 10kg
;pocket-sized: m < 2kg, volume < 1000 cc

;;;;;;;;;; Show what we have wrought:
(setf show-cells nil)
(output-graphviz "minidroid.dot" (/ 11 8.5))
(format t "Params = ~A~%" (count-if (ppf typep ? 'parameter) (all-elements)))
(format t "Dimensions = ~A~%" (count-if (ppf typep ? 'dimension) (all-elements))) ; 7 per box
(format t "Relations = ~A~%" (count-if (ppf typep ? 'relation) (all-elements)))

;;(print-topology-stats)

;; Attach Values
(defun minidroid-initial-values ()
  (parameter-values 
   `((BOT_BODY_LENGTH_X ,(in->m 7.00))	; blueprint spec
     (BOT_BODY_LENGTH_Y ,(in->m 4.5))	; measured from blueprint
     (BOT_BODY_LENGTH_Z ,(in->m (* 0.75 2))) ; blueprint spec
     ;; Derived: BOT_BODY_MASS, 
     (CASE_ROUNDING_DIAMETER ,(in->m 1.0)) ; measured from blueprint
     (STRUCTURE_THICKNESS ,(in->m 0.16)) ; measured from blueprint
     ;; Derived: CASE_ROUNDING_SURFACE_LOSS, STRUCTURE_SURFACE, STRUCTURE_MASS
     
     (WHEEL_CLEARANCE ,(in->m 0.25))	; derived from blueprint spec 0.25
     (WHEEL_DIAMETER ,(in->m 2.20))	; blueprint spec
     (WHEEL_Y ,(in->m (- 6.76 4.5)))	; measure + spec blueprint
     (WHEEL_SEPARATION ,(in->m (- 5.43 2.20))) ; blueprint spec
     (TRACK_THICKNESS ,(in->m 0.13))	; measured from blueprint
     ;; Derived: WHEEL_MASS, TRACK_LENGTH, TRACK_MASS
     
     (FLIPPER_LENGTH_X ,(in->m 5.50))	; blueprint spec
     (FLIPPER_LENGTH_Y ,(in->m (/ (- 7.71 6.76) 2))) ; blueprint spec
     (FLIPPER_LENGTH_Z ,(in->m 1.0))	; measured from blueprint
     (FLIPPER_AXLE_DIAMETER ,(in->m 0.32)) ; measured from blueprint
     (FLIPPER_AXLE_GEAR_DIAMETER ,(in->m 1.15)) ; measured from blueprint
     ;; Derived: FLIPPER_ROTATIONAL_SPEED, FLIPPER_MASS,
     ;;   FLIPPER_AXLE_MASS, FLIPPER_AXLE_TORQUE_LIMIT, FLIPPER_AXLE_Y
     
     ;; Assuming motors are hs-985mg, 6.0v
     (WHEEL_MOTOR_X 39.88e-3)
     (WHEEL_MOTOR_Y 36.83e-3)
     (WHEEL_MOTOR_Z 19.81e-3)
     (WHEEL_MOTOR_MAX_SPEED ,(/ (/ pi 3) 0.13)) ; 60 degrees in 0.13 sec
     (WHEEL_MOTOR_POWER 1.5)		; 7.7mA idle, 180 mA no-load, ~1.5 A at stall
     (WHEEL_MOTOR_TORQUE ,(kg-cm->N-m 12.4))
     (WHEEL_MOTOR_MASS 61.8e-3)
     
     (FLIPPER_MOTOR_X 39.88e-3)
     (FLIPPER_MOTOR_Y 36.83e-3)
     (FLIPPER_MOTOR_Z 19.81e-3)
     (FLIPPER_MOTOR_MAX_SPEED ,(/ (/ pi 3) 0.13)) ; 60 degrees in 0.15 sec
     (FLIPPER_MOTOR_POWER 1.5)		; 7.7mA idle, 180 mA no-load, ~1.5 A at stall
     (FLIPPER_MOTOR_TORQUE ,(kg-cm->N-m 12.4))
     (FLIPPER_MOTOR_MASS 61.8e-3)
     ;; Derived: FLIPPER_MOTOR_GEAR_MULTIPLIER 
     
     ;; each battery pack is 2 AAs: stats taken from AA specs
     (BATTERY_X 50.5e-3)
     (BATTERY_Y 14e-3)
     (BATTERY_Z ,(* 2 14e-3))
     ;; Derived: BATTERY_MASS (~30g), BATTERY_ENERGY (~2000mAh/AA)
     
     ;;; Electronics:
     ;; CPU based on Arduino MEGA board
     (CPU_X 68.6e-3)
     (CPU_Y 53.3e-3)
     (CPU_Z 15e-3)
     (CPU_POWER 10e-3)
     (CPU_MASS 40e-3)
     
     ;; Proximity sensor based on Sharp GP2D120XJ00F, narrowed to match blueprints
     (PROXIMITY_SENSOR_X 14e-3)
     (PROXIMITY_SENSOR_Y 30e-3)
     (PROXIMITY_SENSOR_Z 13e-3)
     (PROXIMITY_SENSOR_RANGE 40e-2)
     (PROXIMITY_SENSOR_POWER 33e-3)
     (PROXIMITY_SENSOR_MASS 3.6e-3)
     
     ;; Flipper encoder based on STEC11B09 encoder
     (FLIPPER_ENCODER_X 11e-3)
     (FLIPPER_ENCODER_Y 20e-3)
     (FLIPPER_ENCODER_Z 11e-3)
     (FLIPPER_ENCODER_POWER 10e-3)
     (FLIPPER_ENCODER_MASS 10e-3)	; guess, based on similar datasheets
     
     (POWER_SUPPLY_BOARD_X 28e-3)	; measured from blueprints
     (POWER_SUPPLY_BOARD_Y 28e-3)	; measured from blueprints
     (POWER_SUPPLY_BOARD_Z 5e-3)	; guess based on CPU
     (POWER_SUPPLY_BOARD_MASS 10e-3)	; guess based on CPU
     
     (ON_OFF_BUTTON_X 18e-3)		; measured from blueprints
     (ON_OFF_BUTTON_Y 18e-3)		; measured from blueprints
     (ON_OFF_BUTTON_Z ,(in->m 0.3))	; guess based on blueprints
     (ON_OFF_BUTTON_MASS 5e-3)		; guess based on CPU
     
     
     ;; Environment parameters:
     (STEP_HEIGHT ,(in->m 4.0))		; initially ~10 cm
     )))

(minidroid-initial-values) ;; comment out to initialize from a different start

(setf show-values-in-parameters t)
(output-graphviz "minidroid-with-values.dot" (/ 11 8.5))
(format t "Values = ~A~%" (count-if (ppf typep ? 'value) (all-elements)))
;; 56 values assigned, out of 111 parameters

;; At beginning, we have the following problems:
;; COMPOSITE-SUM-BOUND (0.797): FLIPPER_ENCODER_POWER = 1.000e-2 CPU_POWER = 1.000e-2 BATTERY_ENERGY = 3.858e+4 FLIPPER_MOTOR_POWER = 1.500e+0 WHEEL_MOTOR_POWER = 1.500e+0 PROXIMITY_SENSOR_POWER = 3.300e-2
;; COMPOSITE-SUM-BOUND (0.309): BOT_BODY_LENGTH_Z = 3.810e-2 STRUCTURE_THICKNESS = 4.064e-3 FLIPPER_AXLE_GEAR_DIAMETER = 2.921e-2
;; COMPOSITE-SUM-BOUND (0.007): BOT_BODY_LENGTH_Z = 3.810e-2 STRUCTURE_THICKNESS = 4.064e-3 BATTERY_Z = 2.800e-2
;; allowing them to resolve gives stress graph:
;; 1.1131 0.8174 0.5282 0.3837 0.1167 0.0349 0.0186 0.0142 0.0000

;; Moving to 3 kJ margin, rather than 1 kJ margin:
;; 0.91535836 0.7492686 0.5917056754672297d0 0.54276127 0.41003895
;; 0.42577733814942714d0 0.4202165943692374d0 0.36344963 0.35131046 0.35760188
;; 0.30544722 0.30518258 0.30306876 0.27389973 0.22038756 0.25912845 0.43342528
;; 0.21910287 0.29229102 0.22886324 0.18780541 0.20258173 0.105313085 0.23694259
;; 0.077010244 0.19642034 0.11145208 0.081208736 0.11546605 0.04414434
;; 0.15067652 0.04251432 0.044390652 0.17134453 0.010442453 0.07431772
;; 0.025303904 0.046094853 0.010468385 0.042821527 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;; 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0
;; 0 0 0 0 0 0

;; example target:
;; should run for around 10000 to 15000 steps before resolving
(defun wheel-clearance-test ()
 (add-user-target 'wheel_clearance 0.04 5)
 ;(setf singleton-step-size 0.005)
 (setf clearancestress (make-stress-graph 12000)))

(defun step-height-test ()
 (add-user-target 'step_height 0.55)
 ;(setf singleton-step-size 0.005)
 (setf stepstress (make-stress-graph 30000)))

;; One the FBs exist, it takes no more than 5 minutes per FB 
;; and often much less
;; The simulations, of course, are a totally different matter...


;;; SIMULATION ENVIRONMENT
; BOT_POSITION_X = 
; BOT_POSITION_Y = 
; BOT_POSITION_Z = 

; STEP_LENGTH_X = 
; STEP_LENGTH_Y =                                     
; STEP_LENGTH_Z = 
; STEP_POSITION_X = 
; STEP_POSITION_Y = 
; STEP_POSITION_Z = 
(parameter-value 'wheel_clearance)
(parameter-value 'step_height)
(problem-summary )
(all-parameter-values )

;; What I think will work:
;; Make unit identities record which ones changed, and only push back on them

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Inspecting likely problems:
;; - I think that the unit identities may need to be handled separately
;;   of 112 relations, 40 are of type unit-identity
;;   others are also effectively the same class, but implemented differently
