;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;; Design Parameters
(clear-topology)
(setf design-name "tetralander")

;; Densities in kg/m^3
(defun g/cc->kg/m^3 (n) (* n 1000))	; water is 1 g/cc, 1000 kg/m^3
(defparameter titanium-density (g/cc->kg/m^3 4.23))
(defparameter electronics-density 1182.9581)
;; Other constants:
;; martian insolation is ~44% of earth's
(defparameter solar-power-density (* 150 0.44)) ; watts/meter^2
(defparameter g0 9.80665)		; standard gravity
(defparameter g_mars 3.711)
(defun kg-cm->N-m (torque) (* g0 0.01 torque))

(ensure-parameter 'bot_body_length 'meters) ; tetrahedral edge
(ensure-parameter 'bot_body_face 'meters^2)
(ensure-parameter 'bot_body_volume 'meters^3)
(ensure-parameter 'bot_body_mass 'kilograms)
;; mass/volume relation more complicated - defined below
(name-it "Equilateral triangle area"
 (make-identity-relation			; length/area relation
  `(,(/ (sqrt 3) 4))			; equilateral triangle
  '(bot_body_length bot_body_length)
  '(bot_body_face)))
(name-it "Tetrahedron edge/volume"
 (make-identity-relation			; length/volume relation
  `(,(/ (sqrt 2) 12))
  '(bot_body_length bot_body_length bot_body_length)
  '(bot_body_volume)))

(ensure-parameter 'structure_mass 'kilograms)
(ensure-parameter 'structure_thickness 'meters)
(ensure-parameter 'structure_surface 'meters^2)
;; surface defined from body covering
(name-it "Tetrahedron surface area"
 (make-identity-relation			; surface is cover of body
  `(4)					; 4 equilateral triangles
  '(bot_body_face)
  '(structure_surface)))
(name-it "Tetrahedron mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,(* 1.1 titanium-density) kilograms meters^-3) ; surface + 10% mounting volume
  '(structure_surface structure_thickness)
  '(structure_mass)))

(ensure-parameter 'instrument_package_volume 'meters^3)
(ensure-parameter 'instrument_package_mass 'kilograms)
(ensure-parameter 'instrument_package_power 'watts)
(name-it "instrument mass/volume"
 (make-identity-relation			; mass/volume relation
  `(,electronics-density kilograms meters^-3)
  '(instrument_package_volume)
  '(instrument_package_mass)))

(ensure-parameter 'solar_cell_area 'meters^2)
(ensure-parameter 'solar_cell_power 'watts)
(name-it "solar-generation"
 (make-identity-relation
  `(,solar-power-density watts meters^-2)
  '(solar_cell_area)
  '(solar_cell_power)))

(extrude-box 'flipper_motor_x 'flipper_motor_y 'flipper_motor_z)
(ensure-parameter 'flipper_motor_mass 'kilograms)
(ensure-parameter 'flipper_motor_torque 'Newton-Meters)
(ensure-parameter 'flipper_motor_power 'amps)
(ensure-parameter 'flipper_motor_max_speed 'radians/second)

;; total body mass is the sum of all component masses
(name-it "total-mass"
 (make-summation 'bot_body_mass
		 'structure_mass 
		 'flipper_motor_mass 'flipper_motor_mass 'flipper_motor_mass ; x3
		 'instrument_package_mass))



;;; Functional blueprints

;; Motor relations
; (make-quantized-component 
;  servo-library 
;  'flipper_motor_x 'flipper_motor_y 'flipper_motor_z 'flipper_motor_mass 
;  'flipper_motor_torque 'flipper_motor_power 'flipper_motor_max_speed)

(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ (* 39.88e-3 36.83e-3 19.81e-3) 1.216))
  '(flipper_motor_torque) '(flipper_motor_xyz)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ 61.8e-3 1.216))
  '(flipper_motor_torque) '(flipper_motor_mass)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ 1.5 1.216))
  '(flipper_motor_torque) '(flipper_motor_power)))
(name-it "quantized: motor"
 (make-identity-relation 
  `(,(/ (/ (/ pi 3) 0.13)))
  '(flipper_motor_max_speed) '()))
;; Connect these together into a component relation, allow topological changes
(name-it "component: flipper motor"
  (bind-into-component
   ;; bounding relations
   '()
   ;; scalable dimensions: implicitly includes all derived units
   '(flipper_motor_x flipper_motor_y flipper_motor_z) 
   ;; cumulative parameters
   '(flipper_motor_mass flipper_motor_power flipper_motor_torque)
   ;; non-cumulative parameters
   '(flipper_motor_max_speed)))
   
;; structure thickness (thus strength) should be proportional to length
(name-it "proportional-structure"
 (make-identity-relation
  `(,(/ 0.2 5e-3))
  '(structure_thickness)
  '(bot_body_length)))

;; instrument volume < body volume
;; (motors assumed to be on outside)
(name-it "layout-geometry"
 (make-sum-bound-relation
  '((1) (1))
  '(instrument_package_volume)
  '(bot_body_volume)))
; -0.0001 0.0002)

;; solar cell up to 80% of surface area
;; size: cell_area < 0.8 * 3 faces
(name-it "panel-area-budget"
 (make-sum-bound-relation
  `((1) (,(* 0.8 3)))
  '(solar_cell_area)
  '(bot_body_face)))

;; solar power is sufficient: instruments < solar
;; (ignores flippers, which are only used on opening)
(name-it "sufficient-power"
 (make-sum-bound-relation
  '((1) (1))
  '(instrument_package_power)
  '(solar_cell_power)))

;; flip self over
(defun flipping-eval (relation vars)
  (destructuring-bind (torque body_x mass speed) vars
    (declare (ignore relation speed))
    (/ (* 0.5 body_x g_mars mass) (* torque))))

(name-it "self-righting"
 (make-modular-relation
  '((flipper_motor_torque bot_body_length bot_body_mass flipper_motor_max_speed))
  '(flipper_motor_torque)
  '(bot_body_length bot_body_mass)
  (let ((plsf (positive-linear-stress-fn 0.7 1.0)))
    (eval-stress #'flipping-eval plsf))))
(mark (find-relation "self-righting") 'simulation)



;;;; ATTACH VALUES
(parameter-values 
 `((BOT_BODY_LENGTH 0.2)
   ;; Derived: BOT_BODY_MASS, BOT_BODY_FACE, BOT_BODY_VOLUME
   (STRUCTURE_THICKNESS 5e-3)
   ;; Derived: STRUCTURE_MASS, STRUCTURE_SURFACE
   
   (INSTRUMENT_PACKAGE_VOLUME 0.0005)	; 500 cm^3
   (INSTRUMENT_PACKAGE_MASS 0.5)	; half a kilograph
   (INSTRUMENT_PACKAGE_POWER 0.5)	; 1/2 watt initial assumption
   
   (SOLAR_CELL_AREA 0.015)
   ;; Derived: SOLAR_CELL_POWER

   (FLIPPER_MOTOR_X 39.88e-3)
   (FLIPPER_MOTOR_Y 36.83e-3)
   (FLIPPER_MOTOR_Z 19.81e-3)
   (FLIPPER_MOTOR_MAX_SPEED ,(/ (/ pi 3) 0.13)) ; 60 degrees in 0.15 sec
   (FLIPPER_MOTOR_POWER 1.5); 7.7mA idle, 180 mA no-load, ~1.5 A at stall
   (FLIPPER_MOTOR_TORQUE ,(kg-cm->N-m 12.4)) ; about 1.21
   (FLIPPER_MOTOR_MASS 61.8e-3)
   ))

(defun test-instruments ()
  (add-user-target 'instrument_package_power 3)
  (add-user-target 'instrument_package_mass 3)
;; should run for ~5000 steps before resolving
  (setf landerstress (make-stress-graph 6000)))
(problem-summary)

;; Note: this took 2 hours to build (~1 to build, 1 to debug)
(output-graphviz "tetralander.dot" (/ 11 8.5))
