;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; GraphViz output
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar show-cells t)
(defvar show-non-extrema nil)
(defvar show-gradient-bounds nil)
(defvar show-values-in-parameters nil)
(defvar use-neato t)
(defvar stress-coloring t)
(defvar neato-length 3) ; or 10
(defvar transitive-constraints nil) ; nil, 'constraints, 'parameters
(defvar show-priority-values nil)

;; Note: the extra ",1" element in the aspect ratio attribute is needed
;; to prevent a graphviz bug that duplicates multi-edges
(defun output-graphviz (&optional (filename "blueprint.dot") (aspect nil))
  (with-open-file (dotfile filename :direction :output :if-exists :new-version)
    (format dotfile "digraph Blueprint { ~@[aspect=\"~A,1\"~] ~%" aspect)
    (when use-neato (format dotfile "layout=\"neato\";~%"))
    ;;(when use-neato (format dotfile "overlap=\"false\";~%"))
    (if transitive-constraints
	(output-transitives dotfile)
      (progn
	(dolist (e (hash-keys elements)) (->graphviz e dotfile))
	(dolist (l (hash-keys links)) (->graphviz l dotfile))))
    (format dotfile "}~%")
    'done))

(defun graphviz-sublabel (label sublabels &optional ports (fontsize 10))
  (concatenate 'string
    (format nil "~%    label=<<TABLE BORDER=\"0\" CELLBORDER=\"0\" CELLSPACING=\"0\">~%")
    (format nil "    <TR><TD COLSPAN=\"~A\">~A</TD></TR>~%" (if sublabels (length sublabels) 1) label)
    (when sublabels
      (unless ports (setf ports (make-list (length sublabels))))
      (let ((argsets (mapcar (lambda (p sl) (list p fontsize sl)) 
			     ports sublabels)))
	(format nil "    <TR>~%\~:{      <TD~@[ PORT=\"~A\"~]><FONT POINT-SIZE=\"~A\">~
                    ~A</FONT></TD>~%~}</TR>~%" argsets)))
    (format nil "    </TABLE>>")))


(defgeneric ->graphviz (object stream) )
(defmethod ->graphviz ((c cell) stream)
  (when show-cells
    (let* ((cvs (coordinates c))
	   (coordinates
	    (if show-non-extrema cvs (remove-if-not #'extremum? cvs)))
	   (cnames (mapcar (cmp name tail) coordinates))
	   (cports (mapcar #'uid coordinates)))
      (format stream "~A [shape=rectangle ~A];~%" (uid c) 
	      (graphviz-sublabel (format nil "~A-Cell" (dimension c))
				 cnames cports))
      ;; create boundary links, if desired
      (when show-gradient-bounds
	(let* ((gcs (remove-if-not (ppf typep ? 'graduated-coordinate) 
				   coordinates))
	       (gcuid (mapcar (lambda (gc) 
				(format nil "~A:~A" (uid c) (uid gc)))
			      gcs))
	       (b-uid (mapcar (cmp uid bound-) gcs))
	       (b+uid (mapcar (cmp uid bound+) gcs)))
	  (format stream "~:{~A:c -> ~A [color=\"red\" style=\"dashed\"];~%~}"
		  (mapcar #'list gcuid b-uid))
	  (format stream "~:{~A:c -> ~A [color=\"black\" style=\"dashed\"];~%~}"
		  (mapcar #'list gcuid b+uid))
	  )))))

(defmethod ->graphviz ((cs coordinate-system) stream)
  (declare (ignore stream)) ) ;; Coordinate systems handled via cell
(defmethod ->graphviz ((cs lattice-element) stream)
  (declare (ignore stream)) ) ;; Coordinate systems handled via cell
(defmethod ->graphviz ((p parameter) stream)
  (let* ((dset (mapcar (lambda (s) (let ((n (count s (deltas p) :key #'car)))
				     (list (unless (zerop n) n) s)))
		       '(+ -)))
	 (dlist (mapcar (lambda (args) (apply #'format nil "~@[~A~A~]" args))
			(remove nil dset :key #'first)))
	 (s (stress p))
	 (abs-s (abs s))
	 (fillcolor 
	  (if stress-coloring
	      (format nil "~,3F ~,3F 1" 
		      (+ 5/6 (* 1/6 (if (plusp abs-s) (/ s abs-s) 0)))
		      (if (plusp abs-s) (/ abs-s current-max-abs-stress) 0))
	    (cond ((and (first (first dset)) (first (second dset))) "deeppink") ; up/down
		  ((first (first dset)) "aquamarine") ; going up
		  ((first (second dset)) "dodgerblue") ; going down
		  (t "lightblue")))))	; unchanged
    (when show-values-in-parameters
      (let ((vlist (get-values p)))
	(cond ((= (length vlist) 1)
	       (push (format nil "Value = ~,4F" (value (first vlist))) dlist))
	      ((> (length vlist) 1)
	       (push "Multiple Values" dlist))
	      (t 
	       (setf fillcolor "blueviolet")))))
    (when show-priority-values
      (let ((pri (mark? p 'priority)))
	(when pri (push (format nil "<FONT POINT-SIZE=\"15\" COLOR=\"green\">Priority = ~A</FONT>" pri) dlist))))
    (format stream "~A [shape=oval style=filled fillcolor=\"~A\" ~A];~%"
	    (uid p) fillcolor (graphviz-sublabel (name p) dlist))))
(defmethod ->graphviz ((v value) stream)
  (unless show-values-in-parameters
    (with-slots (number) v
      (format stream "~A [shape=diamond style=filled fillcolor=\"lightgrey\" label=\"~,3E\"];~%" (uid v) number))))
(defmethod ->graphviz ((r relation) stream)
  (let* ((s (if (plusp current-max-abs-stress)
		(/ (stress r) current-max-abs-stress)
	      0))
	 (stresscolor 
	  (if (not (enabled r))
	      "lightgrey"
	    (if stress-coloring
		(format nil "~,3F ~,3F 1" (if (plusp s) 0 2/3) (abs s))
	      (if (<= (abs (stress r)) 0) "green" 
		(if (>= (abs (stress r)) 1) "red" "yellow")))))
	 (label (or (name r) (type-of r)))
	 (dlist '()))
    (when show-priority-values
      (let ((pri (mark? r 'priority)))
	(when pri (push (format nil "<FONT POINT-SIZE=\"15\" COLOR=\"green\">Priority = ~A</FONT>" pri) dlist))))
    (format stream "~A [shape=~A style=filled fillcolor=\"~A\" ~A];~%" 
	    (uid r) 
	    (if (equality-relation r) "doubleoctagon" 
	      (if (typep r 'functional-blueprint) "hexagon" "octagon"))
	    stresscolor 
	    (if dlist
		(graphviz-sublabel label dlist)
	      (format nil "label=\"~A\"" label)))))

(defmethod ->graphviz ((b boundary) stream)
  (when show-cells
    (with-slots (bounded bound) b
      (format stream "~A -> ~A [color=\"red\"];~%" (uid bounded) (uid bound)))))
(defmethod ->graphviz ((s subspace) stream)
  (when show-cells
    (with-slots (parent child) s
      (format stream "~A -> ~A [color=\"darkorange\"];~%" (uid parent) (uid child)))))
(defmethod ->graphviz ((p derived-parallel) stream)
  (when show-cells
    (with-slots (parent child) p
      (format stream "~A -> ~A [color=\"pink\"];~%" (uid parent) (uid child)))))
(defmethod ->graphviz ((m measurement) stream)
  (when show-cells
    (with-slots (parameter object) m
      (unless (typep object 'lattice-span)
	(format stream "~A -> ~A [color=\"green\"];~%" (uid parameter) (uid object))))))
(defmethod ->graphviz ((m coordinate-value) stream)
  (declare (ignore stream)) ) ;; Coordinate systems handled via cell
(defmethod ->graphviz ((a assignment) stream)
  (unless show-values-in-parameters
    (with-slots (parameter value) a
      (format stream "~A -> ~A [color=\"blueviolet\"];~%" (uid value)(uid parameter)))))
(defmethod ->graphviz ((d derivation) stream)
  (unless show-values-in-parameters
    (with-slots (derived source) d
      (format stream "~A -> ~A [color=\"cyan\"];~%" (uid source)(uid derived)))))
(defmethod ->graphviz ((c constraint) stream)
  (with-slots (relation parameter) c
    (let* ((s (if (plusp current-max-abs-stress)
		  (* (/ (stress relation) current-max-abs-stress)
		     (if (riser relation parameter) 1 -1))
		0))
	   (scolor (format nil "~,3F ~,3F 1" (if (plusp s) 0 2/3) (abs s))))
      (format stream "~A -> ~A [color=\"~A\" weight=~A penwidth=~A~A];~%" 
	      (uid relation) (uid parameter)
	      (if stress-coloring (if (zerop s) "grey" scolor) "blue")
	      (if stress-coloring (round (+ 1 (* 100 (abs s)))) 1)
	      (if stress-coloring (round (+ 1 (* 20 (abs s)))) 1)
	      ;;(round (+ 30 (log (max (expt 2 -9) (abs n)) (expt 2 1/3))))
	      (if use-neato (format nil " len=~A" neato-length) "")))))

(defun output-transitives (stream)
  (flet ((stress-link (stress a b)
	   (let* ((s (if (plusp current-max-abs-stress)
			 (/ stress current-max-abs-stress)
		       0))
		  (scolor (format nil "~,3F ~,3F 1" (if (plusp s) 0 2/3) (abs s))))
	     (format stream "~A -> ~A [color=\"~A\" weight=~A penwidth=~A~A];~%"
		     (uid a) (uid b) 
		     (if stress-coloring (if (zerop s) "grey" scolor) "blue")
		     (if stress-coloring (round (+ 1 (* 100 (abs s)))) 1)
		     (if stress-coloring (round (+ 1 (* 2 (abs s)))) 1)
		     (if use-neato (format nil " len=~A" neato-length) "")))))
    (cond ((eq transitive-constraints 'constraints)
	   (dolist (e (all-relations)) (->graphviz e stream))
	   (dolist (clique (find-transitive-relations))
	     (let ((param (car clique)) (relations (cdr clique)))
	       (pairwise (lambda (a b)
			   (when (lt (uid a) (uid b))
			     (stress-link (stress param) a b)))
			 relations relations))))
	  ((eq transitive-constraints 'parameters)
	   (dolist (clique (find-transitive-parameters))
	     (let ((reln (car clique)) (parameters (second clique)))
	       (pairwise (lambda (a b)
			   (when (lt (uid a) (uid b))
			     (stress-link (stress reln) a b)))
			 parameters parameters)))
	   (dolist (e (all-parameters)) (->graphviz e stream)))
	  (t (error "Unknown transitive mode: ~A" transitive-constraints)))))

(defun find-transitive-relations ()
  (let ((cliques '()))
    (dolist (p (all-parameters))
      (push (cons p (relations p)) cliques))
    cliques))

(defun find-transitive-parameters ()
  (let ((cliques (make-hash-table)))
    (dolist (p (all-parameters))
      (dolist (r (relations p))
	(push p (gethash r cliques))))
    (hash->list cliques)))
	