;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; Mesh output for rendering, e.g. in MeshLab
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun output-mesh (&optional (filename "blueprint.ply"))
  (with-open-file (file filename :direction :output :if-exists :new-version)
    (let ((vertices (find-elements '((type ?x cell) (dimension ?x 0))))
	  (faces (find-elements '((type ?x cell) (dimension ?x 2)))))
      (output-ply-header file vertices faces)
      (let ((vertex-ids (make-hash-table)) (cur-id -1))
	(dolist (v vertices)
	  (setf (gethash v vertex-ids) (incf cur-id))
	  (output-ply-vertex file v coordinates))
	(dolist (f faces)
	  (output-ply-face file f vertex-ids)))
      (output-ply-footer file)
      'done)))

(defun output-ply-header (file vertices faces)
  (format file "ply~%")
  (format file "format ascii 1.0~%")
  (format file "comment Blueprint generated by BBN MADV~%")
  (format file "element vertex ~A~%" (length vertices))
  (format file "property float x~%")
  (format file "property float y~%")
  (format file "property float z~%")
  (format file "element face ~A~%" (length faces))
  (format file "property list uchar int vertex_index~%")
  (format file "end_header~%"))

(defun output-ply-footer (file) 'done)

(defun output-ply-vertex (file cell coordinates)
  (let ((coord (gethash cell coordinates)))
    (unless coord
      (warn "No coordinates assigned to vertex ~A, defaulting to zeros" cell)
      (setf coord (list 0 0 0)))
    (format file "~F ~F ~F~%" (first coord) (or (second coord) 0) 
	    (or (third coord) 0))))


(defun output-ply-face (file cell vertex-ids)
  (let ((edges (boundaries cell)) (vertex nil))
    (format file "~A" (length edges))
    ;; grab an arbitrary vertex to start with
    (setf vertex (first (boundaries (car edges))))
    ;; chase the vertex around the edge list
    (while edges
      (let ((nextedge (find-if (lambda (e) (find vertex (boundaries e))) edges)))
	(format file " ~A" (gethash vertex vertex-ids))
	(setf vertex (first (remove vertex (boundaries nextedge))))
	(setf edges (remove nextedge edges))))
    (format file "~%")))




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Coordinatization of elements
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defvar coordinates (make-hash-table))
(defvar embedding)			; k-dimensional parameters
(defvar line-alignment (make-hash-table)) ; <1-Cell> -> embedding parameter
(defvar worklist)
(defun initialize-coordinatization-variables (root-params)
  (clrhash coordinates) (clrhash line-alignment)
  (setf worklist '()) (setf embedding root-params))

(defun parameter-vertices (parameter)
  (flatten 
   (mapcar (cmp boundaries head)
	   (remove-if (ppf typep ?c 'extremum) 
		      (links-of-type parameter 'measurement nil)))))

(defun assign-coordinate (vertex coord)
  (add-1cells-to-worklist vertex)
  (setf (gethash vertex coordinates) coord))

(defun add-1cells-to-worklist (vertex)
  (let ((1cells (links-of-type vertex 'boundary t)))
    (dolist (e 1cells) (pushnew (tail e) worklist))))

(defun translate (coord dim value)
  (if (= dim 0)
      (cons (+ (first coord) value) (rest coord))
    (cons (first coord) (translate (rest coord) (1- dim) value))))

(defun coordinatize-vertices (root-params)
  (initialize-coordinatization-variables root-params)
  ;; First find an arbitrary 0-cell that touches all of the root parameters
  ;; That cell will become the origin point of the coordinatization
  (let* ((root-vertices (mapcar #'parameter-vertices root-params))
	 (shared (reduce #'intersection root-vertices)))
    (assert (first shared) () "No vertex was linked to all root parameters")
    (assign-coordinate (first shared) 
		       (make-list (length embedding) :initial-element 0)))
  ;; Next, bridge outward from that element by considering connected 1-cells
  (while worklist
    (let* ((1cell (pop worklist)) (vertices (boundaries 1cell))
	   (assigned 
	    (first (remove-if-not (ppf gethash ?x coordinates) vertices)))
	   (unassigned 
	    (first (remove-if (ppf gethash ?x coordinates) vertices))))
      (assert (= (length vertices) 2) (1cell vertices) 
	      "Invalid 1cell: ~A has ~A vertices" 1cell (length vertices))
      (when (and assigned unassigned)	; one good, one bad
	(attempt-to-assign-coordinate unassigned assigned 1cell))))
  ;; Assume we're complete (though we may not be)
  'done)

;; Given a 1cell, attempt to determing which embedding coordinate it
;; is parallel to.  Return nil if not possible, or three values:
;; local coordinate, embedding coordinate, inverted?
;; NOTE: DOESN'T HANDLE ANTIPARALLEL YET
(defun find-1cell-alignment (1cell)
  (or 
   ;; First, check if the answer is already cached
   (gethash 1cell line-alignment)
   ;; Check if the 1-cell has a direct mapping to the embedding
   (let* ((measures (mapcar #'tail (measurements 1cell t)))
	  (embed (first (intersection embedding measures))))
     (when embed
       (setf (gethash 1cell line-alignment) (list embed embed nil))))
   ;; Check if the 1-cell is parallel to a 1-cell already aligned
   (let* ((measures (mapcar #'tail (measurements 1cell t)))
	  (parallels (mapcar #'tail (links-of-type 1cell 'derived-parallel t)))
	  (alignments (mapcar (ppf gethash ?x line-alignment) parallels))
	  (aligned (first (remove nil alignments))))
     (when (and measures aligned)
       (setf (gethash 1cell line-alignment) 
	 (list (first measures) (second aligned) nil))))
   ;; Check if the 1-cell is perpendicular to all but one of the embedding
   ))


(defun attempt-to-assign-coordinate (unassigned assigned 1cell)
  ;; First, check if the 1cell has a direct mapping to the embedding
  (let ((alignment (find-1cell-alignment 1cell)))
    (when alignment
      (destructuring-bind (coord embed-coord inverted?) alignment
	(let ((value (first (get-values coord))))
	  (when value
	    (let* ((dim (position embed-coord embedding))
		   (v (* (value value)
			 (if (extremum? unassigned coord 'maximum) 1 -1)
			 (if inverted? -1 1)))
		   (newcoord (translate (gethash assigned coordinates) dim v)))
	      (assign-coordinate unassigned newcoord))))))))

		
;;; Current state of coordinate inference:
;; - root infers correctly
;; - subspaces infer iff they share a vertex
;; - extrusions don't yet infer correctly

;;; Other means of coordinate inference:
;;; - If there's an upper-bound relation from Sub to X, then Sub may be parallel
;;;   to X.  Of course, there may be multiple ones here...
;;; - A 1-cell marked with extrema is perpendicular to those measures
;;;   in extrusions, we often have things multiply marked... 
;;;   those extra marks will show up on all of the 2-cells the 1-cell bounds, too
;; - a floating subspace can get values from its gap parameters




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Coordinatization of elements
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Given K root parameters, look for a K-cell measured by their product
;; From there, proceed by chaining through the set of K-cells
;; * Any K-cell that shares a K-1 boundary with prior K-cells can be
;;   coordinatized by setting the non-shared dimensions parallel
;; * Any subspace K-cell can be coordinatized by following the subspace
;;   relations on its dimensions

(defvar coordinates (make-hash-table))
(defvar progress t)
(defvar worklist)
(defvar resolved)
(defvar param-alignment
    (make-hash-table :test #'equal))	; (cell, param) -> root-param
(defun initialize-coordinatization-variables (root-params)
  (clrhash coordinates) (setf worklist '()) (setf resolved '())
  (clrhash param-alignment)
  (setf progress t))

(defun cell-dimensions (cell)
  (mapcar #'tail (remove-if-not #'base-span? (measurements cell))))
(defun vertices (cell)
  (remove-if-not #'zerop (cons cell (boundaries* cell)) :key #'dimension))

(defun dim-length (dim)
  (let ((vset (get-values dim)))
    (assert (= (length vset) 1))
    (value (first vset))))

(defun coordinatize-root (cell dimensions)
  (dolist (p dimensions) (setf (gethash (cons cell p) param-alignment) p))
  (let ((lengths (mapcar #'dim-length dimensions))
	(vlist (vertices cell)))
    (flet ((dim-mult (cvlist dim)
	     (if (find-if #'minimum? 
			  (remove dim cvlist :test-not #'eq :key #'tail)
			  :key #'value)
		 -0.5 0.5)))
      (dolist (v vlist)
	(let ((cvs (remove-if-not (ppf typep ?X 'coordinate-value) 
				  (measurements v))))
	  (setf (gethash v coordinates)
	    (mapcar (lambda (d l) (* (dim-mult cvs d) l))
		    dimensions lengths)))))
    (push cell resolved)))

(defun find-minimum-vertex (cell)
  (let ((cell-dims (cell-dimensions cell))
	(vlist (vertices cell)))
    (labels ((min-of-dim (d cv)
	       (and (eq (tail cv) d) (minimum? (value cv))))
	     (all-dims-min (v)
	       (let ((cvlist (remove-if-not (ppf typep ?X 'coordinate-value)
					    (measurements v))))
		 (not 
		  (find-if-not (lambda (d) (find-if (ppf min-of-dim d) cvlist))
			       cell-dims)))))
      (find-if #'all-dims-min vlist))))

(defun displace-if-max (v dim length)
  (labels ((max-of-dim (d cv)
	     (and (eq (tail cv) d) (maximum? (value cv)))))
    (let ((cvlist (remove-if-not (ppf typep ?X 'coordinate-value)
				 (measurements v))))
      (if (find-if (ppf max-of-dim dim) cvlist) length 0))))

;;;;;;;
;; Parameter alignment does not yet properly handle the problem of 
;; dimensions that have different orientations when they are used
;; in different component instances.
;;;;;;;
(defun align-parameters (cell params root-params)
  (flet ((get-param (p &optional (c cell))
	   (gethash (cons c p) param-alignment))
	 (set-param (p v) (setf (gethash (cons cell p) param-alignment) v)))
    (let ((need-alignment (remove-if #'get-param params)))
      (pprint need-alignment)
      (dolist (p need-alignment)
	;; Case 1: derived-parallel relations can be used to bridge
	(let* ((p-link (car (links-of-type p 'derived-parallel t)))
	       (p-cell (car (links-of-type cell 'subspace t)))
	       (p-align (get-param (when p-link (tail p-link))
				   (when p-cell (tail p-cell)))))
	  (when p-align (set-param p p-align))))
      ;; Case 2: only one parameter needs alignment -> process of elimination
      (when (= (length need-alignment) 1)
	(let ((last-root (set-difference root-params
					 (mapcar #'get-param params))))
	  (assert (= (length last-root) 1))
	  (set-param p (car last-root)))))
    ;; if all are now aligned, then sort to match root-params
    (when (not (find-if-not #'get-param params))
      (mapcar (lambda (rp) (find rp params :key #'get-param))
	      root-params))))

(defun resolve-if-possible (cell root-params)
  (let* ((cell-dims (align-parameters cell (cell-dimensions cell) root-params))
	 (foo (pprint (list (cell-dimensions cell) cell-dims)))
	 (lengths (mapcar #'dim-length cell-dims))
	 (vlist (vertices cell))
	 (anchor-vertex '()))
    (when cell-dims			; can't act if can't align
      ;; Case #1: a vertex is already coordinatized
      (setf anchor-vertex (find-if (ppf gethash ? coordinates) vlist))
      ;; Case #2: cell is a subspace of a resolved cell
      (unless anchor-vertex
	(let ((super (car (mapcar #'tail (links-of-type cell 'subspace t))))
	      (s-dims (align-parameters super (cell-dimensions super) root-params)))
	  (when (and super s-dims (find super resolved))
	    ;; Note: there's no shared vertices, else it would be Case #1
	    ;; I am, however, ignoring gap info now
	    (setf anchor-vertex (find-minimum-vertex cell))
	    (let* ((slengths (mapcar #'dim-length s-dims))
		   (gaps (mapcar (lambda (a b) (/ (- a b) 2)) slengths lengths))
		   (sminv (find-minimum-vertex super)))
	      (setf (gethash anchor-vertex coordinates)
		(mapcar #'+ (gethash sminv coordinates) gaps))))))
      ;; If we've got an anchor vertex, assign all the rest
      (when anchor-vertex
	(let ((base (gethash anchor-vertex coordinates)))
	  (dolist (v vlist)
	    (ensure (gethash v coordinates)
		    (mapcar #'+ base
			    (mapcar (ppf displace-if-max v) 
				    cell-dims lengths)))))
	(push cell resolved)))
    ;; return self if unresolved; nil if resolved
    (if anchor-vertex nil cell)))

(defun coordinatize-vertices (root-params)
  ;; turn any symbols into parameters
  (setf root-params
    (mapcar (lambda (p) (if (symbolp p) (find-parameter p) p)) root-params))
  ;; initialize coordinatization system
  (initialize-coordinatization-variables root-params)
  ;; Get all the k-cells, finding one to be the root
  (let* ((k-cells (find-cells `((dimension ?c ,(length root-params)))))
	 (root (find-if (pcf subsetp root-params) k-cells
			:key #'cell-dimensions)))
    (assert root (root-params) "No root cell spanned by ~A" root-params)
    (coordinatize-root root root-params)
    ;; Walk across cells until progress ceases
    (setf worklist (remove root k-cells))
    (while progress
      (setf progress nil)
      (setf worklist (remove nil (mapcar 
				  (ppf resolve-if-possible ?x root-params) 
				  worklist))))
    ;; Report degree of success
    (format t "Assigned ~A out of ~A vertices." (hash-table-count coordinates)
	    (length (find-cells '((dimension ?c 0)))))))
