;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;; return first bound variable 
;; This is going to start out terribly inefficient

;; Four clause types:
;;  (type element class)
;;  (property element value)
;;  (-> tail linktype head)
;;  (boundary* cell element)
;; e.g. ((type ?c cell) (dimension ?c 2) (-> y minimum ?c))

;; check clauses in order, aborting after the first that returns false
(defun check-clauses (clauselist)
  (not (position-if-not #'check-clause clauselist)))

(defun check-clause (clause)
  (cond ((eq (first clause) 'type) (check-type-clause (rest clause)))
	((eq (first clause) '->) (check-link-clause (rest clause)))
	((eq (first clause) 'boundaries*) (check-boundary*-clause (rest clause)))
	(t (check-property-clause clause))))

(defun check-type-clause (clause)	; assumes clause has marker stripped
  (assert (= (length clause) 2))
  (destructuring-bind (element class) clause
    (typep element class)))

(defun check-boundary*-clause (clause)	; assumes clause has marker stripped
  (assert (= (length clause) 2))
  (destructuring-bind (cell element) clause
    (position element (boundaries* cell))))

(defun check-property-clause (clause)
  (assert (= (length clause) 3))
  (destructuring-bind (property variable value) clause
    (eqv (slot-value variable property) value)))

(defun check-link-clause (clause)	; assumes clause has marker stripped
  (assert (= (length clause) 3))
  (destructuring-bind (tail type head) clause
    (let ((linkset (links-of-type tail type nil)))
      (find-if (lambda (l) (or (eq (head l) head) (eq (name (head l)) head)))
	       linkset))))

(defun instantiate-element (bindings symbol)
  (or (second (assoc symbol bindings))
      (find-parameter symbol)
      symbol))

(defun find-elements (pattern)
  (let* ((var-instances (remove-if-not #'call-variablep (flatten pattern)))
	 (variables (remove-duplicates var-instances)))
    (assert (= (length variables) 1) ()"Not yet handling multibinding patterns")
    (remove-if-not
     (lambda (e)
       (let* ((bindings (mapcar #'list variables (list e)))
	      (instance (transform-tree pattern 
					(pcf instantiate-element bindings) 
					#'list)))
	 (check-clauses instance)))
     (all-elements))))


;; Extrema is a list of (param-name +/-)
(defun keep-only-extrema (extrema cells)
  (let ((ht (make-hash-table)) (minmax (make-hash-table))
	(cs-names (mapcar #'first extrema))
	(results '()))
    ;; make search table
    (dolist (e extrema) (setf (gethash (first e) minmax) (eq '+ (second e))))
    ;; populate with all coordinate-values
    (dolist (c cells)
      (dolist (cv (coordinates c))
	(when (find (name (lattice (tail cv))) cs-names)
	  (setf (gethash (tail cv) ht) t))))
    ;; now check whether extrema for each class are there
    (dolist (c cells)
      (let ((c-good t) (c-left cs-names))
	(dolist (cv (coordinates c))
	  (let ((cs-name (name (lattice (tail cv)))))
	    (when (find cs-name cs-names)
	      ;; mark as noticed
	      (setf c-left (remove cs-name c-left))
	      (cond ((typep (tail cv) 'lattice-coordinate)
		     (if (gethash cs-name minmax) ; maximum
			 (dolist (ls (slot-value (tail cv) 'precedes))
			   (when (or (gethash ls ht) (gethash (upper ls) ht))
			     (setf c-good nil)))
		       (dolist (ls (slot-value (tail cv) 'succeeds))
			 (when (or (gethash ls ht) (gethash (lower ls) ht))
			   (setf c-good nil)))))
		    ((typep (tail cv) 'lattice-span)
		     (if (gethash cs-name minmax) ; maximum
			 (when (gethash (upper (tail cv)) ht) (setf c-good nil))
		       (when (gethash (lower (tail cv)) ht) (setf c-good nil))))
		    (t (error "Unknown value ~A" (tail cv)))))))
	(when (and c-good (not c-left)) (push c results))))
    results))

(defmethod find-cells (pattern &optional extrema)
  (let* ((var-instances (remove-if-not #'call-variablep (flatten pattern)))
	 (variables (remove-duplicates var-instances)))
    (assert (<= (length variables) 1) ()"Not yet handling multibinding patterns")
    (keep-only-extrema
     extrema
     (remove-if-not
      (lambda (e)
	(let* ((bindings (mapcar #'list variables (list e)))
	       (instance (transform-tree pattern 
					 (pcf instantiate-element bindings) 
					 #'list)))
	  (check-clauses instance)))
      (remove-if-not (ppf typep ? 'cell) (all-elements))))))

