;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;;; helper functions
(defun sameset (s1 s2) (not (set-exclusive-or s1 s2 :test #'eqv)))

;; Test whether current topology matches expectations
(defun check-stats (stat-values)
  (let ((stats (compute-topology-stats))
	(failures '()))
    (dolist (v stat-values)
      (let ((s (assoc (car v) stats)))
	(assert s (v) "Couldn't find stat ~A" (car v))
	(when (not (equal (cdr s) (cdr v))) (push s failures))))
    failures))

;; Test whether a graduated coordinate matches with the expected
;; pair of edges
(defun check-bounds (cell dimension coordinate b- b+)
  (let* ((gc (find-if (lambda (m) (and (eq (name (lattice (tail m))) dimension)
				       (typep m 'graduated-coordinate)))
		      (coordinates cell))))
    (assert (equal (name (tail gc)) coordinate) 
	(gc coordinate)	"Lattice-span ~A is not expected ~A" 
	(name (tail gc)) coordinate)
    (assert (if b- (eq b- (bound- gc)) t)
	(cell coordinate b-) "~A is not the lower bound of ~A in ~A~%~
                Lower bound is ~A" b- coordinate cell (bound- gc))
    (assert (if b+ (eq b+ (bound+ gc)) t)
	(cell coord b+) "~A is not the upper bound of ~A in ~A~%~
                Upper bound is ~A" b+ coordinate cell (bound+ gc))))
		     



;;; Cell & boundary tests
(defregression test-cells-and-boundaries
  (let* ((c0 (make-cell 0)) (c1 (make-cell 0)) (c2 (make-cell 0))
	 (e01 (make-cell 1 c0 c1)) (e12 (make-cell 1 c1 c2)) 
	 (e02 (make-cell 1 c0 c2)) (triangle (make-cell 2 e01 e12 e02)))
    (assert (sameset (boundaries c1) nil))
    (assert (sameset (boundaries e01) (list c0 c1)))
    (assert (sameset (boundaries e12) (list c2 c1)))
    (assert (sameset (boundaries triangle) (list e01 e02 e12)))
    (assert (sameset (boundaries* c2) nil))
    (assert (sameset (boundaries* e02) (list c0 c2)))
    (assert (sameset (boundaries* triangle) (list c0 c1 c2 e01 e02 e12)))
    'ok))

;;; Coordinate system tests
(defregression test-coordinate-system
  (let* ((cs (make-coordinate-system 'x))
	 (sub1 (refine-span (base-span cs) 3))
	 (sub2 (refine-span (second sub1) 2))
	 (sub3 (refine-span (base-span cs) 2))
	 (sub4 (refine-span (base-span cs) 1))
	 (sub5 (refine-span (second sub3) 4)))
    (assert (sameset (mapcar #'name (coordinate-values cs))
		     '("X-" "X+" "X.1A" "X.1B" "X.1A.1A" "X.2A" "X.2A.1A"
		       "X.2A.1B" "X.2A.1C")))
    (assert (sameset (mapcar #'name (coordinate-spans cs))
		     '("X-::X+" "X-::X.1A" "X.1A::X.1B" "X.1B::X+"
		       "X.1A::X.1A.1A" "X.1A.1A::X.1B" "X-::X.2A"
		       "X.2A::X+" "X.2A::X.2A.1A" "X.2A.1A::X.2A.1B"
		       "X.2A.1B::X.2A.1C" "X.2A.1C::X+")))
    'ok))

;;; Tests of extrusion (and measurement)
(defregression test-extrusion
  ;; Start with a simple test of constructing a rectangle
  (clear-topology)
  (let ((r (extrude-rectangle)))
    (check-stats
     `((0-cells 4) (1-cells 4) (2-cells 1)
       (dimensions 3) (parameters 3) (relations 3) (measurements 4)
       (boundaries 12) (coordinate-values 18) (graduated 6)
       (constraints 1)))
    (let ((x+y+ (find-cells '() '((x +) (y +))))
	  (x-y+ (find-cells '() '((x -) (y +))))
	  (x+y- (find-cells '() '((x +) (y -))))
	  (x-y- (find-cells '() '((x -) (y -))))
	  (x+ (find-cells '((dimension ? 1)) '((x +))))
	  (x- (find-cells '((dimension ? 1)) '((x -))))
	  (y+ (find-cells '((dimension ? 1)) '((y +))))
	  (y- (find-cells '((dimension ? 1)) '((y -))))
	  (xy (find-cells '((dimension ? 2)))))
      (assert (= (length x+y+) 1)) (assert (= (length x-y+) 1))
      (assert (= (length x+y-) 1)) (assert (= (length x-y-) 1))
      (assert (= (length x+) 1)) (assert (= (length x-) 1))
      (assert (= (length y+) 1)) (assert (= (length y-) 1))
      (assert (= (length xy) 1))
      (check-bounds (car xy) 'x "X-::X+" (car x-) (car x+))
      (check-bounds (car xy) 'y "Y-::Y+" (car y-) (car y+))
      (check-bounds (car x+) 'y "Y-::Y+" (car x+y-) (car x+y+))
      (check-bounds (car x-) 'y "Y-::Y+" (car x-y-) (car x-y+))
      (check-bounds (car y+) 'x "X-::X+" (car x-y+) (car x+y+))
      (check-bounds (car y-) 'x "X-::X+" (car x-y-) (car x+y-))))
  (clear-topology)
  (let ((r (extrude-rectangle)))
    (extrude (find-cells '() '((y +))) (ensure-dimension '|Distal| 'meters))
    ;; tests of coordinate boundaries
    (let ((x+d+ (car (find-cells '() '((x +) (|Distal| +)))))
	  (x-d+ (car (find-cells '() '((x -) (|Distal| +)))))
	  (d+ (car (find-cells '((dimension ? 1)) '((|Distal| +)))))
	  (d- (car (find-cells '((dimension ? 1)) '((|Distal| -)))))
	  (newcell (car (find-cells '((dimension ? 2)) '((y +))))))
      (check-bounds d+ 'x "X-::X+" x-d+ x+d+)
      (check-bounds newcell '|Distal| "Distal-::Distal+" d- d+)
      (check-bounds r 'y "Y-::Y+" nil d-)))
  'ok)
(output-graphviz )

;; Linear stress functions
(defregression test-stress-functions
  (let ((fn (linear-stress-fn -3 -1 4 5)))
    (assert (~= (funcall fn 4.3) 0.3 1e-4))
    (assert (~= (funcall fn 1) 0 1e-4))
    (assert (~= (funcall fn 7) 1 1e-4))
    (assert (~= (funcall fn -5) -1 1e-4))
    (assert (~= (funcall fn -2.5) -0.75 1e-4))
    (assert (~= (funcall fn -1) 0 1e-4)))
  'ok)
