;; Copyright (C) 2013-2014, Raytheon BBN Technologies and contributors listed
;; in the AUTHORS file in the MADV distribution's top directory.

;; This file is part of MADV, and is distributed under the terms of the
;; GNU General Public License, with a linking exception, as described in
;; the file LICENSE in the MADV distribution's top directory. 

;; Mechanisms for supporting topology change

(defclass component (relation)
  ((bounds :initarg :bounds :accessor bounds)
   (dimensions :initarg :dims :accessor dimensions)
   (cumulative :initarg :cum :accessor cumulative)
   (noncumulative :initarg :noncum :accessor noncumulative)
   ;; set parameter
   (multiplicity :initarg :mult :initform 1 :accessor multiplicity)
   (set-dimension :initarg :setdim :initform nil :accessor set-dimension)
   (set-relations :initarg :setrel :initform nil :accessor set-relations)
   ))

(defun bind-into-component (bounds dims cum noncum)
  (let ((b (mapcar #'find-relation bounds))
	(d (mapcar #'find-parameter dims))
	(c (mapcar #'find-parameter cum))
	(n (mapcar #'find-parameter noncum)))
    (assert (not (position nil (append b d c n))))
    (let ((cmp (make-instance 'component :bounds b :dims d :cum c :noncum n)))
      (mapcar (pcf make-constraint cmp) (append d c n))
      cmp)))
;; For some reason, whether the constraints are in can sometimes makes
;; a difference, even though they aren't doing anything... need to
;; check on that: it may indicate a duplicate normalization

(defun describe-component-stresses (component)
  (flet ((objstr (o) 
	   (format nil "~A = ~A~@[ ~A~]" (pretty-name o) (sigfig (abs-stress o))
		   (when (typep o 'parameter) (mapcar #'car (deltas o))))))
    (format t "Stresses on ~A:~%  Bounds:~{ ~A~}~%  Dimensions:~{ ~A~}~%  Cumulative:~{ ~A~}~%  Non-Cumulative:~{ ~A~}~%"
	    (pretty-name component)
	    (mapcar #'objstr (bounds component))
	    (mapcar #'objstr (dimensions component))
	    (mapcar #'objstr (cumulative component))
	    (mapcar #'objstr (noncumulative component)))))

;; components aren't directly stressed or resolved
(defmethod resolve ((r component))
  (setf (stress r) 0))

(defun all-components ()
  (remove-if-not (ppf typep ?x 'component) (all-relations)))

(defun high-stress (object)
  (and (> (abs-stress object) 0.01)
       (> (/ (abs-stress object) current-max-abs-stress) 0.1)))

;; returns a list of (component element +/- stress)
(defun find-candidate-changes ()
  (let ((candidates '()))
    (dolist (c (all-components))
      (dolist (b (bounds c))
	(when (high-stress b)
	  (with-slots (risers fallers) b
	    (dolist (p (append risers fallers))
	      (when (and (high-stress p) (not (stress-unopposed? p)))
		(let ((sign 
		       (if (eq (tp (find p risers)) (> (stress b) 0)) '- '+)))
		  (push (list c p sign (stress b)) candidates)))))))
      (dolist (d (dimensions c))
	;; not yet handling dimension-driven changes
	))
    (sort candidates #'> :key #'fourth)))

(defun shift-constraints (relation oldp newp)
  (dolist (ct (links-of-type relation 'constraint nil))
    (switch-link ct oldp newp)))

(defmethod switch-parameter ((r relation) (oldp parameter) (newp parameter))
  (error "don't know how to switch parameters for a generic relation"))

(defmethod switch-parameter :after ((r relation) (oldp parameter) 
				    (newp parameter))
  (shift-constraints r oldp newp))

(defmethod switch-parameter ((s summation) (oldp parameter) (newp parameter))
  (with-slots (total summands) s
    (assert (and (find oldp summands) (not (eq newp total))))
    (setf summands (subst newp oldp summands))))

(defmethod switch-parameter ((m modular-relation) (oldp parameter) (newp parameter))
  (with-slots (varsets risers fallers) m
    (assert (find oldp (flatten varsets)))
    (setf varsets (subst newp oldp varsets))
    (setf risers (subst newp oldp risers))
    (setf fallers (subst newp oldp fallers))))

(defmethod switch-parameter ((ub upper-bound) (oldp parameter) (newp parameter))
  (with-slots (lower upper) ub
    (assert (or (eq oldp lower) (eq oldp upper)))
    (when (eq oldp lower) (setf lower newp))
    (when (eq oldp upper) (setf upper newp))))

(defmethod split-parameter ((c component) (p parameter))
  ;; create the new set parameter
  (let* ((setp-name (intern (format nil "SET:~A" (name p))))
	 (setp (make-parameter setp-name (units p))))
    ;; walk through all existing relations and decide whether to shift to set
    (dolist (r (remove-duplicates (relations p)))
      (cond ((member r (bounds c)) 'do-nothing)
	    ((typep r 'component) 'do-nothing)
	    ((typep r 'unit-identity) 'do-nothing)
	    ((typep r 'identity-relation) 'do-nothing)
	    ((typep r 'user-blueprint) 'do-nothing)
	    ((typep r 'quantized-component) 'do-nothing)
	    ((typep r 'summation) (switch-parameter r p setp))
	    ((typep r 'composite-sum-bound) (switch-parameter r p setp))
	    ((typep r 'upper-bound) (switch-parameter r p setp))
	    ((typep r 'modular-relation) (switch-parameter r p setp))
	    (t (error "Don't know whether to change ~A to use set parameter"
		      (pretty-name r)))))
    ;; create a new relation linking it with the old
    ;; assume n = 2 because this is part of creating a component set
    (push (make-identity-relation '(2) (list p) (list setp)) (set-relations c))
    ))

(defmethod create-component-set ((c component) (split parameter))
  (format t "Splitting component ~A into two along dimension ~A~%" 
	  (pretty-name c) (pretty-name split))
  (assert (= (multiplicity c) 1))
  (assert (typep split 'dimension))
  (mapcar (pcf split-parameter c) (cons split (cumulative c)))
  (incf (multiplicity c)))

(defmethod increment-component-set ((c component))
  (format t "Adding another component ~A to set, bringing number to ~A~%"
	  (pretty-name c) (1+ (multiplicity c)))
  (assert (> (multiplicity c) 1))
  (dolist (r (set-relations c))
    (with-slots (constant) r 
      (assert (= constant (multiplicity c))) 
      (incf constant)))
  (incf (multiplicity c)))

(defmethod decrement-component-set ((c component))
  (format t "Removing a component ~A from set, bringing number to ~A~%"
	  (pretty-name c) (1- (multiplicity c)))
  (assert (> (multiplicity c) 2))
  (dolist (r (set-relations c)) 
    (with-slots (constant) r 
      (assert (= constant (multiplicity c)))
      (decf constant)))
  (decf (multiplicity c)))

(defmethod merge-parameter ((setp parameter) (basep parameter))
  ;; redirect everything pointing at the set to the base
  (dolist (r (remove-duplicates (relations setp)))
    (switch-parameter r setp basep))
  ;; remove all other links as well
  (dolist (l (links setp)) (delete-link l))
  ;; now that all links are gone, delete the set parameter
  (delete-element setp))

(defmethod collapse-component-set ((c component))
  (format t "Collapsing set component ~A into single component~%"
	  (pretty-name c))
  (assert (= (multiplicity c) 2))
  ;; for each set-relation, merge the set parmeter into the base parameter
  (dolist (r (set-relations c))
    (with-slots (risers fallers constant) r
      (assert (and (= constant 2) (= (length risers) 1) (= (length fallers) 1)))
      (let ((setp (first risers)) (basep (first fallers)))
	(merge-parameter setp basep)))
    ;; delete the set relation
    (dolist (l (links r)) (delete-link l))
    (delete-element r))
  ;; leave the component "set-less"
  (setf (set-relations c) nil (set-dimension c) nil)
  (decf (multiplicity c)))

(defun choose-split-dimension (component parameter)
  (assert (dimensions component))
  (if (typep parameter 'dimension)
      parameter
    (first (sort (dimensions component) #'< :key #'abs-stress))))

(defun make-topological-change (component parameter dir)
  (let ((n (multiplicity component)))
    (cond 
     ((eq dir '-)			; decrease set --> remove a component
      (cond ((= n 1) (error "Cannot reduce cardinality of one-component set"))
	    ((= n 2) (collapse-component-set component))
	    (t (decrement-component-set component))))
     ((eq dir '+)			; increase set --> add a component
      (cond ((= n 1)
	     (create-component-set 
	      component (choose-split-dimension component parameter)))
	    (t (increment-component-set component))))
     (t (error "Don't know how to change in direction '~A'" dir)))))

(defun apply-best-candidate-change ()
  (let ((candidates (find-candidate-changes)))
    (when candidates
      (let ((c (first candidates)))
	(make-topological-change (first c) (second c) (third c))))
    (if candidates 'done 'none)))

(defun run-with-topochange-until-stuck (&optional max-sets)
  (let ((problems (diagnose t)))
    (when problems 
      (let ((chart (run-batches-until-stuck))
	    (rest
	     (when (diagnose t)		; still problems
	       (let ((change (apply-best-candidate-change)))
		 (cond ((not (eq change 'none))
			(run-with-topochange-until-stuck max-sets))
		       ((and max-sets (plusp max-sets))
			(run-with-topochange-until-stuck (1- max-sets)))
		       (t nil))))))
	(merge-charts chart rest)))))
